@@ -15,21 +15,18 @@
 #include <sys/time.h>
 
 #include "collie.h"
+#include "farm/farm.h"
 
 static struct sd_option cluster_options[] = {
 	{'b', "store", true, "specify backend store"},
 	{'c', "copies", true, "specify the default data redundancy (number of copies)"},
 	{'m', "mode", true, "mode (safe, quorum, unsafe)"},
 	{'f', "force", false, "do not prompt for confirmation"},
-	{'R', "restore", true, "restore the cluster"},
-	{'l', "list", false, "list the user epoch information"},
 
 	{ 0, NULL, false, NULL },
 };
 
 static struct cluster_cmd_data {
-	uint32_t epoch;
-	bool list;
 	int copies;
 	bool nohalt;
 	bool quorum;
@@ -37,29 +34,21 @@ static struct cluster_cmd_data {
 	char name[STORE_LEN];
 } cluster_cmd_data;
 
-#define DEFAULT_STORE	"farm"
+#define DEFAULT_STORE	"plain"
 
 static int list_store(void)
 {
-	int fd, ret;
+	int ret;
 	struct sd_req hdr;
 	struct sd_rsp *rsp = (struct sd_rsp *)&hdr;
 	char buf[512] = { 0 };
 
-	fd = connect_to(sdhost, sdport);
-	if (fd < 0)
-		return EXIT_SYSFAIL;
-
 	sd_init_req(&hdr, SD_OP_GET_STORE_LIST);
 	hdr.data_length = 512;
 
-	ret = collie_exec_req(fd, &hdr, buf);
-	close(fd);
-
-	if (ret) {
-		fprintf(stderr, "Failed to connect\n");
+	ret = collie_exec_req(sdhost, sdport, &hdr, buf);
+	if (ret < 0)
 		return EXIT_SYSFAIL;
-	}
 
 	if (rsp->result != SD_RES_SUCCESS) {
 		fprintf(stderr, "Restore failed: %s\n",
@@ -73,68 +62,47 @@ static int list_store(void)
 	return EXIT_SYSFAIL;
 }
 
-#define FORMAT_PRINT \
-"\
-    __   \n\
-   ()'`; \n\
-   /\\|`  \n\
-  /  |   Caution! The cluster is not empty.\n\
-(/_)_|_  Are you sure you want to continue? [yes/no]: "
+static bool no_vdi(const unsigned long *vdis)
+{
+	return find_next_bit(vdis, SD_NR_VDIS, 0) == SD_NR_VDIS;
+}
+
+#define FORMAT_PRINT				\
+	"    __\n"				\
+	"   ()'`;\n"				\
+	"   /\\|`\n"				\
+	"  /  |   Caution! The cluster is not empty.\n" \
+	"(/_)_|_  Are you sure you want to continue? [yes/no]: "
 
 static int cluster_format(int argc, char **argv)
 {
-	int fd, ret;
-	struct sd_so_req hdr;
-	struct sd_so_rsp *rsp = (struct sd_so_rsp *)&hdr;
+	int ret;
+	struct sd_req hdr;
+	struct sd_rsp *rsp = (struct sd_rsp *)&hdr;
 	struct timeval tv;
 	char store_name[STORE_LEN];
-	DECLARE_BITMAP(vdi_inuse, SD_NR_VDIS);
-	unsigned long nr;
+	static DECLARE_BITMAP(vdi_inuse, SD_NR_VDIS);
 
-	fd = connect_to(sdhost, sdport);
-	if (fd < 0)
-		return EXIT_SYSFAIL;
-
-	sd_init_req((struct sd_req *)&hdr, SD_OP_READ_VDIS);
+	sd_init_req(&hdr, SD_OP_READ_VDIS);
 	hdr.data_length = sizeof(vdi_inuse);
 
-	ret = collie_exec_req(fd, (struct sd_req *)&hdr, &vdi_inuse);
-	if (ret < 0) {
-		fprintf(stderr, "Failed to read VDIs from %s:%d\n",
-			sdhost, sdport);
-		close(fd);
+	ret = collie_exec_req(sdhost, sdport, &hdr, &vdi_inuse);
+	if (ret < 0)
 		return EXIT_SYSFAIL;
-	}
 
-	for (nr = 0; nr < SD_NR_VDIS; nr++)
-		if (test_bit(nr, vdi_inuse))
-			break;
-
-	if (nr != SD_NR_VDIS) {
-		int i, l;
-		char str[123] = {'\0'};
-
-		printf(FORMAT_PRINT);
-		ret = scanf("%s", str);
-		if (ret < 0)
-			return EXIT_SYSFAIL;
-		l = strlen(str);
-		for (i = 0; i < l; i++)
-			str[i] = tolower(str[i]);
-		if (strncmp(str, "yes", 3) != 0)
-			return EXIT_SUCCESS;
-	}
+	if (!no_vdi(vdi_inuse))
+		confirm(FORMAT_PRINT);
 
 	gettimeofday(&tv, NULL);
 
-	sd_init_req((struct sd_req *)&hdr, SD_OP_MAKE_FS);
-	hdr.copies = cluster_cmd_data.copies;
+	sd_init_req(&hdr, SD_OP_MAKE_FS);
+	hdr.cluster.copies = cluster_cmd_data.copies;
 	if (cluster_cmd_data.nohalt)
 		hdr.flags |= SD_FLAG_NOHALT;
 	if (cluster_cmd_data.quorum)
 		hdr.flags |= SD_FLAG_QUORUM;
 
-	hdr.ctime = (uint64_t) tv.tv_sec << 32 | tv.tv_usec * 1000;
+	hdr.cluster.ctime = (uint64_t) tv.tv_sec << 32 | tv.tv_usec * 1000;
 
 	if (strlen(cluster_cmd_data.name))
 		pstrcpy(store_name, STORE_LEN, cluster_cmd_data.name);
@@ -144,13 +112,9 @@ static int cluster_format(int argc, char **argv)
 	hdr.flags |= SD_FLAG_CMD_WRITE;
 
 	printf("using backend %s store\n", store_name);
-	ret = collie_exec_req(fd, (struct sd_req *)&hdr, store_name);
-	close(fd);
-
-	if (ret) {
-		fprintf(stderr, "Failed to connect\n");
+	ret = collie_exec_req(sdhost, sdport, &hdr, store_name);
+	if (ret < 0)
 		return EXIT_SYSFAIL;
-	}
 
 	if (rsp->result != SD_RES_SUCCESS) {
 		fprintf(stderr, "Format failed: %s\n",
@@ -166,7 +130,7 @@ static int cluster_format(int argc, char **argv)
 
 static int cluster_info(int argc, char **argv)
 {
-	int i, fd, ret;
+	int i, ret;
 	struct sd_req hdr;
 	struct sd_rsp *rsp = (struct sd_rsp *)&hdr;
 	struct epoch_log *logs;
@@ -176,34 +140,20 @@ static int cluster_info(int argc, char **argv)
 	char time_str[128];
 
 	log_length = sd_epoch * sizeof(struct epoch_log);
-again:
-	logs = malloc(log_length);
-	if (!logs) {
-		if (log_length < 10) {
-			fprintf(stderr, "No memory to allocate.\n");
-			return EXIT_SYSFAIL;
-		}
-		log_length /= 2;
-		goto again;
-	}
-
-	fd = connect_to(sdhost, sdport);
-	if (fd < 0)
-		goto error;
+	logs = xmalloc(log_length);
 
 	sd_init_req(&hdr, SD_OP_STAT_CLUSTER);
 	hdr.data_length = log_length;
 
-	ret = collie_exec_req(fd, &hdr, logs);
-	close(fd);
-
-	if (ret != 0)
+	ret = collie_exec_req(sdhost, sdport, &hdr, logs);
+	if (ret < 0)
 		goto error;
 
 	if (!raw_output)
 		printf("Cluster status: ");
 	if (rsp->result == SD_RES_SUCCESS)
-		printf("running\n");
+		printf("running, auto-recovery %s\n", logs->disable_recovery ?
+		       "disabled" : "enabled");
 	else
 		printf("%s\n", sd_strerror(rsp->result));
 
@@ -262,117 +212,158 @@ static int cluster_shutdown(int argc, char **argv)
 	return EXIT_SUCCESS;
 }
 
-static int restore_snap(uint32_t epoch)
-{
-	int ret;
-	struct sd_req hdr;
-
-	sd_init_req(&hdr, SD_OP_RESTORE);
-	hdr.obj.tgt_epoch = epoch;
-
-	ret = send_light_req(&hdr, sdhost, sdport);
-	if (ret) {
-		fprintf(stderr, "failed to execute request\n");
-		return EXIT_FAILURE;
-	}
-
-	printf("Cluster restore to the snapshot %d\n", epoch);
-	return EXIT_SUCCESS;
-}
-
 static void print_list(void *buf, unsigned len)
 {
 	struct snap_log *log_buf = (struct snap_log *)buf;
-	unsigned nr = len / sizeof(struct snap_log), i;
+	unsigned nr = len / sizeof(struct snap_log);
 
-	printf("Index\t\tSnapshot Time\n");
-	for (i = 0; i < nr; i++, log_buf++) {
+	printf("Index\t\tTag\t\tSnapshot Time\n");
+	for (unsigned i = 0; i < nr; i++, log_buf++) {
 		time_t *t = (time_t *)&log_buf->time;
-		printf("%d\t\t", log_buf->epoch);
+		printf("%d\t\t", log_buf->idx);
+		printf("%s\t\t", log_buf->tag);
 		printf("%s", ctime(t));
 	}
 }
 
-static int list_snap(void)
+static int list_snapshot(int argc, char **argv)
 {
-	int fd, ret = EXIT_SYSFAIL;
-	struct sd_req hdr;
-	struct sd_rsp *rsp = (struct sd_rsp *)&hdr;
-	void *buf;
+	char *path = argv[optind++];
+	void *buf = NULL;
+	int log_nr;
+	int ret = EXIT_SYSFAIL;
 
-	buf = malloc(SD_DATA_OBJ_SIZE);
+	if (farm_init(path) != SD_RES_SUCCESS)
+		goto out;
+
+	buf = snap_log_read(&log_nr);
 	if (!buf)
-		return EXIT_SYSFAIL;
-
-	fd = connect_to(sdhost, sdport);
-	if (fd < 0)
 		goto out;
 
-	sd_init_req(&hdr, SD_OP_GET_SNAP_FILE);
-	hdr.data_length = SD_DATA_OBJ_SIZE;
-
-	ret = collie_exec_req(fd, &hdr, buf);
-	close(fd);
-
-	if (ret) {
-		fprintf(stderr, "Failed to connect\n");
-		goto out;
-	}
-
-	if (rsp->result != SD_RES_SUCCESS) {
-		fprintf(stderr, "Listing snapshots failed: %s\n",
-				sd_strerror(rsp->result));
-		ret = EXIT_FAILURE;
-		goto out;
-	}
-
-	print_list(buf, rsp->data_length);
+	print_list(buf, log_nr * sizeof(struct snap_log));
+	ret = EXIT_SUCCESS;
 out:
+	if (ret)
+		fprintf(stderr, "Fail to list snapshot.\n");
 	free(buf);
-	return EXIT_SUCCESS;
+	return ret;
 }
 
-static int do_snapshot(void)
+static void fill_object_tree(uint32_t vid, const char *name, const char *tag,
+			  uint32_t snapid, uint32_t flags,
+			  const struct sd_inode *i, void *data)
 {
-	int ret;
-	struct sd_req hdr;
+	uint64_t vdi_oid = vid_to_vdi_oid(vid);
 
-	sd_init_req(&hdr, SD_OP_SNAPSHOT);
+	/* ignore active vdi */
+	if (!vdi_is_snapshot(i))
+		return;
 
-	ret = send_light_req(&hdr, sdhost, sdport);
-	if (ret) {
-		fprintf(stderr, "failed to execute request\n");
-		return EXIT_FAILURE;
+	object_tree_insert(vdi_oid, i->nr_copies);
+
+	for (uint64_t idx = 0; idx < MAX_DATA_OBJS; idx++) {
+		if (i->data_vdi_id[idx]) {
+			uint64_t oid = vid_to_data_oid(i->data_vdi_id[idx],
+						       idx);
+			object_tree_insert(oid, i->nr_copies);
+		}
+	}
+}
+
+static int save_snapshot(int argc, char **argv)
+{
+	char *tag = argv[optind++];
+	char *path, *p;
+	int ret = EXIT_SYSFAIL, uninitialized_var(unused);
+
+	unused = strtol(tag, &p, 10);
+	if (tag != p) {
+		fprintf(stderr, "Tag should not start with number.\n");
+		return EXIT_USAGE;
 	}
 
-	return EXIT_SUCCESS;
+	if (!argv[optind]) {
+		fprintf(stderr, "Please specify the path to save snapshot.\n");
+		return EXIT_USAGE;
+	}
+	path = argv[optind];
+
+	if (farm_init(path) != SD_RES_SUCCESS)
+		goto out;
+
+	if (farm_contain_snapshot(0, tag)) {
+		fprintf(stderr, "Snapshot tag has already been used for another"
+			" snapshot, please, use another one.\n");
+		goto out;
+	}
+
+	if (parse_vdi(fill_object_tree, SD_INODE_SIZE, NULL) != SD_RES_SUCCESS)
+		goto out;
+
+	if (farm_save_snapshot(tag) != SD_RES_SUCCESS)
+		goto out;
+
+	ret = EXIT_SUCCESS;
+out:
+	if (ret)
+		fprintf(stderr, "Fail to save snapshot to path: %s.\n", path);
+	object_tree_free();
+	return ret;
 }
 
-static int cluster_snapshot(int argc, char **argv)
+static int load_snapshot(int argc, char **argv)
 {
-	int ret, epoch = cluster_cmd_data.epoch;
-	if (epoch)
-		ret = restore_snap(epoch);
-	else if (cluster_cmd_data.list)
-		ret = list_snap();
-	else
-		ret = do_snapshot();
+	char *tag = argv[optind++];
+	char *path, *p;
+	uint32_t idx;
+	int ret = EXIT_SYSFAIL;
+
+	idx = strtol(tag, &p, 10);
+	if (tag == p)
+		idx = 0;
+
+	if (!argv[optind]) {
+		fprintf(stderr, "Please specify the path to save snapshot.\n");
+		return EXIT_USAGE;
+	}
+	path = argv[optind];
+
+	if (farm_init(path) != SD_RES_SUCCESS)
+		goto out;
+
+	if (!farm_contain_snapshot(idx, tag)) {
+		fprintf(stderr, "Snapshot index or tag does not exist.\n");
+		goto out;
+	}
+
+	if (cluster_format(0, NULL) != SD_RES_SUCCESS)
+		goto out;
+
+	if (farm_load_snapshot(idx, tag) != SD_RES_SUCCESS)
+		goto out;
+
+	ret = EXIT_SUCCESS;
+out:
+	if (ret)
+		fprintf(stderr, "Fail to load snapshot\n");
 	return ret;
 }
 
 #define RECOVER_PRINT \
-"Caution! Please try starting all the cluster nodes normally before\n\
-running this command.\n\n\
-The cluster may need to be force recovered if:\n\
-  - the master node fails to start because of epoch mismatch; or\n\
-  - some nodes fail to start after a cluster shutdown.\n\n\
-Are you sure you want to continue? [yes/no]: "
+	"Caution! Please try starting all the cluster nodes normally before\n" \
+	"running this command.\n\n" \
+	"The cluster may need to be force recovered if:\n" \
+	"  - the master node fails to start because of epoch mismatch; or\n" \
+	"  - some nodes fail to start after a cluster shutdown.\n\n" \
+	"Are you sure you want to continue? [yes/no]: "
 
 static int cluster_force_recover(int argc, char **argv)
 {
 	int ret;
 	struct sd_req hdr;
+	struct sd_rsp *rsp = (struct sd_rsp *)&hdr;
 	char str[123] = {'\0'};
+	struct sd_node nodes[SD_MAX_NODES];
 
 	if (!cluster_cmd_data.force) {
 		int i, l;
@@ -388,10 +379,15 @@ static int cluster_force_recover(int argc, char **argv)
 	}
 
 	sd_init_req(&hdr, SD_OP_FORCE_RECOVER);
+	hdr.data_length = sizeof(nodes);
 
-	ret = send_light_req(&hdr, sdhost, sdport);
-	if (ret) {
-		fprintf(stderr, "failed to execute request\n");
+	ret = collie_exec_req(sdhost, sdport, &hdr, nodes);
+	if (ret < 0)
+		return EXIT_SYSFAIL;
+
+	if (rsp->result != SD_RES_SUCCESS) {
+		fprintf(stderr, "failed to execute request, %s\n",
+			sd_strerror(rsp->result));
 		return EXIT_FAILURE;
 	}
 
@@ -442,15 +438,35 @@ static struct subcommand cluster_recover_cmd[] = {
 
 static int cluster_recover(int argc, char **argv)
 {
-	int i;
+	return do_generic_subcommand(cluster_recover_cmd, argc, argv);
+}
 
-	for (i = 0; cluster_recover_cmd[i].name; i++) {
-		if (!strcmp(cluster_recover_cmd[i].name, argv[3]))
-			return cluster_recover_cmd[i].fn(argc, argv);
-	}
+/* Subcommand list of snapshot */
+static struct subcommand cluster_snapshot_cmd[] = {
+	{"save", NULL, "h", "save snapshot to localpath",
+	 NULL, SUBCMD_FLAG_NEED_ARG, save_snapshot, NULL},
+	{"list", NULL, "h", "list snapshot of localpath",
+	 NULL, SUBCMD_FLAG_NEED_ARG, list_snapshot, NULL},
+	{"load", NULL, "h", "list snapshot of localpath",
+	 NULL, SUBCMD_FLAG_NEED_ARG, load_snapshot, NULL},
+	{NULL},
+};
 
-	subcommand_usage(argv[1], argv[2], EXIT_FAILURE);
-	return EXIT_FAILURE;
+static int cluster_snapshot(int argc, char **argv)
+{
+	return do_generic_subcommand(cluster_snapshot_cmd, argc, argv);
+}
+
+static int cluster_reweight(int argc, char **argv)
+{
+	int ret;
+	struct sd_req hdr;
+
+	sd_init_req(&hdr, SD_OP_REWEIGHT);
+	ret = send_light_req(&hdr, sdhost, sdport);
+	if (ret)
+		return EXIT_FAILURE;
+	return EXIT_SUCCESS;
 }
 
 static struct subcommand cluster_cmd[] = {
@@ -460,12 +476,15 @@ static struct subcommand cluster_cmd[] = {
 	 NULL, 0, cluster_format, cluster_options},
 	{"shutdown", NULL, "aph", "stop Sheepdog",
 	 NULL, 0, cluster_shutdown, cluster_options},
-	{"snapshot", NULL, "aRlph", "snapshot/restore the cluster",
-	 NULL, 0, cluster_snapshot, cluster_options},
+	{"snapshot", "<tag|idx> <path>", "aph", "snapshot/restore the cluster",
+	 cluster_snapshot_cmd, SUBCMD_FLAG_NEED_ARG,
+	 cluster_snapshot, cluster_options},
 	{"recover", NULL, "afph",
 	 "See 'collie cluster recover' for more information\n",
-	 cluster_recover_cmd, SUBCMD_FLAG_NEED_THIRD_ARG,
+	 cluster_recover_cmd, SUBCMD_FLAG_NEED_ARG,
 	 cluster_recover, cluster_options},
+	{"reweight", NULL, "aph", "reweight the cluster", NULL, 0,
+	 cluster_reweight, cluster_options},
 	{NULL,},
 };
 
@@ -509,20 +528,6 @@ static int cluster_parser(int ch, char *opt)
 	case 'f':
 		cluster_cmd_data.force = true;
 		break;
-	case 'R':
-		cluster_cmd_data.epoch = strtol(opt, &p, 10);
-		if (opt == p) {
-			fprintf(stderr, "The epoch must be an integer\n");
-			exit(EXIT_FAILURE);
-		}
-		if (cluster_cmd_data.epoch < 1) {
-			fprintf(stderr, "The epoch must be greater than 0\n");
-			exit(EXIT_FAILURE);
-		}
-		break;
-	case 'l':
-		cluster_cmd_data.list = true;
-		break;
 	}
 
 	return 0;
