@@ -11,13 +11,17 @@
 #include <inttypes.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <signal.h>
 
 #include "sheepdog_proto.h"
 #include "sheep.h"
 #include "collie.h"
+#include "util.h"
+
+#define EPOLL_SIZE 4096
 
 static const char program_name[] = "collie";
-const char *sdhost = "localhost";
+const char *sdhost = "127.0.0.1";
 int sdport = SD_LISTEN_PORT;
 bool highlight = true;
 bool raw_output;
@@ -27,8 +31,8 @@ static const struct sd_option collie_options[] = {
 	/* common options for all collie commands */
 	{'a', "address", true, "specify the daemon address (default: localhost)"},
 	{'p', "port", true, "specify the daemon port"},
-	{'r', "raw", false, "raw output mode: omit headers, separate fields with\n\
-                          single spaces and print all sizes in decimal bytes"},
+	{'r', "raw", false, "raw output mode: omit headers, separate fields with\n"
+	 "                          single spaces and print all sizes in decimal bytes"},
 	{'h', "help", false, "display this help and exit"},
 
 	{ 0, NULL, false, NULL },
@@ -43,31 +47,24 @@ struct sd_vnode sd_vnodes[SD_MAX_VNODES];
 int sd_nodes_nr, sd_vnodes_nr;
 unsigned master_idx;
 
-static int update_node_list(int max_nodes, uint32_t epoch)
+int update_node_list(int max_nodes)
 {
-	int fd, ret;
+	int ret;
 	unsigned int size;
 	char *buf = NULL;
 	struct sd_node *ent;
-	struct sd_node_req hdr;
-	struct sd_node_rsp *rsp = (struct sd_node_rsp *)&hdr;
-
-	fd = connect_to(sdhost, sdport);
-	if (fd < 0)
-		return -1;
+	struct sd_req hdr;
+	struct sd_rsp *rsp = (struct sd_rsp *)&hdr;
 
 	size = sizeof(*ent) * max_nodes;
 	buf = xzalloc(size);
-	sd_init_req((struct sd_req *)&hdr, SD_OP_GET_NODE_LIST);
-	hdr.request_ver = epoch;
+	sd_init_req(&hdr, SD_OP_GET_NODE_LIST);
 
 	hdr.data_length = size;
 
-	ret = collie_exec_req(fd, (struct sd_req *)&hdr, buf);
-	if (ret) {
-		ret = -1;
+	ret = collie_exec_req(sdhost, sdport, &hdr, buf);
+	if (ret < 0)
 		goto out;
-	}
 
 	if (rsp->result != SD_RES_SUCCESS) {
 		fprintf(stderr, "Failed to update node list: %s\n",
@@ -92,12 +89,10 @@ static int update_node_list(int max_nodes, uint32_t epoch)
 	memcpy(sd_nodes, buf, size);
 	sd_vnodes_nr = nodes_to_vnodes(sd_nodes, sd_nodes_nr, sd_vnodes);
 	sd_epoch = hdr.epoch;
-	master_idx = rsp->master_idx;
+	master_idx = rsp->node.master_idx;
 out:
 	if (buf)
 		free(buf);
-	if (fd >= 0)
-		close(fd);
 
 	return ret;
 }
@@ -143,7 +138,7 @@ static void init_commands(const struct command **commands)
 	};
 
 	if (!cmds) {
-		cmds = (struct command *)malloc(sizeof(command_list));
+		cmds = (struct command *)xmalloc(sizeof(command_list));
 		memcpy(cmds, command_list, sizeof(command_list));
 	}
 
@@ -190,11 +185,15 @@ static unsigned long setup_commands(const struct command *commands,
 	}
 
 	if (!found) {
-		fprintf(stderr, "Invalid command '%s'\n", cmd);
-		usage(commands, EXIT_USAGE);
+		if (cmd && strcmp(cmd, "help") && strcmp(cmd, "--help") &&
+		    strcmp(cmd, "-h")) {
+			fprintf(stderr, "Invalid command '%s'\n", cmd);
+			usage(commands, EXIT_USAGE);
+		}
+		usage(commands, 0);
 	}
 
-	for (s = commands[i].sub; s->name; s++) {
+	for (s = commands[i].sub; subcmd && s->name; s++) {
 		if (!strcmp(s->name, subcmd)) {
 			command_fn = s->fn;
 			command_opts = s->opts;
@@ -207,7 +206,8 @@ static unsigned long setup_commands(const struct command *commands,
 	}
 
 	if (!command_fn) {
-		if (strcmp(subcmd, "help") && strcmp(subcmd, "--help"))
+		if (subcmd && strcmp(subcmd, "help") &&
+		    strcmp(subcmd, "--help") && strcmp(subcmd, "-h"))
 			fprintf(stderr, "Invalid command '%s %s'\n", cmd, subcmd);
 		fprintf(stderr, "Available %s commands:\n", cmd);
 		for (s = commands[i].sub; s->name; s++)
@@ -232,7 +232,8 @@ static void usage(const struct command *commands, int status)
 		printf("\nAvailable commands:\n");
 		for (i = 0; commands[i].name; i++) {
 			for (s = commands[i].sub; s->name; s++) {
-				sprintf(name, "%s %s", commands[i].name, s->name);
+				snprintf(name, sizeof(name), "%s %s",
+					 commands[i].name, s->name);
 				printf("  %-24s%s\n", name, s->desc);
 			}
 		}
@@ -290,7 +291,8 @@ void subcommand_usage(char *cmd, char *subcmd, int status)
 	printf("Options:\n");
 	for (i = 0; i < len; i++) {
 		sd_opt = find_opt(command_opts[i]);
-		sprintf(name, "-%c, --%s", sd_opt->ch, sd_opt->name);
+		snprintf(name, sizeof(name), "-%c, --%s",
+			 sd_opt->ch, sd_opt->name);
 		printf("  %-24s%s\n", name, sd_opt->desc);
 	}
 
@@ -310,6 +312,28 @@ static const struct sd_option *build_sd_options(const char *opts)
 	return sd_opts;
 }
 
+static void crash_handler(int signo)
+{
+	fprintf(stderr, "collie exits unexpectedly (%s).\n", strsignal(signo));
+
+	sd_backtrace();
+
+	/*
+	 * OOM raises SIGABRT in xmalloc but the administrator expects
+	 * that collie exits with EXIT_SYSFAIL.  We have to give up
+	 * dumping a core file in this case.
+	 */
+	if (signo == SIGABRT)
+		exit(EXIT_SYSFAIL);
+
+	reraise_crash_signal(signo, EXIT_SYSFAIL);
+}
+
+static size_t get_nr_nodes(void)
+{
+	return sd_nodes_nr;
+}
+
 int main(int argc, char **argv)
 {
 	int ch, longindex, ret;
@@ -320,9 +344,11 @@ int main(int argc, char **argv)
 	char *p;
 	const struct sd_option *sd_opts;
 
+	install_crash_handler(crash_handler);
+
 	init_commands(&commands);
 
-	if (argc < 3)
+	if (argc < 2)
 		usage(commands, 0);
 
 	flags = setup_commands(commands, argv[1], argv[2]);
@@ -365,19 +391,30 @@ int main(int argc, char **argv)
 		}
 	}
 
-	if (!isatty(STDOUT_FILENO) || raw_output)
+	if (!is_stdout_console() || raw_output)
 		highlight = false;
 
 	if (flags & SUBCMD_FLAG_NEED_NODELIST) {
-		ret = update_node_list(SD_MAX_NODES, 0);
+		ret = update_node_list(SD_MAX_NODES);
 		if (ret < 0) {
 			fprintf(stderr, "Failed to get node list\n");
 			exit(EXIT_SYSFAIL);
 		}
 	}
 
-	if (flags & SUBCMD_FLAG_NEED_THIRD_ARG && argc == optind)
+	if (flags & SUBCMD_FLAG_NEED_ARG && argc == optind)
 		subcommand_usage(argv[1], argv[2], EXIT_USAGE);
 
-	return command_fn(argc, argv);
+	if (init_event(EPOLL_SIZE) < 0)
+		exit(EXIT_SYSFAIL);
+
+	if (init_work_queue(get_nr_nodes, NULL, NULL) != 0) {
+		fprintf(stderr, "Failed to init work queue\n");
+		exit(EXIT_SYSFAIL);
+	}
+
+	ret = command_fn(argc, argv);
+	if (ret == EXIT_USAGE)
+		subcommand_usage(argv[1], argv[2], EXIT_USAGE);
+	return ret;
 }
