@@ -21,12 +21,11 @@
 #include "sheep.h"
 #include "exits.h"
 #include "option.h"
+#include "work.h"
+#include "event.h"
 
 #define SUBCMD_FLAG_NEED_NODELIST (1 << 0)
-#define SUBCMD_FLAG_NEED_THIRD_ARG (1 << 1)
-
-#define TEXT_NORMAL "\033[0m"
-#define TEXT_BOLD   "\033[1m"
+#define SUBCMD_FLAG_NEED_ARG (1 << 1)
 
 #define UINT64_DECIMAL_SIZE 21
 
@@ -59,21 +58,28 @@ extern struct sd_vnode sd_vnodes[SD_MAX_VNODES];
 extern int sd_nodes_nr, sd_vnodes_nr;
 extern unsigned master_idx;
 
-bool is_current(const struct sheepdog_inode *i);
+bool is_current(const struct sd_inode *i);
 char *size_to_str(uint64_t _size, char *str, int str_size);
 typedef void (*vdi_parser_func_t)(uint32_t vid, const char *name,
 				  const char *tag, uint32_t snapid,
 				  uint32_t flags,
-				  const struct sheepdog_inode *i, void *data);
+				  const struct sd_inode *i, void *data);
 int parse_vdi(vdi_parser_func_t func, size_t size, void *data);
 int sd_read_object(uint64_t oid, void *data, unsigned int datalen,
 		   uint64_t offset, bool direct);
 int sd_write_object(uint64_t oid, uint64_t cow_oid, void *data,
 		    unsigned int datalen, uint64_t offset, uint32_t flags,
 		    int copies, bool create, bool direct);
+int collie_exec_req(const char *host, int port, struct sd_req *hdr, void *data);
 int send_light_req(struct sd_req *hdr, const char *host, int port);
-int send_light_req_get_response(struct sd_req *hdr, const char *host, int port);
-int collie_exec_req(int sockfd, struct sd_req *hdr, void *data);
+int do_generic_subcommand(struct subcommand *sub, int argc, char **argv);
+int update_node_list(int max_nodes);
+void confirm(const char *message);
+void work_queue_wait(struct work_queue *q);
+int do_vdi_create(const char *vdiname, int64_t vdi_size,
+		  uint32_t base_vid, uint32_t *vdi_id, bool snapshot,
+		  int nr_copies);
+
 
 extern struct command vdi_command;
 extern struct command node_command;
