@@ -14,6 +14,7 @@
 #include <sys/time.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <sys/ioctl.h>
 #include <fcntl.h>
 
 #include "collie.h"
@@ -28,6 +29,7 @@ static struct sd_option vdi_options[] = {
 	{'w', "writeback", false, "use writeback mode"},
 	{'c', "copies", true, "specify the data redundancy (number of copies)"},
 	{'F', "from", true, "create a differential backup from the snapshot"},
+	{'f', "force", false, "do operation forcibly"},
 	{ 0, NULL, false, NULL },
 };
 
@@ -42,6 +44,7 @@ static struct vdi_cmd_data {
 	bool writeback;
 	int from_snapshot_id;
 	char from_snapshot_tag[SD_MAX_VDI_TAG_LEN];
+	bool force;
 } vdi_cmd_data = { ~0, };
 
 struct get_vdi_info {
@@ -82,9 +85,71 @@ static int parse_option_size(const char *value, uint64_t *ret)
 	return 0;
 }
 
+#define DEFAULT_SCREEN_WIDTH 80
+
+static int get_screen_width(void)
+{
+	struct winsize wsz;
+
+	if (ioctl(STDOUT_FILENO, TIOCGWINSZ, &wsz) < 0)
+		return DEFAULT_SCREEN_WIDTH;
+
+	return wsz.ws_col;
+}
+
+/*
+ * Show prograss bar as follows.
+ *
+ *  45.0 % [===============>                  ] 180 MB / 400 MB
+ */
+static void show_progress(uint64_t done, uint64_t total)
+{
+	char done_str[256], total_str[256];
+	int screen_width = get_screen_width();
+	int bar_length = screen_width - 30;
+	char *buf;
+
+	if (!is_stdout_console())
+		return;
+	if (screen_width <= 0)
+		return;
+
+	printf("\r"); /* move to the beginning of the line */
+
+	size_to_str(done, done_str, sizeof(done_str));
+	size_to_str(total, total_str, sizeof(total_str));
+
+	buf = xmalloc(screen_width + 1);
+	snprintf(buf, screen_width, "%5.1lf %% [", (double)done / total * 100);
+
+	for (int i = 0; i < bar_length; i++) {
+		if (total * (i + 1) / bar_length <= done)
+			strcat(buf, "=");
+		else if (total * i / bar_length <= done &&
+			 done < total * (i + 1) / bar_length)
+			strcat(buf, ">");
+		else
+			strcat(buf, " ");
+	}
+	snprintf(buf + strlen(buf), screen_width - strlen(buf),
+		 "] %s / %s", done_str, total_str);
+
+	/* fill the rest of buffer with blank characters */
+	memset(buf + strlen(buf), ' ', screen_width - strlen(buf));
+	buf[screen_width] = '\0';
+	printf("%s", buf);
+
+	if (done == total)
+		printf("\n");
+
+	fflush(stdout);
+
+	free(buf);
+}
+
 static void print_vdi_list(uint32_t vid, const char *name, const char *tag,
 			   uint32_t snapid, uint32_t flags,
-			   const struct sheepdog_inode *i, void *data)
+			   const struct sd_inode *i, void *data)
 {
 	int idx;
 	bool is_clone = false;
@@ -126,7 +191,7 @@ static void print_vdi_list(uint32_t vid, const char *name, const char *tag,
 		is_clone = true;
 
 	if (raw_output) {
-		printf("%c ", is_current(i) ? (is_clone ? 'c' : '=') : 's');
+		printf("%c ", vdi_is_snapshot(i) ? 's' : (is_clone ? 'c' : '='));
 		while (*name) {
 			if (isspace(*name) || *name == '\\')
 				putchar('\\');
@@ -137,7 +202,7 @@ static void print_vdi_list(uint32_t vid, const char *name, const char *tag,
 				i->nr_copies, i->tag);
 	} else {
 		printf("%c %-8s %5d %7s %7s %7s %s  %7" PRIx32 " %5d %13s\n",
-				is_current(i) ? (is_clone ? 'c' : ' ') : 's',
+				vdi_is_snapshot(i) ? 's' : (is_clone ? 'c' : ' '),
 				name, snapid, vdi_size_str, my_objs_str, cow_objs_str,
 				dbuf, vid, i->nr_copies, i->tag);
 	}
@@ -145,28 +210,28 @@ static void print_vdi_list(uint32_t vid, const char *name, const char *tag,
 
 static void print_vdi_tree(uint32_t vid, const char *name, const char *tag,
 			   uint32_t snapid, uint32_t flags,
-			   const struct sheepdog_inode *i, void *data)
+			   const struct sd_inode *i, void *data)
 {
 	time_t ti;
 	struct tm tm;
 	char buf[128];
 
-	if (is_current(i))
-		pstrcpy(buf, sizeof(buf), "(you are here)");
-	else {
+	if (vdi_is_snapshot(i)) {
 		ti = i->create_time >> 32;
 		localtime_r(&ti, &tm);
 
 		strftime(buf, sizeof(buf),
 			 "[%Y-%m-%d %H:%M]", &tm);
-	}
+	} else
+		pstrcpy(buf, sizeof(buf), "(you are here)");
 
-	add_vdi_tree(name, buf, vid, i->parent_vdi_id, highlight && is_current(i));
+	add_vdi_tree(name, buf, vid, i->parent_vdi_id,
+		     highlight && !vdi_is_snapshot(i));
 }
 
 static void print_vdi_graph(uint32_t vid, const char *name, const char *tag,
 			    uint32_t snapid, uint32_t flags,
-			    const struct sheepdog_inode *i, void *data)
+			    const struct sd_inode *i, void *data)
 {
 	time_t ti;
 	struct tm tm;
@@ -191,16 +256,16 @@ static void print_vdi_graph(uint32_t vid, const char *name, const char *tag,
 	       "Time: %10s",
 	       name, snapid, size_str, dbuf, tbuf);
 
-	if (is_current(i))
-		printf("\",\n    color=\"red\"\n  ];\n\n");
-	else
+	if (vdi_is_snapshot(i))
 		printf("\"\n  ];\n\n");
+	else
+		printf("\",\n    color=\"red\"\n  ];\n\n");
 
 }
 
 static void get_oid(uint32_t vid, const char *name, const char *tag,
 		    uint32_t snapid, uint32_t flags,
-		    const struct sheepdog_inode *i, void *data)
+		    const struct sd_inode *i, void *data)
 {
 	struct get_vdi_info *info = data;
 
@@ -245,8 +310,8 @@ static int do_print_obj(char *sheep, uint64_t oid, struct sd_rsp *rsp,
 		fprintf(stderr, "The node list has changed: please try again\n");
 		break;
 	default:
-		fprintf(stderr, "%s: hit an unexpected error (%d)\n",
-		       sheep, rsp->result);
+		fprintf(stderr, "%s: hit an unexpected error (%s)\n",
+			sheep, sd_strerror(rsp->result));
 		break;
 	}
 
@@ -263,7 +328,7 @@ static int get_data_oid(char *sheep, uint64_t oid, struct sd_rsp *rsp,
 			 char *buf, void *data)
 {
 	struct get_data_oid_info *info = data;
-	struct sheepdog_inode *inode = (struct sheepdog_inode *)buf;
+	struct sd_inode *inode = (struct sd_inode *)buf;
 
 	switch (rsp->result) {
 	case SD_RES_SUCCESS:
@@ -282,8 +347,8 @@ static int get_data_oid(char *sheep, uint64_t oid, struct sd_rsp *rsp,
 		fprintf(stderr, "The node list has changed: please try again\n");
 		break;
 	default:
-		fprintf(stderr, "%s: hit an unexpected error (%d)\n",
-		       sheep, rsp->result);
+		fprintf(stderr, "%s: hit an unexpected error (%s)\n",
+			sheep, sd_strerror(rsp->result));
 		break;
 	}
 
@@ -293,7 +358,7 @@ static int get_data_oid(char *sheep, uint64_t oid, struct sd_rsp *rsp,
 static void parse_objs(uint64_t oid, obj_parser_func_t func, void *data, unsigned size)
 {
 	char name[128];
-	int i, fd, ret, cb_ret;
+	int i, ret, cb_ret;
 	char *buf;
 
 	buf = xzalloc(size);
@@ -301,12 +366,6 @@ static void parse_objs(uint64_t oid, obj_parser_func_t func, void *data, unsigne
 		struct sd_req hdr;
 		struct sd_rsp *rsp = (struct sd_rsp *)&hdr;
 
-		addr_to_str(name, sizeof(name), sd_nodes[i].nid.addr, 0);
-
-		fd = connect_to(name, sd_nodes[i].nid.port);
-		if (fd < 0)
-			break;
-
 		sd_init_req(&hdr, SD_OP_READ_PEER);
 		hdr.data_length = size;
 		hdr.flags = 0;
@@ -314,20 +373,19 @@ static void parse_objs(uint64_t oid, obj_parser_func_t func, void *data, unsigne
 
 		hdr.obj.oid = oid;
 
-		ret = collie_exec_req(fd, &hdr, buf);
-		close(fd);
+		addr_to_str(name, sizeof(name), sd_nodes[i].nid.addr, 0);
+		ret = collie_exec_req(name, sd_nodes[i].nid.port, &hdr, buf);
+		if (ret < 0)
+			continue;
 
-		sprintf(name + strlen(name), ":%d", sd_nodes[i].nid.port);
+		snprintf(name + strlen(name), sizeof(name) - strlen(name),
+			 ":%d", sd_nodes[i].nid.port);
 
-		if (ret)
-			fprintf(stderr, "Failed to connect to %s\n", name);
-		else {
-			untrim_zero_sectors(buf, rsp->obj.offset,
-					    rsp->data_length, size);
-			cb_ret = func(name, oid, rsp, buf, data);
-			if (cb_ret)
-				break;
-		}
+		untrim_zero_sectors(buf, rsp->obj.offset,
+				    rsp->data_length, size);
+		cb_ret = func(name, oid, rsp, buf, data);
+		if (cb_ret)
+			break;
 	}
 
 	free(buf);
@@ -384,18 +442,15 @@ static int vdi_graph(int argc, char **argv)
 static int find_vdi_name(const char *vdiname, uint32_t snapid, const char *tag,
 			 uint32_t *vid, int for_snapshot)
 {
-	int ret, fd;
+	int ret;
 	struct sd_req hdr;
 	struct sd_rsp *rsp = (struct sd_rsp *)&hdr;
 	char buf[SD_MAX_VDI_LEN + SD_MAX_VDI_TAG_LEN];
 
-	fd = connect_to(sdhost, sdport);
-	if (fd < 0)
-		return -1;
-
 	memset(buf, 0, sizeof(buf));
 	pstrcpy(buf, SD_MAX_VDI_LEN, vdiname);
-	pstrcpy(buf + SD_MAX_VDI_LEN, SD_MAX_VDI_TAG_LEN, tag);
+	if (tag)
+		pstrcpy(buf + SD_MAX_VDI_LEN, SD_MAX_VDI_TAG_LEN, tag);
 
 	if (for_snapshot)
 		sd_init_req(&hdr, SD_OP_GET_VDI_INFO);
@@ -405,28 +460,22 @@ static int find_vdi_name(const char *vdiname, uint32_t snapid, const char *tag,
 	hdr.flags = SD_FLAG_CMD_WRITE;
 	hdr.vdi.snapid = snapid;
 
-	ret = collie_exec_req(fd, &hdr, buf);
-	if (ret) {
-		ret = -1;
-		goto out;
-	}
+	ret = collie_exec_req(sdhost, sdport, &hdr, buf);
+	if (ret < 0)
+		return -1;
 
 	if (rsp->result != SD_RES_SUCCESS) {
 		fprintf(stderr, "Cannot get VDI info for %s %d %s: %s\n",
 			vdiname, snapid, tag, sd_strerror(rsp->result));
-		ret = -1;
-		goto out;
+		return -1;
 	}
 	*vid = rsp->vdi.vdi_id;
 
-	ret = 0;
-out:
-	close(fd);
-	return ret;
+	return 0;
 }
 
 static int read_vdi_obj(const char *vdiname, int snapid, const char *tag,
-			uint32_t *pvid, struct sheepdog_inode *inode,
+			uint32_t *pvid, struct sd_inode *inode,
 			size_t size)
 {
 	int ret;
@@ -458,21 +507,15 @@ static int read_vdi_obj(const char *vdiname, int snapid, const char *tag,
 	return EXIT_SUCCESS;
 }
 
-static int do_vdi_create(const char *vdiname, int64_t vdi_size,
+int do_vdi_create(const char *vdiname, int64_t vdi_size,
 			 uint32_t base_vid, uint32_t *vdi_id, bool snapshot,
 			 int nr_copies)
 {
 	struct sd_req hdr;
 	struct sd_rsp *rsp = (struct sd_rsp *)&hdr;
-	int fd, ret;
+	int ret;
 	char buf[SD_MAX_VDI_LEN];
 
-	fd = connect_to(sdhost, sdport);
-	if (fd < 0) {
-		fprintf(stderr, "Failed to connect\n");
-		return EXIT_SYSFAIL;
-	}
-
 	memset(buf, 0, sizeof(buf));
 	pstrcpy(buf, SD_MAX_VDI_LEN, vdiname);
 
@@ -482,17 +525,12 @@ static int do_vdi_create(const char *vdiname, int64_t vdi_size,
 
 	hdr.vdi.base_vdi_id = base_vid;
 	hdr.vdi.snapid = snapshot ? 1 : 0;
-	hdr.vdi.vdi_size = roundup(vdi_size, 512);
+	hdr.vdi.vdi_size = vdi_size;
 	hdr.vdi.copies = nr_copies;
 
-	ret = collie_exec_req(fd, &hdr, buf);
-
-	close(fd);
-
-	if (ret) {
-		fprintf(stderr, "Failed to send a request\n");
+	ret = collie_exec_req(sdhost, sdport, &hdr, buf);
+	if (ret < 0)
 		return EXIT_SYSFAIL;
-	}
 
 	if (rsp->result != SD_RES_SUCCESS) {
 		fprintf(stderr, "Failed to create VDI %s: %s\n", vdiname,
@@ -512,8 +550,8 @@ static int vdi_create(int argc, char **argv)
 	uint64_t size;
 	uint32_t vid;
 	uint64_t oid;
-	int idx, max_idx, ret;
-	struct sheepdog_inode *inode = NULL;
+	int idx, max_idx, ret, nr_copies = vdi_cmd_data.nr_copies;
+	struct sd_inode *inode = NULL;
 
 	if (!argv[optind]) {
 		fprintf(stderr, "Please specify the VDI size\n");
@@ -527,17 +565,18 @@ static int vdi_create(int argc, char **argv)
 		return EXIT_USAGE;
 	}
 
+	if (nr_copies > sd_nodes_nr) {
+		fprintf(stderr, "There are not enough nodes(%d) to hold "
+			"the copies(%d)\n", sd_nodes_nr, nr_copies);
+		return EXIT_USAGE;
+	}
+
 	ret = do_vdi_create(vdiname, size, 0, &vid, false,
 			    vdi_cmd_data.nr_copies);
 	if (ret != EXIT_SUCCESS || !vdi_cmd_data.prealloc)
 		goto out;
 
-	inode = malloc(sizeof(*inode));
-	if (!inode) {
-		fprintf(stderr, "Failed to allocate memory\n");
-		ret = EXIT_SYSFAIL;
-		goto out;
-	}
+	inode = xmalloc(sizeof(*inode));
 
 	ret = sd_read_object(vid_to_vdi_oid(vid), inode, sizeof(*inode), 0, true);
 	if (ret != SD_RES_SUCCESS) {
@@ -548,6 +587,7 @@ static int vdi_create(int argc, char **argv)
 	max_idx = DIV_ROUND_UP(size, SD_DATA_OBJ_SIZE);
 
 	for (idx = 0; idx < max_idx; idx++) {
+		show_progress(idx * SD_DATA_OBJ_SIZE, inode->vdi_size);
 		oid = vid_to_data_oid(vid, idx);
 
 		ret = sd_write_object(oid, 0, NULL, 0, 0, 0, inode->nr_copies,
@@ -566,6 +606,7 @@ static int vdi_create(int argc, char **argv)
 			goto out;
 		}
 	}
+	show_progress(idx * SD_DATA_OBJ_SIZE, inode->vdi_size);
 	ret = EXIT_SUCCESS;
 out:
 	free(inode);
@@ -578,7 +619,7 @@ static int vdi_snapshot(int argc, char **argv)
 	uint32_t vid;
 	int ret;
 	char buf[SD_INODE_HEADER_SIZE];
-	struct sheepdog_inode *inode = (struct sheepdog_inode *)buf;
+	struct sd_inode *inode = (struct sd_inode *)buf;
 
 	if (vdi_cmd_data.snapshot_id != 0) {
 		fprintf(stderr, "Please specify a non-integer value for "
@@ -593,7 +634,7 @@ static int vdi_snapshot(int argc, char **argv)
 	if (vdi_cmd_data.snapshot_tag[0]) {
 		ret = sd_write_object(vid_to_vdi_oid(vid), 0, vdi_cmd_data.snapshot_tag,
 				      SD_MAX_VDI_TAG_LEN,
-				      offsetof(struct sheepdog_inode, tag),
+				      offsetof(struct sd_inode, tag),
 				      0, inode->nr_copies, false, true);
 	}
 
@@ -607,7 +648,7 @@ static int vdi_clone(int argc, char **argv)
 	uint32_t base_vid, new_vid;
 	uint64_t oid;
 	int idx, max_idx, ret;
-	struct sheepdog_inode *inode = NULL;
+	struct sd_inode *inode = NULL;
 	char *buf = NULL;
 
 	dst_vdi = argv[optind];
@@ -624,12 +665,7 @@ static int vdi_clone(int argc, char **argv)
 		goto out;
 	}
 
-	inode = malloc(sizeof(*inode));
-	if (!inode) {
-		fprintf(stderr, "Failed to allocate memory\n");
-		ret = EXIT_SYSFAIL;
-		goto out;
-	}
+	inode = xmalloc(sizeof(*inode));
 
 	ret = read_vdi_obj(src_vdi, vdi_cmd_data.snapshot_id,
 			   vdi_cmd_data.snapshot_tag, &base_vid, inode,
@@ -646,6 +682,7 @@ static int vdi_clone(int argc, char **argv)
 	max_idx = DIV_ROUND_UP(inode->vdi_size, SD_DATA_OBJ_SIZE);
 
 	for (idx = 0; idx < max_idx; idx++) {
+		show_progress(idx * SD_DATA_OBJ_SIZE, inode->vdi_size);
 		if (inode->data_vdi_id[idx]) {
 			oid = vid_to_data_oid(inode->data_vdi_id[idx], idx);
 			ret = sd_read_object(oid, buf, SD_DATA_OBJ_SIZE, 0, true);
@@ -672,6 +709,7 @@ static int vdi_clone(int argc, char **argv)
 			goto out;
 		}
 	}
+	show_progress(idx * SD_DATA_OBJ_SIZE, inode->vdi_size);
 	ret = EXIT_SUCCESS;
 out:
 	free(inode);
@@ -686,7 +724,7 @@ static int vdi_resize(int argc, char **argv)
 	uint32_t vid;
 	int ret;
 	char buf[SD_INODE_HEADER_SIZE];
-	struct sheepdog_inode *inode = (struct sheepdog_inode *)buf;
+	struct sd_inode *inode = (struct sd_inode *)buf;
 
 	if (!argv[optind]) {
 		fprintf(stderr, "Please specify the new size for the VDI\n");
@@ -722,14 +760,26 @@ static int vdi_resize(int argc, char **argv)
 
 static int do_vdi_delete(const char *vdiname, int snap_id, const char *snap_tag)
 {
-	int fd, ret;
+	int ret;
 	struct sd_req hdr;
 	struct sd_rsp *rsp = (struct sd_rsp *)&hdr;
 	char data[SD_MAX_VDI_LEN + SD_MAX_VDI_TAG_LEN];
+	uint32_t vid;
 
-	fd = connect_to(sdhost, sdport);
-	if (fd < 0)
-		return EXIT_SYSFAIL;
+	ret = find_vdi_name(vdiname, snap_id, snap_tag, &vid, 0);
+	if (ret < 0) {
+		fprintf(stderr, "Failed to open VDI %s\n", vdiname);
+		return EXIT_FAILURE;
+	}
+
+	sd_init_req(&hdr, SD_OP_DELETE_CACHE);
+	hdr.obj.oid = vid_to_vdi_oid(vid);
+
+	ret = send_light_req(&hdr, sdhost, sdport);
+	if (ret) {
+		fprintf(stderr, "failed to execute request\n");
+		return EXIT_FAILURE;
+	}
 
 	sd_init_req(&hdr, SD_OP_DEL_VDI);
 	hdr.flags = SD_FLAG_CMD_WRITE;
@@ -740,13 +790,9 @@ static int do_vdi_delete(const char *vdiname, int snap_id, const char *snap_tag)
 	if (snap_tag)
 		pstrcpy(data + SD_MAX_VDI_LEN, SD_MAX_VDI_TAG_LEN, snap_tag);
 
-	ret = collie_exec_req(fd, &hdr, data);
-	close(fd);
-
-	if (ret) {
-		fprintf(stderr, "Failed to connect\n");
+	ret = collie_exec_req(sdhost, sdport, &hdr, data);
+	if (ret < 0)
 		return EXIT_SYSFAIL;
-	}
 
 	if (rsp->result != SD_RES_SUCCESS) {
 		fprintf(stderr, "Failed to delete %s: %s\n", vdiname,
@@ -774,7 +820,7 @@ static int vdi_rollback(int argc, char **argv)
 	uint32_t base_vid;
 	int ret;
 	char buf[SD_INODE_HEADER_SIZE];
-	struct sheepdog_inode *inode = (struct sheepdog_inode *)buf;
+	struct sd_inode *inode = (struct sd_inode *)buf;
 
 	if (!vdi_cmd_data.snapshot_id && !vdi_cmd_data.snapshot_tag[0]) {
 		fprintf(stderr, "Please specify the '-s' option\n");
@@ -784,8 +830,12 @@ static int vdi_rollback(int argc, char **argv)
 	ret = read_vdi_obj(vdiname, vdi_cmd_data.snapshot_id,
 			   vdi_cmd_data.snapshot_tag, &base_vid, inode,
 			   SD_INODE_HEADER_SIZE);
-	if (ret < 0)
-		return EXIT_FAILURE;
+	if (ret != EXIT_SUCCESS)
+		return ret;
+
+	if (!vdi_cmd_data.force)
+		confirm("This operation dicards any changes made since the"
+			" previous\nsnapshot was taken.  Continue? [yes/no]: ");
 
 	ret = do_vdi_delete(vdiname, 0, NULL);
 	if (ret != SD_RES_SUCCESS) {
@@ -855,10 +905,10 @@ static int vdi_object(int argc, char **argv)
 
 static int do_track_object(uint64_t oid, uint8_t nr_copies)
 {
-	int i, j, fd, ret;
+	int i, j, ret;
 	struct sd_req hdr;
 	struct sd_rsp *rsp = (struct sd_rsp *)&hdr;
-	struct sd_vnode vnodes[SD_MAX_VNODES];
+	struct sd_vnode *vnodes;
 	const struct sd_vnode *vnode_buf[SD_MAX_COPIES];
 	struct epoch_log *logs;
 	int vnodes_nr, nr_logs, log_length;
@@ -866,17 +916,13 @@ static int do_track_object(uint64_t oid, uint8_t nr_copies)
 
 	log_length = sd_epoch * sizeof(struct epoch_log);
 	logs = xmalloc(log_length);
-	fd = connect_to(sdhost, sdport);
-	if (fd < 0)
-		goto error;
+	vnodes = xmalloc(sizeof(*vnodes) * SD_MAX_VNODES);
 
 	sd_init_req(&hdr, SD_OP_STAT_CLUSTER);
 	hdr.data_length = log_length;
 
-	ret = collie_exec_req(fd, &hdr, logs);
-	close(fd);
-
-	if (ret != 0)
+	ret = collie_exec_req(sdhost, sdport, &hdr, logs);
+	if (ret < 0)
 		goto error;
 
 	if (rsp->result != SD_RES_SUCCESS) {
@@ -914,9 +960,11 @@ static int do_track_object(uint64_t oid, uint8_t nr_copies)
 	}
 
 	free(logs);
+	free(vnodes);
 	return EXIT_SUCCESS;
 error:
 	free(logs);
+	free(vnodes);
 	return EXIT_SYSFAIL;
 }
 
@@ -988,7 +1036,7 @@ static int find_vdi_attr_oid(const char *vdiname, const char *tag, uint32_t snap
 {
 	struct sd_req hdr;
 	struct sd_rsp *rsp = (struct sd_rsp *)&hdr;
-	int fd, ret;
+	int ret;
 	struct sheepdog_vdi_attr vattr;
 
 	memset(&vattr, 0, sizeof(vattr));
@@ -1001,12 +1049,6 @@ static int find_vdi_attr_oid(const char *vdiname, const char *tag, uint32_t snap
 		memcpy(vattr.value, value, value_len);
 	}
 
-	fd = connect_to(sdhost, sdport);
-	if (fd < 0) {
-		fprintf(stderr, "Failed to connect\n\n");
-		return SD_RES_EIO;
-	}
-
 	sd_init_req(&hdr, SD_OP_GET_VDI_ATTR);
 	hdr.flags = SD_FLAG_CMD_WRITE;
 	hdr.data_length = SD_ATTR_OBJ_SIZE;
@@ -1019,25 +1061,18 @@ static int find_vdi_attr_oid(const char *vdiname, const char *tag, uint32_t snap
 	if (delete)
 		hdr.flags |= SD_FLAG_CMD_DEL;
 
-	ret = collie_exec_req(fd, &hdr, &vattr);
-	if (ret) {
-		ret = SD_RES_EIO;
-		goto out;
-	}
+	ret = collie_exec_req(sdhost, sdport, &hdr, &vattr);
+	if (ret < 0)
+		return SD_RES_EIO;
 
-	if (rsp->result != SD_RES_SUCCESS) {
-		ret = rsp->result;
-		goto out;
-	}
+	if (rsp->result != SD_RES_SUCCESS)
+		return rsp->result;
 
 	*vid = rsp->vdi.vdi_id;
 	*oid = vid_to_attr_oid(rsp->vdi.vdi_id, rsp->vdi.attr_id);
 	*nr_copies = rsp->vdi.copies;
 
-	ret = SD_RES_SUCCESS;
-out:
-	close(fd);
-	return ret;
+	return SD_RES_SUCCESS;
 }
 
 static int vdi_setattr(int argc, char **argv)
@@ -1057,11 +1092,7 @@ static int vdi_setattr(int argc, char **argv)
 
 	value = argv[optind++];
 	if (!value && !vdi_cmd_data.delete) {
-		value = malloc(SD_MAX_VDI_ATTR_VALUE_LEN);
-		if (!value) {
-			fprintf(stderr, "Failed to allocate memory\n");
-			return EXIT_SYSFAIL;
-		}
+		value = xmalloc(SD_MAX_VDI_ATTR_VALUE_LEN);
 
 		offset = 0;
 reread:
@@ -1150,7 +1181,7 @@ static int vdi_read(int argc, char **argv)
 {
 	const char *vdiname = argv[optind++];
 	int ret, idx;
-	struct sheepdog_inode *inode = NULL;
+	struct sd_inode *inode = NULL;
 	uint64_t offset = 0, oid, done = 0, total = (uint64_t) -1;
 	unsigned int len, remain;
 	char *buf = NULL;
@@ -1159,10 +1190,6 @@ static int vdi_read(int argc, char **argv)
 		ret = parse_option_size(argv[optind++], &offset);
 		if (ret < 0)
 			return EXIT_USAGE;
-		if (offset % 512 != 0) {
-			fprintf(stderr, "Read offset must be block-aligned\n");
-			return EXIT_USAGE;
-		}
 		if (argv[optind]) {
 			ret = parse_option_size(argv[optind++], &total);
 			if (ret < 0)
@@ -1171,12 +1198,7 @@ static int vdi_read(int argc, char **argv)
 	}
 
 	inode = malloc(sizeof(*inode));
-	buf = malloc(SD_DATA_OBJ_SIZE);
-	if (!inode || !buf) {
-		fprintf(stderr, "Failed to allocate memory\n");
-		ret = EXIT_SYSFAIL;
-		goto out;
-	}
+	buf = xmalloc(SD_DATA_OBJ_SIZE);
 
 	ret = read_vdi_obj(vdiname, vdi_cmd_data.snapshot_id,
 			   vdi_cmd_data.snapshot_tag, NULL, inode,
@@ -1191,7 +1213,6 @@ static int vdi_read(int argc, char **argv)
 	}
 
 	total = min(total, inode->vdi_size - offset);
-	total = roundup(total, 512);
 	idx = offset / SD_DATA_OBJ_SIZE;
 	offset %= SD_DATA_OBJ_SIZE;
 	while (done < total) {
@@ -1237,7 +1258,7 @@ static int vdi_write(int argc, char **argv)
 	const char *vdiname = argv[optind++];
 	uint32_t vid, flags;
 	int ret, idx;
-	struct sheepdog_inode *inode = NULL;
+	struct sd_inode *inode = NULL;
 	uint64_t offset = 0, oid, old_oid, done = 0, total = (uint64_t) -1;
 	unsigned int len, remain;
 	char *buf = NULL;
@@ -1247,10 +1268,6 @@ static int vdi_write(int argc, char **argv)
 		ret = parse_option_size(argv[optind++], &offset);
 		if (ret < 0)
 			return EXIT_USAGE;
-		if (offset % 512 != 0) {
-			fprintf(stderr, "Write offset must be block-aligned\n");
-			return EXIT_USAGE;
-		}
 		if (argv[optind]) {
 			ret = parse_option_size(argv[optind++], &total);
 			if (ret < 0)
@@ -1258,13 +1275,8 @@ static int vdi_write(int argc, char **argv)
 		}
 	}
 
-	inode = malloc(sizeof(*inode));
-	buf = malloc(SD_DATA_OBJ_SIZE);
-	if (!inode || !buf) {
-		fprintf(stderr, "Failed to allocate memory\n");
-		ret = EXIT_SYSFAIL;
-		goto out;
-	}
+	inode = xmalloc(sizeof(*inode));
+	buf = xmalloc(SD_DATA_OBJ_SIZE);
 
 	ret = read_vdi_obj(vdiname, 0, "", &vid, inode, SD_INODE_SIZE);
 	if (ret != EXIT_SUCCESS)
@@ -1277,7 +1289,6 @@ static int vdi_write(int argc, char **argv)
 	}
 
 	total = min(total, inode->vdi_size - offset);
-	total = roundup(total, 512);
 	idx = offset / SD_DATA_OBJ_SIZE;
 	offset %= SD_DATA_OBJ_SIZE;
 	while (done < total) {
@@ -1355,43 +1366,30 @@ static void *read_object_from(const struct sd_vnode *vnode, uint64_t oid)
 {
 	struct sd_req hdr;
 	struct sd_rsp *rsp = (struct sd_rsp *)&hdr;
-	int fd, ret;
+	int ret;
 	char name[128];
 	void *buf;
+	size_t size = get_objsize(oid);
 
-	buf = malloc(SD_DATA_OBJ_SIZE);
-	if (!buf) {
-		fprintf(stderr, "Failed to allocate memory\n");
-		exit(EXIT_SYSFAIL);
-	}
-
-	addr_to_str(name, sizeof(name), vnode->nid.addr, 0);
-	fd = connect_to(name, vnode->nid.port);
-	if (fd < 0) {
-		fprintf(stderr, "FATAL: failed to connect to %s:%"PRIu32"\n",
-			name, vnode->nid.port);
-		exit(EXIT_FAILURE);
-	}
+	buf = xmalloc(size);
 
 	sd_init_req(&hdr, SD_OP_READ_PEER);
 	hdr.epoch = sd_epoch;
 	hdr.flags = 0;
-	hdr.data_length = SD_DATA_OBJ_SIZE;
+	hdr.data_length = size;
 
 	hdr.obj.oid = oid;
 
-	ret = collie_exec_req(fd, &hdr, buf);
-	close(fd);
+	addr_to_str(name, sizeof(name), vnode->nid.addr, 0);
+	ret = collie_exec_req(name, vnode->nid.port, &hdr, buf);
 
-	if (ret) {
-		fprintf(stderr, "FATAL: failed to execute request\n");
-		exit(EXIT_FAILURE);
-	}
+	if (ret < 0)
+		exit(EXIT_SYSFAIL);
 
-	switch (rsp->result)  {
+	switch (rsp->result) {
 	case SD_RES_SUCCESS:
 		untrim_zero_sectors(buf, rsp->obj.offset, rsp->data_length,
-				    SD_DATA_OBJ_SIZE);
+				    size);
 		break;
 	case SD_RES_NO_OBJ:
 		free(buf);
@@ -1409,33 +1407,23 @@ static void write_object_to(const struct sd_vnode *vnode, uint64_t oid,
 {
 	struct sd_req hdr;
 	struct sd_rsp *rsp = (struct sd_rsp *)&hdr;
-	int fd, ret;
+	int ret;
 	char name[128];
 
-	addr_to_str(name, sizeof(name), vnode->nid.addr, 0);
-	fd = connect_to(name, vnode->nid.port);
-	if (fd < 0) {
-		fprintf(stderr, "FATAL: failed to connect to %s:%"PRIu32"\n",
-			name, vnode->nid.port);
-		exit(EXIT_FAILURE);
-	}
-
 	if (create)
 		sd_init_req(&hdr, SD_OP_CREATE_AND_WRITE_PEER);
 	else
 		sd_init_req(&hdr, SD_OP_WRITE_PEER);
 	hdr.epoch = sd_epoch;
 	hdr.flags = SD_FLAG_CMD_WRITE;
-	hdr.data_length = SD_DATA_OBJ_SIZE;
+	hdr.data_length = get_objsize(oid);
 	hdr.obj.oid = oid;
 
-	ret = collie_exec_req(fd, &hdr, buf);
-	close(fd);
+	addr_to_str(name, sizeof(name), vnode->nid.addr, 0);
+	ret = collie_exec_req(name, vnode->nid.port, &hdr, buf);
 
-	if (ret) {
-		fprintf(stderr, "FATAL: failed to execute request\n");
-		exit(EXIT_FAILURE);
-	}
+	if (ret < 0)
+		exit(EXIT_SYSFAIL);
 
 	if (rsp->result != SD_RES_SUCCESS) {
 		fprintf(stderr, "FATAL: failed to write, %s\n",
@@ -1444,49 +1432,164 @@ static void write_object_to(const struct sd_vnode *vnode, uint64_t oid,
 	}
 }
 
-/*
- * Fix consistency of the replica of oid.
- *
- * XXX: The fix is rather dumb, just read the random copy and write it
- * to other replica.
- */
-static void do_check_repair(uint64_t oid, int nr_copies)
-{
-	const struct sd_vnode *tgt_vnodes[SD_MAX_COPIES];
-	void *buf = xmalloc(SD_DATA_OBJ_SIZE), *buf_cmp;
-	int i, ret;
+struct vdi_check_work {
+	struct vdi_check_info *info;
+	const struct sd_vnode *vnode;
+	uint8_t hash[SHA1_LEN];
+	bool object_found;
+	struct work work;
+};
 
-	ret = sd_read_object(oid, buf, SD_DATA_OBJ_SIZE, 0, true);
-	if (ret != SD_RES_SUCCESS) {
-		fprintf(stderr, "FATAL: read %"PRIx64" failed\n", oid);
+struct vdi_check_info {
+	uint64_t oid;
+	int nr_copies;
+	uint64_t total;
+	uint64_t *done;
+	int refcnt;
+	struct work_queue *wq;
+	struct vdi_check_work *base;
+	struct vdi_check_work vcw[0];
+};
+
+static void free_vdi_check_info(struct vdi_check_info *info)
+{
+	if (info->done) {
+		*info->done += SD_DATA_OBJ_SIZE;
+		show_progress(*info->done, info->total);
+	}
+	free(info);
+}
+
+static void vdi_repair_work(struct work *work)
+{
+	struct vdi_check_work *vcw = container_of(work, struct vdi_check_work,
+						  work);
+	struct vdi_check_info *info = vcw->info;
+	void *buf;
+
+	buf = read_object_from(info->base->vnode, info->oid);
+	write_object_to(vcw->vnode, info->oid, buf, !vcw->object_found);
+	free(buf);
+}
+
+static void vdi_repair_main(struct work *work)
+{
+	struct vdi_check_work *vcw = container_of(work, struct vdi_check_work,
+						  work);
+	struct vdi_check_info *info = vcw->info;
+
+	if (vcw->object_found)
+		fprintf(stdout, "fixed replica %"PRIx64"\n", info->oid);
+	else
+		fprintf(stdout, "fixed missing %"PRIx64"\n", info->oid);
+
+	info->refcnt--;
+	if (info->refcnt == 0)
+		free_vdi_check_info(info);
+}
+
+static void vdi_hash_check_work(struct work *work)
+{
+	struct vdi_check_work *vcw = container_of(work, struct vdi_check_work,
+						  work);
+	struct vdi_check_info *info = vcw->info;
+	char host[HOST_NAME_MAX];
+	int ret;
+	struct sd_req hdr;
+	struct sd_rsp *rsp = (struct sd_rsp *)&hdr;
+
+	sd_init_req(&hdr, SD_OP_GET_HASH);
+	hdr.obj.oid = info->oid;
+	hdr.obj.tgt_epoch = sd_epoch;
+
+	addr_to_str(host, sizeof(host), vcw->vnode->nid.addr, 0);
+	ret = collie_exec_req(host, vcw->vnode->nid.port, &hdr, NULL);
+	if (ret < 0)
+		exit(EXIT_SYSFAIL);
+
+	switch (ret) {
+	case SD_RES_SUCCESS:
+		vcw->object_found = true;
+		memcpy(vcw->hash, rsp->hash.digest, sizeof(vcw->hash));
+		uatomic_set(&info->base, vcw);
+		break;
+	case SD_RES_NO_OBJ:
+		vcw->object_found = false;
+		break;
+	default:
+		fprintf(stderr, "failed to read %"PRIx64" from %s:%d, %s\n",
+			info->oid, host, vcw->vnode->nid.port,
+			sd_strerror(ret));
+		exit(EXIT_FAILURE);
+	}
+}
+
+static void vdi_hash_check_main(struct work *work)
+{
+	struct vdi_check_work *vcw = container_of(work, struct vdi_check_work,
+						  work);
+	struct vdi_check_info *info = vcw->info;
+
+	info->refcnt--;
+	if (info->refcnt > 0)
+		return;
+
+	if (info->base  == NULL) {
+		fprintf(stderr, "no node has %"PRIx64"\n", info->oid);
 		exit(EXIT_FAILURE);
 	}
 
-	oid_to_vnodes(sd_vnodes, sd_vnodes_nr, oid, nr_copies, tgt_vnodes);
-	for (i = 0; i < nr_copies; i++) {
-		buf_cmp = read_object_from(tgt_vnodes[i], oid);
-		if (!buf_cmp) {
-			write_object_to(tgt_vnodes[i], oid, buf, true);
-			fprintf(stdout, "fixed missing %"PRIx64"\n", oid);
+	for (int i = 0; i < info->nr_copies; i++) {
+		if (&info->vcw[i] == info->base)
 			continue;
+		/* need repair when object not found or consistency broken */
+		if (!info->vcw[i].object_found ||
+		    memcmp(info->base->hash, info->vcw[i].hash,
+			   sizeof(info->base->hash)) != 0) {
+			info->vcw[i].work.fn = vdi_repair_work;
+			info->vcw[i].work.done = vdi_repair_main;
+			info->refcnt++;
+			queue_work(info->wq, &info->vcw[i].work);
 		}
-		if (memcmp(buf, buf_cmp, SD_DATA_OBJ_SIZE)) {
-			write_object_to(tgt_vnodes[i], oid, buf, false);
-			fprintf(stdout, "fixed replica %"PRIx64"\n", oid);
-		}
-		free(buf_cmp);
 	}
-	free(buf);
-	return;
+
+	if (info->refcnt == 0)
+		free_vdi_check_info(info);
+}
+
+static void queue_vdi_check_work(struct sd_inode *inode, uint64_t oid,
+				 uint64_t *done, struct work_queue *wq)
+{
+	struct vdi_check_info *info;
+	const struct sd_vnode *tgt_vnodes[SD_MAX_COPIES];
+	int nr_copies = inode->nr_copies;
+
+	info = xzalloc(sizeof(*info) + sizeof(info->vcw[0]) * nr_copies);
+	info->oid = oid;
+	info->nr_copies = nr_copies;
+	info->total = inode->vdi_size;
+	info->done = done;
+	info->wq = wq;
+
+	oid_to_vnodes(sd_vnodes, sd_vnodes_nr, oid, nr_copies, tgt_vnodes);
+	for (int i = 0; i < nr_copies; i++) {
+		info->vcw[i].info = info;
+		info->vcw[i].vnode = tgt_vnodes[i];
+		info->vcw[i].work.fn = vdi_hash_check_work;
+		info->vcw[i].work.done = vdi_hash_check_main;
+		info->refcnt++;
+		queue_work(info->wq, &info->vcw[i].work);
+	}
 }
 
 static int vdi_check(int argc, char **argv)
 {
 	const char *vdiname = argv[optind++];
-	int ret;
-	uint64_t total, done = 0, oid;
-	uint32_t idx = 0, vid;
-	struct sheepdog_inode *inode = xmalloc(sizeof(*inode));
+	int ret, max_idx;
+	uint64_t done = 0, oid;
+	uint32_t vid;
+	struct sd_inode *inode = xmalloc(sizeof(*inode));
+	struct work_queue *wq;
 
 	ret = read_vdi_obj(vdiname, vdi_cmd_data.snapshot_id,
 			   vdi_cmd_data.snapshot_tag, &vid, inode,
@@ -1495,24 +1598,31 @@ static int vdi_check(int argc, char **argv)
 		fprintf(stderr, "FATAL: no inode objects\n");
 		goto out;
 	}
-	if (sd_nodes_nr < inode->nr_copies){
-		fprintf(stderr,"ABORT: Not enough active nodes for "
+	if (sd_nodes_nr < inode->nr_copies) {
+		fprintf(stderr, "ABORT: Not enough active nodes for "
 			"consistency-check\n");
 		return EXIT_FAILURE;
 	}
 
-	do_check_repair(vid_to_vdi_oid(vid), inode->nr_copies);
-	total = inode->vdi_size;
-	while (done < total) {
+	wq = create_work_queue("vdi check", WQ_DYNAMIC);
+
+	queue_vdi_check_work(inode, vid_to_vdi_oid(vid), NULL, wq);
+
+	max_idx = DIV_ROUND_UP(inode->vdi_size, SD_DATA_OBJ_SIZE);
+	show_progress(done, inode->vdi_size);
+	for (int idx = 0; idx < max_idx; idx++) {
 		vid = inode->data_vdi_id[idx];
 		if (vid) {
 			oid = vid_to_data_oid(vid, idx);
-			do_check_repair(oid, inode->nr_copies);
+			queue_vdi_check_work(inode, oid, &done, wq);
+		} else {
+			done += SD_DATA_OBJ_SIZE;
+			show_progress(done, inode->vdi_size);
 		}
-		done += SD_DATA_OBJ_SIZE;
-		idx++;
 	}
 
+	work_queue_wait(wq);
+
 	fprintf(stdout, "finish check&repair %s\n", vdiname);
 	return EXIT_SUCCESS;
 out:
@@ -1563,9 +1673,7 @@ struct obj_backup {
 	uint8_t data[SD_DATA_OBJ_SIZE];
 };
 
-/*
- * discards redundant area from backup data
- */
+/* discards redundant area from backup data */
 static void compact_obj_backup(struct obj_backup *backup, uint8_t *from_data)
 {
 	uint8_t *p1, *p2;
@@ -1630,8 +1738,8 @@ static int vdi_backup(int argc, char **argv)
 {
 	const char *vdiname = argv[optind++];
 	int ret = EXIT_SUCCESS, idx, nr_objs;
-	struct sheepdog_inode *from_inode = xzalloc(sizeof(*from_inode));
-	struct sheepdog_inode *to_inode = xzalloc(sizeof(*to_inode));
+	struct sd_inode *from_inode = xzalloc(sizeof(*from_inode));
+	struct sd_inode *to_inode = xzalloc(sizeof(*to_inode));
 	struct backup_hdr hdr = {
 		.version = VDI_BACKUP_FORMAT_VERSION,
 		.magic = VDI_BACKUP_MAGIC,
@@ -1719,7 +1827,7 @@ out:
 
 /* restore backup data to vdi */
 static int restore_obj(struct obj_backup *backup, uint32_t vid,
-		       struct sheepdog_inode *parent_inode)
+		       struct sd_inode *parent_inode)
 {
 	int ret;
 	uint32_t parent_vid = parent_inode->data_vdi_id[backup->idx];
@@ -1746,7 +1854,7 @@ static uint32_t do_restore(const char *vdiname, int snapid, const char *tag)
 	uint32_t vid;
 	struct backup_hdr hdr;
 	struct obj_backup *backup = xzalloc(sizeof(*backup));
-	struct sheepdog_inode *inode = xzalloc(sizeof(*inode));
+	struct sd_inode *inode = xzalloc(sizeof(*inode));
 
 	ret = xread(STDIN_FILENO, &hdr, sizeof(hdr));
 	if (ret != sizeof(hdr))
@@ -1811,8 +1919,8 @@ static int vdi_restore(int argc, char **argv)
 	const char *vdiname = argv[optind++];
 	int ret;
 	char buf[SD_INODE_HEADER_SIZE] = {0};
-	struct sheepdog_inode *current_inode = xzalloc(sizeof(*current_inode));
-	struct sheepdog_inode *parent_inode = (struct sheepdog_inode *)buf;
+	struct sd_inode *current_inode = xzalloc(sizeof(*current_inode));
+	struct sd_inode *parent_inode = (struct sd_inode *)buf;
 	bool need_current_recovery = false;
 
 	if (!vdi_cmd_data.snapshot_id && !vdi_cmd_data.snapshot_tag[0]) {
@@ -1823,8 +1931,10 @@ static int vdi_restore(int argc, char **argv)
 		goto out;
 	}
 
-	/* delete the current vdi temporarily first to avoid making
-	 * the current state become snapshot */
+	/*
+	 * delete the current vdi temporarily first to avoid making
+	 * the current state become snapshot
+	 */
 	ret = read_vdi_obj(vdiname, 0, "", NULL, current_inode,
 			   SD_INODE_HEADER_SIZE);
 	if (ret != EXIT_SUCCESS)
@@ -1837,10 +1947,15 @@ static int vdi_restore(int argc, char **argv)
 		goto out;
 	}
 
+	if (is_stdin_console()) {
+		fprintf(stderr, "stdin must be pipe\n");
+		ret = EXIT_USAGE;
+		goto out;
+	}
+
 	ret = do_vdi_delete(vdiname, 0, NULL);
-	if (ret != SD_RES_SUCCESS) {
+	if (ret != EXIT_SUCCESS) {
 		fprintf(stderr, "Failed to delete the current state\n");
-		ret = EXIT_FAILURE;
 		goto out;
 	}
 	need_current_recovery = true;
@@ -1920,45 +2035,35 @@ out:
 
 static struct subcommand vdi_cache_cmd[] = {
 	{"flush", NULL, NULL, "flush the cache of the vdi specified.",
-	 NULL, 0, vdi_cache_flush},
+	 NULL, SUBCMD_FLAG_NEED_ARG, vdi_cache_flush},
 	{"delete", NULL, NULL, "delete the cache of the vdi specified in all nodes.",
-	 NULL, 0, vdi_cache_delete},
+	 NULL, SUBCMD_FLAG_NEED_ARG, vdi_cache_delete},
 	{NULL,},
 };
 
 static int vdi_cache(int argc, char **argv)
 {
-	int i;
-
-	for (i = 0; vdi_cache_cmd[i].name; i++) {
-		if (!strcmp(vdi_cache_cmd[i].name, argv[3])) {
-			optind++;
-			return vdi_cache_cmd[i].fn(argc, argv);
-		}
-	}
-
-	subcommand_usage(argv[1], argv[2], EXIT_FAILURE);
-	return EXIT_FAILURE;
+	return do_generic_subcommand(vdi_cache_cmd, argc, argv);
 }
 
 static struct subcommand vdi_cmd[] = {
 	{"check", "<vdiname>", "saph", "check and repair image's consistency",
-	 NULL, SUBCMD_FLAG_NEED_NODELIST|SUBCMD_FLAG_NEED_THIRD_ARG,
+	 NULL, SUBCMD_FLAG_NEED_NODELIST|SUBCMD_FLAG_NEED_ARG,
 	 vdi_check, vdi_options},
 	{"create", "<vdiname> <size>", "Pcaph", "create an image",
-	 NULL, SUBCMD_FLAG_NEED_THIRD_ARG,
+	 NULL, SUBCMD_FLAG_NEED_NODELIST|SUBCMD_FLAG_NEED_ARG,
 	 vdi_create, vdi_options},
 	{"snapshot", "<vdiname>", "saph", "create a snapshot",
-	 NULL, SUBCMD_FLAG_NEED_THIRD_ARG,
+	 NULL, SUBCMD_FLAG_NEED_ARG,
 	 vdi_snapshot, vdi_options},
 	{"clone", "<src vdi> <dst vdi>", "sPcaph", "clone an image",
-	 NULL, SUBCMD_FLAG_NEED_THIRD_ARG,
+	 NULL, SUBCMD_FLAG_NEED_ARG,
 	 vdi_clone, vdi_options},
 	{"delete", "<vdiname>", "saph", "delete an image",
-	 NULL, SUBCMD_FLAG_NEED_THIRD_ARG,
+	 NULL, SUBCMD_FLAG_NEED_ARG,
 	 vdi_delete, vdi_options},
-	{"rollback", "<vdiname>", "saph", "rollback to a snapshot",
-	 NULL, SUBCMD_FLAG_NEED_THIRD_ARG,
+	{"rollback", "<vdiname>", "saphf", "rollback to a snapshot",
+	 NULL, SUBCMD_FLAG_NEED_ARG,
 	 vdi_rollback, vdi_options},
 	{"list", "[vdiname]", "aprh", "list images",
 	 NULL, 0, vdi_list, vdi_options},
@@ -1967,37 +2072,37 @@ static struct subcommand vdi_cmd[] = {
 	{"graph", NULL, "aph", "show images in Graphviz dot format",
 	 NULL, 0, vdi_graph, vdi_options},
 	{"object", "<vdiname>", "isaph", "show object information in the image",
-	 NULL, SUBCMD_FLAG_NEED_NODELIST|SUBCMD_FLAG_NEED_THIRD_ARG,
+	 NULL, SUBCMD_FLAG_NEED_NODELIST|SUBCMD_FLAG_NEED_ARG,
 	 vdi_object, vdi_options},
 	{"track", "<vdiname>", "isaph", "show the object epoch trace in the image",
-	 NULL, SUBCMD_FLAG_NEED_NODELIST|SUBCMD_FLAG_NEED_THIRD_ARG,
+	 NULL, SUBCMD_FLAG_NEED_NODELIST|SUBCMD_FLAG_NEED_ARG,
 	 vdi_track, vdi_options},
 	{"setattr", "<vdiname> <key> [value]", "dxaph", "set a VDI attribute",
-	 NULL, SUBCMD_FLAG_NEED_THIRD_ARG,
+	 NULL, SUBCMD_FLAG_NEED_ARG,
 	 vdi_setattr, vdi_options},
 	{"getattr", "<vdiname> <key>", "aph", "get a VDI attribute",
-	 NULL, SUBCMD_FLAG_NEED_THIRD_ARG,
+	 NULL, SUBCMD_FLAG_NEED_ARG,
 	 vdi_getattr, vdi_options},
 	{"resize", "<vdiname> <new size>", "aph", "resize an image",
-	 NULL, SUBCMD_FLAG_NEED_THIRD_ARG,
+	 NULL, SUBCMD_FLAG_NEED_ARG,
 	 vdi_resize, vdi_options},
 	{"read", "<vdiname> [<offset> [<len>]]", "saph", "read data from an image",
-	 NULL, SUBCMD_FLAG_NEED_THIRD_ARG,
+	 NULL, SUBCMD_FLAG_NEED_ARG,
 	 vdi_read, vdi_options},
 	{"write", "<vdiname> [<offset> [<len>]]", "apwh", "write data to an image",
-	 NULL, SUBCMD_FLAG_NEED_THIRD_ARG,
+	 NULL, SUBCMD_FLAG_NEED_ARG,
 	 vdi_write, vdi_options},
 	{"flush", "<vdiname>", "aph", "flush data to cluster",
-	 NULL, SUBCMD_FLAG_NEED_THIRD_ARG,
+	 NULL, SUBCMD_FLAG_NEED_ARG,
 	 vdi_flush, vdi_options},
 	{"backup", "<vdiname> <backup>", "sFaph", "create an incremental backup between two snapshots",
-	 NULL, SUBCMD_FLAG_NEED_NODELIST|SUBCMD_FLAG_NEED_THIRD_ARG,
+	 NULL, SUBCMD_FLAG_NEED_NODELIST|SUBCMD_FLAG_NEED_ARG,
 	 vdi_backup, vdi_options},
 	{"restore", "<vdiname> <backup>", "saph", "restore snapshot images from a backup",
-	 NULL, SUBCMD_FLAG_NEED_NODELIST|SUBCMD_FLAG_NEED_THIRD_ARG,
+	 NULL, SUBCMD_FLAG_NEED_NODELIST|SUBCMD_FLAG_NEED_ARG,
 	 vdi_restore, vdi_options},
-	{"cache", NULL, "saph", "Run 'collie vdi cache' for more information\n",
-	 vdi_cache_cmd, SUBCMD_FLAG_NEED_THIRD_ARG,
+	{"cache", "<vdiname>", "saph", "Run 'collie vdi cache' for more information\n",
+	 vdi_cache_cmd, SUBCMD_FLAG_NEED_ARG,
 	 vdi_cache, vdi_options},
 	{NULL,},
 };
@@ -2024,6 +2129,10 @@ static int vdi_parser(int ch, char *opt)
 			vdi_cmd_data.snapshot_id = 0;
 			pstrcpy(vdi_cmd_data.snapshot_tag,
 				sizeof(vdi_cmd_data.snapshot_tag), opt);
+		} else if (vdi_cmd_data.snapshot_id == 0) {
+			fprintf(stderr,
+				"The snapshot id must be larger than zero\n");
+			exit(EXIT_FAILURE);
 		}
 		break;
 	case 'x':
@@ -2052,6 +2161,9 @@ static int vdi_parser(int ch, char *opt)
 				sizeof(vdi_cmd_data.from_snapshot_tag), opt);
 		}
 		break;
+	case 'f':
+		vdi_cmd_data.force = true;
+		break;
 	}
 
 	return 0;
