@@ -6,8 +6,8 @@
 #include <stddef.h>
 
 #define container_of(ptr, type, member) ({			\
-	const typeof( ((type *)0)->member ) *__mptr = (ptr);	\
-	(type *)( (char *)__mptr - offsetof(type,member) );})
+	const typeof(((type *)0)->member) *__mptr = (ptr);	\
+	(type *)((char *)__mptr - offsetof(type, member)); })
 
 struct list_head {
 	struct list_head *next, *prev;
@@ -40,13 +40,13 @@ static inline int list_empty(const struct list_head *head)
 
 #define list_for_each_entry(pos, head, member)				\
 	for (pos = list_entry((head)->next, typeof(*pos), member);	\
-	     &pos->member != (head);				 	\
+	     &pos->member != (head);					\
 	     pos = list_entry(pos->member.next, typeof(*pos), member))
 
 #define list_for_each_entry_safe(pos, n, head, member)			\
 	for (pos = list_entry((head)->next, typeof(*pos), member),	\
 		n = list_entry(pos->member.next, typeof(*pos), member);	\
-	     &pos->member != (head); 					\
+	     &pos->member != (head);					\
 	     pos = n, n = list_entry(n->member.next, typeof(*n), member))
 
 static inline void __list_add(struct list_head *new,
@@ -217,50 +217,39 @@ static inline void hlist_add_after(struct hlist_node *n,
 		next->next->pprev  = &next->next;
 }
 
-#define hlist_entry(ptr, type, member) container_of(ptr,type,member)
+#define hlist_entry(ptr, type, member) container_of(ptr, type, member)
 
 #define hlist_for_each(pos, head) \
-        for (pos = (head)->first; pos ; pos = pos->next)
+	for (pos = (head)->first; pos ; pos = pos->next)
 
 #define hlist_for_each_safe(pos, n, head) \
-        for (pos = (head)->first; pos && ({ n = pos->next; 1; }); \
-             pos = n)
+	for (pos = (head)->first; pos && ({ n = pos->next; 1; }); pos = n)
 
-/**
+/*
  * hlist_for_each_entry - iterate over list of given type
  * @tpos:       the type * to use as a loop cursor.
  * @pos:        the &struct hlist_node to use as a loop cursor.
  * @head:       the head for your list.
  * @member:     the name of the hlist_node within the struct.
  */
-#define hlist_for_each_entry(tpos, pos, head, member)                    \
-        for (pos = (head)->first;                                        \
-             pos &&                                                      \
-                ({ tpos = hlist_entry(pos, typeof(*tpos), member); 1;}); \
-             pos = pos->next)
+#define hlist_for_each_entry(tpos, pos, head, member)			\
+	for (pos = (head)->first;					\
+	     pos && ({ tpos = hlist_entry(pos, typeof(*tpos), member); 1; }); \
+	     pos = pos->next)
 
-/**
- * hlist_for_each_entry_safe - iterate over list of given type safe against removal of list entry
+/*
+ * hlist_for_each_entry_safe - iterate over list of given type safe against
+ * removal of list entry
  * @tpos:       the type * to use as a loop cursor.
  * @pos:        the &struct hlist_node to use as a loop cursor.
  * @n:          another &struct hlist_node to use as temporary storage
  * @head:       the head for your list.
  * @member:     the name of the hlist_node within the struct.
  */
-#define hlist_for_each_entry_safe(tpos, pos, n, head, member)            \
-        for (pos = (head)->first;                                        \
-             pos && ({ n = pos->next; 1; }) &&                           \
-                ({ tpos = hlist_entry(pos, typeof(*tpos), member); 1;}); \
-             pos = n)
+#define hlist_for_each_entry_safe(tpos, pos, n, head, member)		\
+	for (pos = (head)->first;					\
+	     pos && ({ n = pos->next; 1; }) &&				\
+		     ({ tpos = hlist_entry(pos, typeof(*tpos), member); 1; }); \
+	     pos = n)
 
-#endif
-
-
-#define list_for_each_entry_revert_safe_rcu(pos, n, head, member)          \
-	for (pos = cds_list_entry(rcu_dereference((head)->prev),           \
-				  typeof(*pos), member),                   \
-	     n = cds_list_entry(rcu_dereference(pos->member.prev),         \
-				typeof(*pos), member);                     \
-				&pos->member != (head);                    \
-	     pos = n, n = cds_list_entry(rcu_dereference(pos->member.prev),\
-					 typeof(*pos), member))
+#endif	/* __LIST_H__ */
