@@ -17,21 +17,35 @@
 #include <stdbool.h>
 #include <sys/syslog.h>
 
-#define LOG_SPACE_SIZE (32 * 1024 * 1024)
-#define MAX_MSG_SIZE 256
+#include "util.h"
 
+#define LOG_SPACE_SIZE (32 * 1024 * 1024)
+#define MAX_MSG_SIZE 1024
+#define MAX_THREAD_NAME_LEN	20
+
+struct logger_user_info {
+	int port;
+};
+
+void early_log_init(const char *format_name,
+		struct logger_user_info *user_info);
 int log_init(const char *progname, int size, bool to_stdout, int level,
-		    char *outfile);
+	char *outfile);
 void log_close(void);
 void dump_logmsg(void *);
 void log_write(int prio, const char *func, int line, const char *fmt, ...)
-	__attribute__ ((format (printf, 4, 5)));
-void set_thread_name(const char *name, int idx);
+	__printf(4, 5);
+void set_thread_name(const char *name, bool show_idx);
 void get_thread_name(char *name);
 
-/*
-+ * sheep log priorities, comliant with syslog spec
-+ */
+#define sd_dump_variable(var) ({		\
+	register void *current_sp asm("rsp");	\
+	__sd_dump_variable(#var, current_sp);	\
+})
+int __sd_dump_variable(const char *var, const void *base_sp);
+void sd_backtrace(void);
+
+/* sheep log priorities, comliant with syslog spec */
 #define	SDOG_EMERG		LOG_EMERG
 #define	SDOG_ALERT		LOG_ALERT
 #define	SDOG_CRIT		LOG_CRIT
