@@ -13,6 +13,8 @@
 
 #include <inttypes.h>
 #include <stdint.h>
+#include <linux/limits.h>
+
 #include "util.h"
 
 #define SD_PROTO_VER 0x02
@@ -26,6 +28,7 @@
 #define SD_OP_READ_OBJ       0x02
 #define SD_OP_WRITE_OBJ      0x03
 #define SD_OP_REMOVE_OBJ     0x04
+#define SD_OP_DISCARD_OBJ    0x05
 
 #define SD_OP_NEW_VDI        0x11
 #define SD_OP_LOCK_VDI       0x12
@@ -33,6 +36,7 @@
 #define SD_OP_GET_VDI_INFO   0x14
 #define SD_OP_READ_VDIS      0x15
 #define SD_OP_FLUSH_VDI      0x16
+#define SD_OP_DEL_VDI        0x17
 
 #define SD_FLAG_CMD_WRITE    0x01
 #define SD_FLAG_CMD_COW      0x02
@@ -63,17 +67,10 @@
 #define SD_RES_VER_MISMATCH  0x14 /* Protocol version mismatch */
 #define SD_RES_NO_SPACE      0x15 /* Server has no room for new objects */
 #define SD_RES_WAIT_FOR_FORMAT  0x16 /* Sheepdog is waiting for a format operation */
-#define SD_RES_WAIT_FOR_JOIN    0x17 /* Sheepdog is waiting for other nodes joining */
+#define SD_RES_WAIT_FOR_JOIN 0x17 /* Sheepdog is waiting for other nodes joining */
 #define SD_RES_JOIN_FAILED   0x18 /* Target node had failed to join sheepdog */
-#define SD_RES_HALT 0x19 /* Sheepdog is stopped doing IO */
-#define SD_RES_FORCE_RECOVER    0x1A /* Users should not force recover this cluster */
-#define SD_RES_NO_STORE         0x20 /* No targeted backend store */
-#define SD_RES_NO_SUPPORT       0x21 /* Operation is not supported by backend store */
-#define SD_RES_NODE_IN_RECOVERY 0x22 /*	Targeted node is in recovery */
-#define SD_RES_OBJ_RECOVERING     0x23 /* Object is recovering */
-#define SD_RES_KILLED           0x24 /* Node is killed */
-#define SD_RES_OID_EXIST        0x25 /* Object ID exists already */
-#define SD_RES_AGAIN            0x26 /* Ask to try again */
+#define SD_RES_HALT          0x19 /* Sheepdog is stopped doing IO */
+#define SD_RES_READONLY      0x1A /* Object is read-only */
 
 /* errors above 0x80 are sheepdog-internal */
 
@@ -88,6 +85,7 @@
  */
 
 #define VDI_SPACE_SHIFT   32
+#define SD_VDI_MASK 0x00FFFFFF00000000
 #define VDI_BIT (UINT64_C(1) << 63)
 #define VMSTATE_BIT (UINT64_C(1) << 62)
 #define VDI_ATTR_BIT (UINT64_C(1) << 61)
@@ -97,18 +95,22 @@
 #define SD_MAX_VDI_TAG_LEN 256U
 #define SD_MAX_VDI_ATTR_KEY_LEN 256U
 #define SD_MAX_VDI_ATTR_VALUE_LEN 65536U
+#define SD_MAX_SNAPSHOT_TAG_LEN 256U
 #define SD_NR_VDIS   (1U << 24)
 #define SD_DATA_OBJ_SIZE (UINT64_C(1) << 22)
 #define SD_MAX_VDI_SIZE (SD_DATA_OBJ_SIZE * MAX_DATA_OBJS)
 
-#define SD_INODE_SIZE (sizeof(struct sheepdog_inode))
-#define SD_INODE_HEADER_SIZE (sizeof(struct sheepdog_inode) - \
+#define SD_INODE_SIZE (sizeof(struct sd_inode))
+#define SD_INODE_HEADER_SIZE (sizeof(struct sd_inode) - \
 			      sizeof(uint32_t) * MAX_DATA_OBJS)
 #define SD_ATTR_OBJ_SIZE (sizeof(struct sheepdog_vdi_attr))
 #define CURRENT_VDI_ID 0
 
 #define STORE_LEN 16
 
+#define SD_REQ_SIZE 48
+#define SD_RSP_SIZE 48
+
 struct sd_req {
 	uint8_t		proto_ver;
 	uint8_t		opcode;
@@ -130,6 +132,22 @@ struct sd_req {
 			uint32_t	copies;
 			uint32_t	snapid;
 		} vdi;
+
+		/* sheepdog-internal */
+		struct {
+			uint64_t	oid;
+			uint64_t	ctime;
+			uint32_t	copies;
+			uint32_t	tag;
+		} cluster;
+		struct {
+			uint32_t	old_vid;
+			uint32_t	new_vid;
+			uint32_t	copies;
+			uint8_t		set_bitmap; /* 0 means false */
+						    /* others mean true */
+		} vdi_state;
+
 		uint32_t		__pad[8];
 	};
 };
@@ -155,11 +173,27 @@ struct sd_rsp {
 			uint32_t	attr_id;
 			uint32_t	copies;
 		} vdi;
+
+		/* sheepdog-internal */
+		struct {
+			uint32_t	__pad;
+			uint32_t	nr_nodes;
+			uint32_t	local_idx;
+			uint32_t	master_idx;
+			uint64_t	store_size;
+			uint64_t	store_free;
+		} node;
+		struct {
+			uint32_t	__pad1;
+			uint32_t	__pad2;
+			uint8_t		digest[20];
+		} hash;
+
 		uint32_t		__pad[8];
 	};
 };
 
-struct sheepdog_inode {
+struct sd_inode {
 	char name[SD_MAX_VDI_LEN];
 	char tag[SD_MAX_VDI_TAG_LEN];
 	uint64_t create_time;
@@ -190,19 +224,16 @@ struct sheepdog_vdi_attr {
 #define SHA1_LEN        20
 
 struct snap_log {
-	uint32_t epoch;
+	uint32_t idx;
+	char tag[SD_MAX_SNAPSHOT_TAG_LEN];
 	uint64_t time;
 	unsigned char sha1[SHA1_LEN];
 };
 
-/*
- * 64 bit FNV-1a non-zero initial basis
- */
+/* 64 bit FNV-1a non-zero initial basis */
 #define FNV1A_64_INIT ((uint64_t) 0xcbf29ce484222325ULL)
 
-/*
- * 64 bit Fowler/Noll/Vo FNV-1a hash code
- */
+/* 64 bit Fowler/Noll/Vo FNV-1a hash code */
 static inline uint64_t fnv_64a_buf(const void *buf, size_t len, uint64_t hval)
 {
 	unsigned char *bp = (unsigned char *) buf;
@@ -222,7 +253,7 @@ static inline uint64_t hash_64(uint64_t val, unsigned int bits)
 	return hash & ((1 << bits) - 1);
 }
 
-static inline bool is_data_obj_writeable(const struct sheepdog_inode *inode,
+static inline bool is_data_obj_writeable(const struct sd_inode *inode,
 					 int idx)
 {
 	return inode->vdi_id == inode->data_vdi_id[idx];
@@ -277,7 +308,7 @@ static inline uint64_t vid_to_data_oid(uint32_t vid, uint32_t idx)
 
 static inline uint32_t oid_to_vid(uint64_t oid)
 {
-	return (~VDI_BIT & oid) >> VDI_SPACE_SHIFT;
+	return (oid & SD_VDI_MASK) >> VDI_SPACE_SHIFT;
 }
 
 static inline uint64_t vid_to_attr_oid(uint32_t vid, uint32_t attrid)
@@ -285,9 +316,16 @@ static inline uint64_t vid_to_attr_oid(uint32_t vid, uint32_t attrid)
 	return ((uint64_t)vid << VDI_SPACE_SHIFT) | VDI_ATTR_BIT | attrid;
 }
 
-static inline uint32_t attr_oid_to_vid(uint64_t oid)
+static inline bool vdi_is_snapshot(const struct sd_inode *inode)
 {
-	return (~VDI_ATTR_BIT & oid) >> VDI_SPACE_SHIFT;
+	return !!inode->snap_ctime;
+}
+
+static inline __attribute__((used)) void __sd_proto_build_bug_ons(void)
+{
+	/* never called, only for checking BUILD_BUG_ON()s */
+	BUILD_BUG_ON(sizeof(struct sd_req) != SD_REQ_SIZE);
+	BUILD_BUG_ON(sizeof(struct sd_rsp) != SD_RSP_SIZE);
 }
 
 #endif
