@@ -24,8 +24,6 @@
 static int efd;
 static LIST_HEAD(events_list);
 
-#define TICK 1
-
 static void timer_handler(int fd, int events, void *data)
 {
 	struct timer *t = data;
@@ -47,7 +45,7 @@ void add_timer(struct timer *t, unsigned int mseconds)
 
 	tfd = timerfd_create(CLOCK_MONOTONIC, TFD_NONBLOCK);
 	if (tfd < 0) {
-		sd_eprintf("timerfd_create: %m\n");
+		sd_eprintf("timerfd_create: %m");
 		return;
 	}
 
@@ -56,12 +54,12 @@ void add_timer(struct timer *t, unsigned int mseconds)
 	it.it_value.tv_nsec = (mseconds % 1000) * 1000000;
 
 	if (timerfd_settime(tfd, 0, &it, NULL) < 0) {
-		sd_eprintf("timerfd_settime: %m\n");
+		sd_eprintf("timerfd_settime: %m");
 		return;
 	}
 
 	if (register_event(tfd, timer_handler, t) < 0)
-		sd_eprintf("failed to register timer fd\n");
+		sd_eprintf("failed to register timer fd");
 }
 
 struct event_info {
@@ -69,13 +67,20 @@ struct event_info {
 	int fd;
 	void *data;
 	struct list_head ei_list;
+	int prio;
 };
 
+static struct epoll_event *events;
+static int nr_events;
+
 int init_event(int nr)
 {
+	nr_events = nr;
+	events = xcalloc(nr_events, sizeof(struct epoll_event));
+
 	efd = epoll_create(nr);
 	if (efd < 0) {
-		sd_eprintf("failed to create epoll fd\n");
+		sd_eprintf("failed to create epoll fd");
 		return -1;
 	}
 	return 0;
@@ -92,7 +97,7 @@ static struct event_info *lookup_event(int fd)
 	return NULL;
 }
 
-int register_event(int fd, event_handler_t h, void *data)
+int register_event_prio(int fd, event_handler_t h, void *data, int prio)
 {
 	int ret;
 	struct epoll_event ev;
@@ -102,6 +107,7 @@ int register_event(int fd, event_handler_t h, void *data)
 	ei->fd = fd;
 	ei->handler = h;
 	ei->data = data;
+	ei->prio = prio;
 
 	memset(&ev, 0, sizeof(ev));
 	ev.events = EPOLLIN;
@@ -109,7 +115,7 @@ int register_event(int fd, event_handler_t h, void *data)
 
 	ret = epoll_ctl(efd, EPOLL_CTL_ADD, fd, &ev);
 	if (ret) {
-		sd_eprintf("failed to add epoll event: %m\n");
+		sd_eprintf("failed to add epoll event: %m");
 		free(ei);
 	} else
 		list_add(&ei->ei_list, &events_list);
@@ -128,13 +134,13 @@ void unregister_event(int fd)
 
 	ret = epoll_ctl(efd, EPOLL_CTL_DEL, fd, NULL);
 	if (ret)
-		sd_eprintf("failed to delete epoll event for fd %d: %m\n", fd);
+		sd_eprintf("failed to delete epoll event for fd %d: %m", fd);
 
 	list_del(&ei->ei_list);
 	free(ei);
 }
 
-int modify_event(int fd, unsigned int events)
+int modify_event(int fd, unsigned int new_events)
 {
 	int ret;
 	struct epoll_event ev;
@@ -142,32 +148,58 @@ int modify_event(int fd, unsigned int events)
 
 	ei = lookup_event(fd);
 	if (!ei) {
-		sd_eprintf("event info for fd %d not found\n", fd);
+		sd_eprintf("event info for fd %d not found", fd);
 		return 1;
 	}
 
 	memset(&ev, 0, sizeof(ev));
-	ev.events = events;
+	ev.events = new_events;
 	ev.data.ptr = ei;
 
 	ret = epoll_ctl(efd, EPOLL_CTL_MOD, fd, &ev);
 	if (ret) {
-		sd_eprintf("failed to delete epoll event for fd %d: %m\n", fd);
+		sd_eprintf("failed to delete epoll event for fd %d: %m", fd);
 		return 1;
 	}
 	return 0;
 }
 
-void event_loop(int timeout)
+static bool event_loop_refresh;
+
+void event_force_refresh(void)
+{
+	event_loop_refresh = true;
+}
+
+static int epoll_event_cmp(const struct epoll_event *_a, struct epoll_event *_b)
+{
+	struct event_info *a, *b;
+
+	a = (struct event_info *)_a->data.ptr;
+	b = (struct event_info *)_b->data.ptr;
+
+	/* we need sort event_info array in reverse order */
+	if (a->prio < b->prio)
+		return 1;
+	else if (b->prio < a->prio)
+		return -1;
+
+	return 0;
+}
+
+static void do_event_loop(int timeout, bool sort_with_prio)
 {
 	int i, nr;
-	struct epoll_event events[128];
 
-	nr = epoll_wait(efd, events, ARRAY_SIZE(events), TICK * 1000);
+refresh:
+	nr = epoll_wait(efd, events, nr_events, timeout);
+	if (sort_with_prio)
+		xqsort(events, nr, epoll_event_cmp);
+
 	if (nr < 0) {
 		if (errno == EINTR)
 			return;
-		sd_eprintf("epoll_wait failed: %m\n");
+		sd_eprintf("epoll_wait failed: %m");
 		exit(1);
 	} else if (nr) {
 		for (i = 0; i < nr; i++) {
@@ -175,6 +207,21 @@ void event_loop(int timeout)
 
 			ei = (struct event_info *)events[i].data.ptr;
 			ei->handler(ei->fd, events[i].events, ei->data);
+
+			if (event_loop_refresh) {
+				event_loop_refresh = false;
+				goto refresh;
+			}
 		}
 	}
 }
+
+void event_loop(int timeout)
+{
+	do_event_loop(timeout, false);
+}
+
+void event_loop_prio(int timeout)
+{
+	do_event_loop(timeout, true);
+}
