@@ -25,6 +25,8 @@
 #include <sys/un.h>
 #include <sys/stat.h>
 #include <sys/types.h>
+#include <sys/uio.h>
+
 #include "sheepdog_proto.h"
 #include "sheep.h"
 #include "util.h"
@@ -130,7 +132,7 @@ int create_listen_ports(const char *bindaddr, int port,
 
 	ret = getaddrinfo(bindaddr, servname, &hints, &res0);
 	if (ret) {
-		sd_eprintf("failed to get address info: %m\n");
+		sd_eprintf("failed to get address info: %m");
 		return 1;
 	}
 
@@ -143,7 +145,7 @@ int create_listen_ports(const char *bindaddr, int port,
 		ret = setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &opt,
 				 sizeof(opt));
 		if (ret)
-			sd_eprintf("failed to set SO_REUSEADDR: %m\n");
+			sd_eprintf("failed to set SO_REUSEADDR: %m");
 
 		opt = 1;
 		if (res->ai_family == AF_INET6) {
@@ -157,14 +159,14 @@ int create_listen_ports(const char *bindaddr, int port,
 
 		ret = bind(fd, res->ai_addr, res->ai_addrlen);
 		if (ret) {
-			sd_eprintf("failed to bind server socket: %m\n");
+			sd_eprintf("failed to bind server socket: %m");
 			close(fd);
 			continue;
 		}
 
 		ret = listen(fd, SOMAXCONN);
 		if (ret) {
-			sd_eprintf("failed to listen on server socket: %m\n");
+			sd_eprintf("failed to listen on server socket: %m");
 			close(fd);
 			continue;
 		}
@@ -187,7 +189,7 @@ int create_listen_ports(const char *bindaddr, int port,
 	freeaddrinfo(res0);
 
 	if (!success)
-		sd_eprintf("failed to create a listening port\n");
+		sd_eprintf("failed to create a listening port");
 
 	return !success;
 }
@@ -207,7 +209,7 @@ int connect_to(const char *name, int port)
 
 	ret = getaddrinfo(name, buf, &hints, &res0);
 	if (ret) {
-		sd_eprintf("failed to get address info: %m\n");
+		sd_eprintf("failed to get address info: %m");
 		return -1;
 	}
 
@@ -225,36 +227,38 @@ int connect_to(const char *name, int port)
 		ret = setsockopt(fd, SOL_SOCKET, SO_LINGER, &linger_opt,
 				 sizeof(linger_opt));
 		if (ret) {
-			sd_eprintf("failed to set SO_LINGER: %m\n");
+			sd_eprintf("failed to set SO_LINGER: %m");
 			close(fd);
 			continue;
 		}
 
 		ret = set_snd_timeout(fd);
 		if (ret) {
-			sd_eprintf("failed to set send timeout: %m\n");
+			sd_eprintf("failed to set send timeout: %m");
 			close(fd);
 			break;
 		}
 
 		ret = set_rcv_timeout(fd);
 		if (ret) {
-			sd_eprintf("failed to set recv timeout: %m\n");
+			sd_eprintf("failed to set recv timeout: %m");
 			close(fd);
 			break;
 		}
-
+reconnect:
 		ret = connect(fd, res->ai_addr, res->ai_addrlen);
 		if (ret) {
-			sd_eprintf("failed to connect to %s:%d: %m\n",
-				name, port);
+			if (errno == EINTR)
+				goto reconnect;
+			sd_eprintf("failed to connect to %s:%d: %m", name,
+				   port);
 			close(fd);
 			continue;
 		}
 
 		ret = set_nodelay(fd);
 		if (ret) {
-			sd_eprintf("%m\n");
+			sd_eprintf("%m");
 			close(fd);
 			break;
 		} else
@@ -263,7 +267,7 @@ int connect_to(const char *name, int port)
 	fd = -1;
 success:
 	freeaddrinfo(res0);
-	sd_dprintf("%d, %s:%d\n", fd, name, port);
+	sd_dprintf("%d, %s:%d", fd, name, port);
 	return fd;
 }
 
@@ -273,7 +277,11 @@ int do_read(int sockfd, void *buf, int len, bool (*need_retry)(uint32_t epoch),
 	int ret, repeat = MAX_RETRY_COUNT;
 reread:
 	ret = read(sockfd, buf, len);
-	if (ret < 0 || !ret) {
+	if (ret == 0) {
+		sd_eprintf("connection is closed (%d bytes left)", len);
+		return 1;
+	}
+	if (ret < 0) {
 		if (errno == EINTR)
 			goto reread;
 		/*
@@ -286,8 +294,7 @@ reread:
 			goto reread;
 		}
 
-		sd_eprintf("failed to read from socket: %d, %d(%m)\n",
-			ret, errno);
+		sd_eprintf("failed to read from socket: %d, %m", ret);
 		return 1;
 	}
 
@@ -331,7 +338,7 @@ rewrite:
 			goto rewrite;
 		}
 
-		sd_eprintf("failed to write to socket: %m\n");
+		sd_eprintf("failed to write to socket: %m");
 		return 1;
 	}
 
@@ -367,8 +374,8 @@ int send_req(int sockfd, struct sd_req *hdr, void *data, unsigned int wlen,
 
 	ret = do_write(sockfd, &msg, sizeof(*hdr) + wlen, need_retry, epoch);
 	if (ret) {
-		sd_eprintf("failed to send request %x, %d: %m\n", hdr->opcode,
-			wlen);
+		sd_eprintf("failed to send request %x, %d: %m", hdr->opcode,
+			   wlen);
 		ret = -1;
 	}
 
@@ -395,7 +402,7 @@ int exec_req(int sockfd, struct sd_req *hdr, void *data,
 
 	ret = do_read(sockfd, rsp, sizeof(*rsp), need_retry, epoch);
 	if (ret) {
-		sd_eprintf("failed to read a response\n");
+		sd_eprintf("failed to read a response");
 		return 1;
 	}
 
@@ -405,7 +412,7 @@ int exec_req(int sockfd, struct sd_req *hdr, void *data,
 	if (rlen) {
 		ret = do_read(sockfd, data, rlen, need_retry, epoch);
 		if (ret) {
-			sd_eprintf("failed to read the response data\n");
+			sd_eprintf("failed to read the response data");
 			return 1;
 		}
 	}
@@ -415,8 +422,9 @@ int exec_req(int sockfd, struct sd_req *hdr, void *data,
 
 char *addr_to_str(char *str, int size, const uint8_t *addr, uint16_t port)
 {
-	int  af = AF_INET6;
-	int  addr_start_idx = 0;
+	int af = AF_INET6;
+	int addr_start_idx = 0;
+	const char *ret;
 
 	/* Find address family type */
 	if (addr[12]) {
@@ -428,7 +436,10 @@ char *addr_to_str(char *str, int size, const uint8_t *addr, uint16_t port)
 			addr_start_idx = 12;
 		}
 	}
-	inet_ntop(af, addr + addr_start_idx, str, size);
+	ret = inet_ntop(af, addr + addr_start_idx, str, size);
+	if (ret == NULL)
+		panic("failed to convert addr to string, %m");
+
 	if (port) {
 		int  len = strlen(str);
 		snprintf(str + len, size - len, ":%d", port);
@@ -437,6 +448,25 @@ char *addr_to_str(char *str, int size, const uint8_t *addr, uint16_t port)
 	return str;
 }
 
+char *sockaddr_in_to_str(struct sockaddr_in *sockaddr)
+{
+	int i, si;
+	static char str[32];
+	uint8_t *addr;
+
+	si = 0;
+	memset(str, 0, 32);
+
+	addr = (uint8_t *)&sockaddr->sin_addr.s_addr;
+	for (i = 0; i < 4; i++) {
+		si += snprintf(str + si, 32 - si,
+			i != 3 ? "%d." : "%d", addr[i]);
+	}
+	snprintf(str + si, 32 - si, ":%u", sockaddr->sin_port);
+
+	return str;
+}
+
 uint8_t *str_to_addr(const char *ipstr, uint8_t *addr)
 {
 	int addr_start_idx = 0, af = strstr(ipstr, ":") ? AF_INET6 : AF_INET;
@@ -456,12 +486,12 @@ int set_nonblocking(int fd)
 
 	ret = fcntl(fd, F_GETFL);
 	if (ret < 0) {
-		sd_eprintf("fcntl F_GETFL failed: %m\n");
+		sd_eprintf("fcntl F_GETFL failed: %m");
 		close(fd);
 	} else {
 		ret = fcntl(fd, F_SETFL, ret | O_NONBLOCK);
 		if (ret < 0)
-			sd_eprintf("fcntl O_NONBLOCK failed: %m\n");
+			sd_eprintf("fcntl O_NONBLOCK failed: %m");
 	}
 
 	return ret;
@@ -512,22 +542,22 @@ int set_keepalive(int fd)
 	int val = 1;
 
 	if (setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, &val, sizeof(val)) < 0) {
-		sd_dprintf("%m\n");
+		sd_dprintf("%m");
 		return -1;
 	}
 	val = 5;
 	if (setsockopt(fd, SOL_TCP, TCP_KEEPIDLE, &val, sizeof(val)) < 0) {
-		sd_dprintf("%m\n");
+		sd_dprintf("%m");
 		return -1;
 	}
 	val = 1;
 	if (setsockopt(fd, SOL_TCP, TCP_KEEPINTVL, &val, sizeof(val)) < 0) {
-		sd_dprintf("%m\n");
+		sd_dprintf("%m");
 		return -1;
 	}
 	val = 3;
 	if (setsockopt(fd, SOL_TCP, TCP_KEEPCNT, &val, sizeof(val)) < 0) {
-		sd_dprintf("%m\n");
+		sd_dprintf("%m");
 		return -1;
 	}
 	return 0;
@@ -539,7 +569,7 @@ int get_local_addr(uint8_t *bytes)
 	int ret = 0;
 
 	if (getifaddrs(&ifaddr) == -1) {
-		sd_eprintf("getifaddrs failed: %m\n");
+		sd_eprintf("getifaddrs failed: %m");
 		return -1;
 	}
 
@@ -559,17 +589,17 @@ int get_local_addr(uint8_t *bytes)
 			memset(bytes, 0, 12);
 			memcpy(bytes + 12, &sin->sin_addr, 4);
 			memcpy(bytes + 12, &sin->sin_addr, 4);
-			sd_eprintf("found IPv4 address\n");
+			sd_eprintf("found IPv4 address");
 			goto out;
 		case AF_INET6:
 			sin6 = (struct sockaddr_in6 *)ifa->ifa_addr;
 			memcpy(bytes, &sin6->sin6_addr, 16);
-			sd_eprintf("found IPv6 address\n");
+			sd_eprintf("found IPv6 address");
 			goto out;
 		}
 	}
 
-	sd_eprintf("no valid interface found\n");
+	sd_eprintf("no valid interface found");
 	ret = -1;
 out:
 	freeifaddrs(ifaddr);
@@ -587,19 +617,19 @@ int create_unix_domain_socket(const char *unix_path,
 
 	fd = socket(addr.sun_family, SOCK_STREAM, 0);
 	if (fd < 0) {
-		sd_eprintf("failed to create socket, %m\n");
+		sd_eprintf("failed to create socket, %m");
 		return -1;
 	}
 
 	ret = bind(fd, &addr, sizeof(addr));
 	if (ret) {
-		sd_eprintf("failed to bind socket: %m\n");
+		sd_eprintf("failed to bind socket: %m");
 		goto err;
 	}
 
 	ret = listen(fd, SOMAXCONN);
 	if (ret) {
-		sd_eprintf("failed to listen on socket: %m\n");
+		sd_eprintf("failed to listen on socket: %m");
 		goto err;
 	}
 
@@ -625,8 +655,21 @@ bool inetaddr_is_valid(char *addr)
 
 	af = strstr(addr, ":") ? AF_INET6 : AF_INET;
 	if (!inet_pton(af, addr, buf)) {
-		sd_eprintf("Bad address '%s'\n", addr);
+		sd_eprintf("Bad address '%s'", addr);
 		return false;
 	}
 	return true;
 }
+
+int do_writev2(int fd, void *hdr, size_t hdr_len, void *body, size_t body_len)
+{
+	struct iovec iov[2];
+
+	iov[0].iov_base = hdr;
+	iov[0].iov_len = hdr_len;
+
+	iov[1].iov_base = body;
+	iov[1].iov_len = body_len;
+
+	return writev(fd, iov, 2);
+}
