@@ -81,8 +81,8 @@ struct corosync_event {
 
 struct corosync_message {
 	struct cpg_node sender;
-	enum corosync_message_type type : 4;
-	enum cluster_join_result result : 4;
+	enum corosync_message_type type:4;
+	enum cluster_join_result result:4;
 	uint32_t msg_len;
 	uint32_t nr_nodes;
 	struct cpg_node nodes[SD_MAX_NODES];
@@ -106,6 +106,18 @@ static inline int find_cpg_node(struct cpg_node *nodes, size_t nr_nodes,
 	return -1;
 }
 
+static inline int find_sd_node(struct cpg_node *nodes, size_t nr_nodes,
+			       struct sd_node *key)
+{
+	int i;
+
+	for (i = 0; i < nr_nodes; i++)
+		if (node_eq(&nodes[i].ent, key))
+			return i;
+
+	return -1;
+}
+
 static inline void add_cpg_node(struct cpg_node *nodes, size_t nr_nodes,
 				struct cpg_node *added)
 {
@@ -119,7 +131,7 @@ static inline void del_cpg_node(struct cpg_node *nodes, size_t nr_nodes,
 
 	idx = find_cpg_node(nodes, nr_nodes, deled);
 	if (idx < 0) {
-		sd_dprintf("cannot find node\n");
+		sd_dprintf("cannot find node");
 		return;
 	}
 
@@ -139,12 +151,12 @@ static int corosync_get_local_addr(uint8_t *addr)
 	ret = corosync_cfg_get_node_addrs(cfg_handle, this_node.nodeid, 1,
 					  &nr, &caddr);
 	if (ret != CS_OK) {
-		sd_printf(SDOG_ERR, "failed to get node addresses (%d)\n", ret);
+		sd_printf(SDOG_ERR, "failed to get node addresses (%d)", ret);
 		return -1;
 	}
 
 	if (!nr) {
-		sd_printf(SDOG_ERR, "no node addresses found\n");
+		sd_printf(SDOG_ERR, "no node addresses found");
 		return -1;
 	}
 
@@ -156,7 +168,7 @@ static int corosync_get_local_addr(uint8_t *addr)
 		memset(addr, 0, 16);
 		memcpy(addr + 12, saddr, 4);
 	} else {
-		sd_printf(SDOG_ERR, "unknown protocol %d\n", ss->ss_family);
+		sd_printf(SDOG_ERR, "unknown protocol %d", ss->ss_family);
 		return -1;
 	}
 
@@ -194,11 +206,11 @@ retry:
 	case CS_OK:
 		break;
 	case CS_ERR_TRY_AGAIN:
-		sd_dprintf("failed to send message: retrying\n");
+		sd_dprintf("failed to send message: retrying");
 		sleep(1);
 		goto retry;
 	default:
-		sd_eprintf("failed to send message (%d)\n", ret);
+		sd_eprintf("failed to send message (%d)", ret);
 		return -1;
 	}
 	return 0;
@@ -306,8 +318,8 @@ static bool __corosync_dispatch_one(struct corosync_event *cevent)
 			     cevent->msg, cevent->msg_len);
 
 		if (res == CJ_RES_MASTER_TRANSFER) {
-			sd_eprintf("failed to join sheepdog cluster:"	\
-				" please retry when master is up\n");
+			sd_eprintf("failed to join sheepdog cluster:"
+				   " please retry when master is up");
 			exit(1);
 		}
 
@@ -324,8 +336,8 @@ static bool __corosync_dispatch_one(struct corosync_event *cevent)
 		case CJ_RES_FAIL:
 			build_node_list(cpg_nodes, nr_cpg_nodes, entries);
 			sd_join_handler(&cevent->sender.ent, entries,
-						       nr_cpg_nodes, cevent->result,
-						       cevent->msg);
+					nr_cpg_nodes, cevent->result,
+					cevent->msg);
 			break;
 		}
 		break;
@@ -342,8 +354,10 @@ static bool __corosync_dispatch_one(struct corosync_event *cevent)
 		break;
 	case COROSYNC_EVENT_TYPE_BLOCK:
 		if (cevent->callbacked)
-			/* block events until the unblock message
-			   removes this event */
+			/*
+			 * block events until the unblock message
+			 * removes this event
+			 */
 			return false;
 		cevent->callbacked = sd_block_handler(&cevent->sender.ent);
 		return false;
@@ -365,11 +379,13 @@ static void __corosync_dispatch(void)
 	};
 
 	if (poll(&pfd, 1, 0)) {
-		/* Corosync dispatches leave events one by one even
+		/*
+		 * Corosync dispatches leave events one by one even
 		 * when network partition has occured.  To count the
 		 * number of alive nodes correctly, we postpone
-		 * processsing events if there are incoming ones. */
-		sd_dprintf("wait for a next dispatch event\n");
+		 * processsing events if there are incoming ones.
+		 */
+		sd_dprintf("wait for a next dispatch event");
 		return;
 	}
 
@@ -442,7 +458,7 @@ update_event(enum corosync_event_type type, struct cpg_node *sender, void *msg,
 	if (msg_len) {
 		cevent->msg = realloc(cevent->msg, msg_len);
 		if (!cevent->msg)
-			panic("failed to allocate memory\n");
+			panic("failed to allocate memory");
 		memcpy(cevent->msg, msg, msg_len);
 	} else {
 		free(cevent->msg);
@@ -469,7 +485,7 @@ static void cdrv_cpg_deliver(cpg_handle_t handle,
 	struct corosync_message *cmsg = msg;
 	int master;
 
-	sd_dprintf("%d\n", cmsg->type);
+	sd_dprintf("%d", cmsg->type);
 
 	switch (cmsg->type) {
 	case COROSYNC_MSG_TYPE_JOIN_REQUEST:
@@ -514,9 +530,11 @@ static void cdrv_cpg_deliver(cpg_handle_t handle,
 
 		master = is_master(&cmsg->sender);
 		if (master >= 0)
-			/* Master is down before new nodes finish joining.
+			/*
+			 * Master is down before new nodes finish joining.
 			 * We have to revoke its mastership to avoid cluster
-			 * hanging */
+			 * hanging
+			 */
 			cpg_nodes[master].gone = 1;
 
 		cevent->sender = cmsg->sender;
@@ -574,25 +592,26 @@ static void cdrv_cpg_confchg(cpg_handle_t handle,
 	struct cpg_node left_sheep[SD_MAX_NODES];
 	bool promote = true;
 
-	sd_dprintf("mem:%zu, joined:%zu, left:%zu\n",
-		member_list_entries, joined_list_entries,
-		left_list_entries);
+	sd_dprintf("mem:%zu, joined:%zu, left:%zu", member_list_entries,
+		   joined_list_entries, left_list_entries);
 
 	/* check network partition */
 	if (left_list_entries) {
 		if (nr_majority == 0) {
 			size_t total = member_list_entries + left_list_entries;
 
-			/* we need at least 3 nodes to handle network
-			 * partition failure */
+			/*
+			 * we need at least 3 nodes to handle network
+			 * partition failure
+			 */
 			if (total > 2)
 				nr_majority = total / 2 + 1;
 		}
 
 		if (member_list_entries == 0)
-			panic("NIC failure?\n");
+			panic("NIC failure?");
 		if (member_list_entries < nr_majority)
-			panic("Network partition is detected\n");
+			panic("Network partition is detected");
 	}
 
 	/* convert cpg_address to cpg_node */
@@ -624,9 +643,11 @@ static void cdrv_cpg_confchg(cpg_handle_t handle,
 		cevent = xzalloc(sizeof(*cevent));
 		master = is_master(&left_sheep[i]);
 		if (master >= 0)
-		/* Master is down before new nodes finish joining.
-		 * We have to revoke its mastership to avoid cluster hanging
-		 */
+			/*
+			 * Master is down before new nodes finish joining.
+			 * We have to revoke its mastership to avoid cluster
+			 * hanging
+			 */
 			cpg_nodes[master].gone = 1;
 
 		cevent->type = COROSYNC_EVENT_TYPE_LEAVE;
@@ -653,7 +674,7 @@ static void cdrv_cpg_confchg(cpg_handle_t handle,
 					    &member_sheep[i]);
 			if (!cevent) {
 				sd_dprintf("Not promoting because member is "
-					"not in our event list.\n");
+					   "not in our event list.");
 				promote = false;
 				break;
 			}
@@ -680,14 +701,14 @@ retry:
 	case CS_OK:
 		break;
 	case CS_ERR_TRY_AGAIN:
-		sd_dprintf("failed to join the sheepdog group: retrying\n");
+		sd_dprintf("failed to join the sheepdog group: retrying");
 		sleep(1);
 		goto retry;
 	case CS_ERR_SECURITY:
-		sd_eprintf("permission denied to join the sheepdog group\n");
+		sd_eprintf("permission denied to join the sheepdog group");
 		return -1;
 	default:
-		sd_eprintf("failed to join the sheepdog group (%d)\n", ret);
+		sd_eprintf("failed to join the sheepdog group (%d)", ret);
 		return -1;
 	}
 
@@ -728,14 +749,13 @@ static void corosync_handler(int listen_fd, int events, void *data)
 	int ret;
 
 	if (events & EPOLLHUP) {
-		sd_eprintf("corosync driver received EPOLLHUP event,"	\
-			" exiting.\n");
+		sd_eprintf("corosync driver received EPOLLHUP event, exiting.");
 		goto out;
 	}
 
 	ret = cpg_dispatch(cpg_handle, CS_DISPATCH_ALL);
 	if (ret != CS_OK) {
-		sd_eprintf("cpg_dispatch returned %d\n", ret);
+		sd_eprintf("cpg_dispatch returned %d", ret);
 		goto out;
 	}
 
@@ -763,27 +783,27 @@ again:
 	case CS_ERR_TRY_AGAIN:
 		if (retry_cnt++ == CPG_INIT_RETRY_CNT) {
 			sd_eprintf("failed to initialize cpg (%d) - "
-				"is corosync running?\n", ret);
+				   "is corosync running?", ret);
 			return -1;
 		}
-		sd_dprintf("retry cpg_initialize\n");
+		sd_dprintf("retry cpg_initialize");
 		usleep(200000);
 		goto again;
 	default:
 		sd_eprintf("failed to initialize cpg (%d) - "
-			"is corosync running?\n", ret);
+			   "is corosync running?", ret);
 		return -1;
 	}
 
 	ret = corosync_cfg_initialize(&cfg_handle, NULL);
 	if (ret != CS_OK) {
-		sd_printf(SDOG_ERR, "failed to initialize cfg (%d)\n", ret);
+		sd_printf(SDOG_ERR, "failed to initialize cfg (%d)", ret);
 		return -1;
 	}
 
 	ret = corosync_cfg_local_get(cfg_handle, &nodeid);
 	if (ret != CS_OK) {
-		sd_printf(SDOG_ERR, "failed to get node id (%d)\n", ret);
+		sd_printf(SDOG_ERR, "failed to get node id (%d)", ret);
 		return -1;
 	}
 
@@ -792,20 +812,30 @@ again:
 
 	ret = cpg_fd_get(cpg_handle, &cpg_fd);
 	if (ret != CS_OK) {
-		sd_eprintf("failed to get cpg file descriptor (%d)\n", ret);
+		sd_eprintf("failed to get cpg file descriptor (%d)", ret);
 		return -1;
 	}
 
 	ret = register_event(cpg_fd, corosync_handler, NULL);
 	if (ret) {
-		sd_eprintf("failed to register corosync event handler (%d)\n",
-			ret);
+		sd_eprintf("failed to register corosync event handler (%d)",
+			   ret);
 		return -1;
 	}
 
 	return 0;
 }
 
+static void corosync_update_node(struct sd_node *node)
+{
+	int idx;
+
+	idx = find_sd_node(cpg_nodes, nr_cpg_nodes, node);
+	if (idx < 0)
+		return;
+	cpg_nodes[idx].ent = *node;
+}
+
 static struct cluster_driver cdrv_corosync = {
 	.name		= "corosync",
 
@@ -816,6 +846,7 @@ static struct cluster_driver cdrv_corosync = {
 	.notify		= corosync_notify,
 	.block		= corosync_block,
 	.unblock	= corosync_unblock,
+	.update_node	= corosync_update_node,
 };
 
 cdrv_register(cdrv_corosync);
