@@ -36,8 +36,8 @@
 	for (zk_get_children(parent, strs),		               \
 		     (strs)->data += (strs)->count;		       \
 	     (strs)->count-- ?					       \
-		     sprintf(path, "%s/%s", parent, *--(strs)->data) : \
-		     (free((strs)->data), 0);			       \
+		     snprintf(path, sizeof(path), "%s/%s", parent,     \
+			      *--(strs)->data) : (free((strs)->data), 0); \
 	     free(*(strs)->data))
 
 enum zk_event_type {
@@ -58,6 +58,7 @@ struct zk_node {
 };
 
 struct zk_event {
+	uint64_t id;
 	enum zk_event_type type;
 	struct zk_node sender;
 	enum cluster_join_result join_result;
@@ -85,7 +86,7 @@ static struct zk_node *zk_tree_insert(struct zk_node *new)
 		parent = *p;
 		entry = rb_entry(parent, struct zk_node, rb);
 
-		cmp = node_id_cmp(&new->node.nid, &entry->node.nid);
+		cmp = node_cmp(&new->node, &entry->node);
 		if (cmp < 0)
 			p = &(*p)->rb_left;
 		else if (cmp > 0)
@@ -141,7 +142,7 @@ static inline ZOOAPI int zk_delete_node(const char *path, int version)
 		rc = zoo_delete(zhandle, path, version);
 	} while (rc == ZOPERATIONTIMEOUT || rc == ZCONNECTIONLOSS);
 	if (rc != ZOK)
-		sd_eprintf("failed, path:%s, %s\n", path, zerror(rc));
+		sd_eprintf("failed, path:%s, %s", path, zerror(rc));
 	return rc;
 }
 
@@ -155,7 +156,7 @@ zk_init_node(const char *path)
 	} while (rc == ZOPERATIONTIMEOUT || rc == ZCONNECTIONLOSS);
 
 	if (rc != ZOK && rc != ZNODEEXISTS)
-		panic("failed, path:%s, %s\n", path, zerror(rc));
+		panic("failed, path:%s, %s", path, zerror(rc));
 }
 
 static inline ZOOAPI int
@@ -168,17 +169,19 @@ zk_create_node(const char *path, const char *value, int valuelen,
 		rc = zoo_create(zhandle, path, value, valuelen, acl,
 				flags, path_buffer, path_buffer_len);
 		if (rc != ZOK && rc != ZNODEEXISTS)
-			sd_eprintf("failed, path:%s, %s\n", path, zerror(rc));
+			sd_eprintf("failed, path:%s, %s", path, zerror(rc));
 	} while (rc == ZOPERATIONTIMEOUT || rc == ZCONNECTIONLOSS);
 	return rc;
 }
 
 /*
- * FIXME: ZOPERATIONTIMEOUT or ZCONNECTIONLOSS will cause duplicate seq node of
- * the same event that we can't handle at all. Panic out in any error case would
- * be the safest option.
+ * Create a znode after adding a unique monotonically increasing sequence number
+ * to the path name.
+ *
+ * Note that the caller has to retry this function when this returns
+ * ZOPERATIONTIMEOUT or ZCONNECTIONLOSS and the znode is not created.
  */
-static inline ZOOAPI void
+static inline ZOOAPI int
 zk_create_seq_node(const char *path, const char *value, int valuelen,
 		   char *path_buffer, int path_buffer_len)
 {
@@ -186,7 +189,9 @@ zk_create_seq_node(const char *path, const char *value, int valuelen,
 	rc = zoo_create(zhandle, path, value, valuelen, &ZOO_OPEN_ACL_UNSAFE,
 			ZOO_SEQUENCE, path_buffer, path_buffer_len);
 	if (rc != ZOK)
-		panic("failed, path:%s, %s\n", path, zerror(rc));
+		sd_iprintf("failed, path:%s, %s", path, zerror(rc));
+
+	return rc;
 }
 
 static inline ZOOAPI int zk_get_data(const char *path, void *buffer,
@@ -197,7 +202,7 @@ static inline ZOOAPI int zk_get_data(const char *path, void *buffer,
 		rc = zoo_get(zhandle, path, 1, (char *)buffer,
 			     buffer_len, NULL);
 		if (rc != ZOK)
-			sd_eprintf("failed, path:%s, %s\n", path, zerror(rc));
+			sd_eprintf("failed, path:%s, %s", path, zerror(rc));
 	} while (rc == ZOPERATIONTIMEOUT || rc == ZCONNECTIONLOSS);
 	return rc;
 }
@@ -210,7 +215,7 @@ zk_set_data(const char *path, const char *buffer, int buflen, int version)
 		rc = zoo_set(zhandle, path, buffer, buflen, version);
 	} while (rc == ZOPERATIONTIMEOUT || rc == ZCONNECTIONLOSS);
 	if (rc != ZOK)
-		panic("failed, path:%s, %s\n", path, zerror(rc));
+		panic("failed, path:%s, %s", path, zerror(rc));
 	return rc;
 }
 
@@ -220,7 +225,7 @@ static inline ZOOAPI int zk_node_exists(const char *path)
 	do {
 		rc = zoo_exists(zhandle, path, 1, NULL);
 		if (rc != ZOK && rc != ZNONODE)
-			sd_eprintf("failed, path:%s, %s\n", path, zerror(rc));
+			sd_eprintf("failed, path:%s, %s", path, zerror(rc));
 	} while (rc == ZOPERATIONTIMEOUT || rc == ZCONNECTIONLOSS);
 
 	return rc;
@@ -234,7 +239,7 @@ static inline ZOOAPI void zk_get_children(const char *path,
 		rc = zoo_get_children(zhandle, path, 1, strings);
 	} while (rc == ZOPERATIONTIMEOUT || rc == ZCONNECTIONLOSS);
 	if (rc != ZOK)
-		panic("failed, path:%s, %s\n", path, zerror(rc));
+		panic("failed, path:%s, %s", path, zerror(rc));
 }
 
 /* ZooKeeper-based queue give us an totally ordered events */
@@ -244,9 +249,9 @@ static int32_t queue_pos;
 static bool zk_queue_peek(void)
 {
 	int rc;
-	char path[256];
+	char path[MAX_NODE_STR_LEN];
 
-	sprintf(path, QUEUE_ZNODE "/%010"PRId32, queue_pos);
+	snprintf(path, sizeof(path), QUEUE_ZNODE "/%010"PRId32, queue_pos);
 
 	rc = zk_node_exists(path);
 	if (rc == ZOK)
@@ -255,15 +260,60 @@ static bool zk_queue_peek(void)
 	return false;
 }
 
+/* return true if there is a node with 'id' in the queue. */
+static bool zk_find_seq_node(uint64_t id)
+{
+	int rc, len;
+
+	for (int seq = queue_pos; ; seq++) {
+		char seq_path[MAX_NODE_STR_LEN];
+		struct zk_event ev;
+
+		snprintf(seq_path, sizeof(seq_path), QUEUE_ZNODE"/%010"PRId32,
+			 seq);
+		len = offsetof(typeof(ev), id) + sizeof(ev.id);
+		rc = zk_get_data(seq_path, &ev, &len);
+		switch (rc) {
+		case ZOK:
+			if (ev.id == id) {
+				sd_dprintf("id %"PRIx64" is found in %s", id,
+					   seq_path);
+				return true;
+			}
+			break;
+		case ZNONODE:
+			sd_dprintf("id %"PRIx64" is not found", id);
+			return false;
+		default:
+			panic("failed, %s", zerror(rc));
+			break;
+		}
+	}
+}
+
 static void zk_queue_push(struct zk_event *ev)
 {
 	static bool first_push = true;
-	int len;
-	char path[256], buf[256];
+	int rc, len;
+	char path[MAX_NODE_STR_LEN], buf[MAX_NODE_STR_LEN];
 
-	len = (char *)(ev->buf) - (char *)ev + ev->buf_len;
-	sprintf(path, "%s/", QUEUE_ZNODE);
-	zk_create_seq_node(path, (char *)ev, len, buf, sizeof(buf));
+	len = offsetof(typeof(*ev), buf) + ev->buf_len;
+	snprintf(path, sizeof(path), "%s/", QUEUE_ZNODE);
+again:
+	rc = zk_create_seq_node(path, (char *)ev, len, buf, sizeof(buf));
+	switch (rc) {
+	case ZOK:
+		/* Success */
+		break;
+	case ZOPERATIONTIMEOUT:
+	case ZCONNECTIONLOSS:
+		if (!zk_find_seq_node(ev->id))
+			/* retry if seq_node was not created */
+			goto again;
+		break;
+	default:
+		panic("failed, path:%s, %s", path, zerror(rc));
+	}
 	if (first_push) {
 		int32_t seq;
 
@@ -273,8 +323,8 @@ static void zk_queue_push(struct zk_event *ev)
 		first_push = false;
 	}
 
-	sd_dprintf("create path:%s, queue_pos:%010"PRId32", len:%d\n",
-		buf, queue_pos, len);
+	sd_dprintf("create path:%s, queue_pos:%010"PRId32", len:%d",
+		   buf, queue_pos, len);
 }
 
 static inline void *zk_event_sd_nodes(struct zk_event *ev)
@@ -285,7 +335,7 @@ static inline void *zk_event_sd_nodes(struct zk_event *ev)
 /* Change the join event in place and piggyback the nodes information. */
 static void push_join_response(struct zk_event *ev)
 {
-	char path[256];
+	char path[MAX_NODE_STR_LEN];
 	int len;
 
 	ev->type = EVENT_JOIN_RESPONSE;
@@ -294,23 +344,25 @@ static void push_join_response(struct zk_event *ev)
 	       nr_sd_nodes * sizeof(struct sd_node));
 	queue_pos--;
 
-	len = (char *)(ev->buf) - (char *)ev + ev->buf_len;
-	sprintf(path, QUEUE_ZNODE "/%010"PRId32, queue_pos);
+	len = offsetof(typeof(*ev), buf) + ev->buf_len;
+	snprintf(path, sizeof(path), QUEUE_ZNODE "/%010"PRId32, queue_pos);
 	zk_set_data(path, (char *)ev, len, -1);
-	sd_dprintf("update path:%s, queue_pos:%010"PRId32", len:%d\n",
-		path, queue_pos, len);
+	sd_dprintf("update path:%s, queue_pos:%010"PRId32", len:%d",
+		   path, queue_pos, len);
 }
 
 static void zk_queue_pop_advance(struct zk_event *ev)
 {
-	int len;
-	char path[256];
+	int rc, len;
+	char path[MAX_NODE_STR_LEN];
 
 	len = sizeof(*ev);
-	sprintf(path, QUEUE_ZNODE "/%010"PRId32, queue_pos);
-	assert(zk_get_data(path, ev, &len) == ZOK);
-	sd_dprintf("%s, type:%d, len:%d, pos:%"PRId32"\n",
-		path, ev->type, len, queue_pos);
+	snprintf(path, sizeof(path), QUEUE_ZNODE "/%010"PRId32, queue_pos);
+	rc = zk_get_data(path, ev, &len);
+	if (rc != ZOK)
+		panic("failed to get data from %s, %s", path, zerror(rc));
+	sd_dprintf("%s, type:%d, len:%d, pos:%"PRId32, path, ev->type, len,
+		   queue_pos);
 	queue_pos++;
 }
 
@@ -377,7 +429,7 @@ static inline void build_node_list(void)
 		zk = rb_entry(n, struct zk_node, rb);
 		sd_nodes[nr_sd_nodes++] = zk->node;
 	}
-	sd_dprintf("nr_sd_nodes:%zu\n", nr_sd_nodes);
+	sd_dprintf("nr_sd_nodes:%zu", nr_sd_nodes);
 }
 
 static inline int zk_master_create(void)
@@ -416,11 +468,24 @@ static void zk_queue_init(void)
 	zk_init_node(MEMBER_ZNODE);
 }
 
+/* Calculate a unique 64 bit integer from this_node and the sequence number. */
+static uint64_t get_uniq_id(void)
+{
+	static int seq;
+	uint64_t id, n = uatomic_add_return(&seq, 1);
+
+	id = fnv_64a_buf(&this_node, sizeof(this_node), FNV1A_64_INIT);
+	id = fnv_64a_buf(&n, sizeof(n), id);
+
+	return id;
+}
+
 static int add_event(enum zk_event_type type, struct zk_node *znode, void *buf,
 		     size_t buf_len)
 {
 	struct zk_event ev;
 
+	ev.id = get_uniq_id();
 	ev.type = type;
 	ev.sender = *znode;
 	ev.buf_len = buf_len;
@@ -434,11 +499,11 @@ static void zk_watcher(zhandle_t *zh, int type, int state, const char *path,
 		       void *ctx)
 {
 	struct zk_node znode;
-	char str[256], *p;
+	char str[MAX_NODE_STR_LEN], *p;
 	int ret;
 
 /* CREATED_EVENT 1, DELETED_EVENT 2, CHANGED_EVENT 3, CHILD_EVENT 4 */
-	sd_dprintf("path:%s, type:%d\n", path, type);
+	sd_dprintf("path:%s, type:%d", path, type);
 	if (type == ZOO_CREATED_EVENT || type == ZOO_CHANGED_EVENT) {
 		ret = sscanf(path, MEMBER_ZNODE "/%s", str);
 		if (ret == 1)
@@ -476,6 +541,7 @@ static int add_join_event(void *msg, size_t msg_len)
 	size_t len = msg_len + sizeof(struct sd_node) * SD_MAX_NODES;
 
 	assert(len <= SD_MAX_EVENT_BUF_SIZE);
+	ev.id = get_uniq_id();
 	ev.type = EVENT_JOIN_REQUEST;
 	ev.sender = this_node;
 	ev.msg_len = msg_len;
@@ -490,14 +556,14 @@ static int zk_join(const struct sd_node *myself,
 		   void *opaque, size_t opaque_len)
 {
 	int rc;
-	char path[256];
+	char path[MAX_NODE_STR_LEN];
 
 	this_node.node = *myself;
 
-	sprintf(path, MEMBER_ZNODE "/%s", node_to_str(myself));
+	snprintf(path, sizeof(path), MEMBER_ZNODE "/%s", node_to_str(myself));
 	rc = zk_node_exists(path);
 	if (rc == ZOK) {
-		sd_eprintf("Previous zookeeper session exist, shoot myself.\n");
+		sd_eprintf("Previous zookeeper session exist, shoot myself.");
 		exit(1);
 	}
 
@@ -510,7 +576,12 @@ static int zk_join(const struct sd_node *myself,
 
 static int zk_leave(void)
 {
-	return add_event(EVENT_LEAVE, &this_node, NULL, 0);
+	char path[PATH_MAX];
+	snprintf(path, sizeof(path), MEMBER_ZNODE"/%s",
+			node_to_str(&this_node.node));
+	add_event(EVENT_LEAVE, &this_node, NULL, 0);
+	zk_delete_node(path, -1);
+	return 0;
 }
 
 static int zk_notify(void *msg, size_t msg_len)
@@ -532,7 +603,7 @@ static void zk_handle_join_request(struct zk_event *ev)
 {
 	enum cluster_join_result res;
 
-	sd_dprintf("sender: %s\n", node_to_str(&ev->sender.node));
+	sd_dprintf("sender: %s", node_to_str(&ev->sender.node));
 	if (!is_master()) {
 		/* Let's await master acking the join-request */
 		queue_pos--;
@@ -544,18 +615,18 @@ static void zk_handle_join_request(struct zk_event *ev)
 	push_join_response(ev);
 	if (res == CJ_RES_MASTER_TRANSFER) {
 		sd_eprintf("failed to join sheepdog cluster: "
-			"please retry when master is up\n");
+			   "please retry when master is up");
 		add_event(EVENT_LEAVE, &this_node, NULL, 0);
 		exit(1);
 	}
-	sd_dprintf("I'm the master now\n");
+	sd_dprintf("I'm the master now");
 }
 
 static void watch_all_nodes(void)
 {
 	struct String_vector strs;
 	struct zk_node znode;
-	char path[256];
+	char path[MAX_NODE_STR_LEN];
 	int len = sizeof(znode);
 
 	if (zk_member_empty())
@@ -572,7 +643,7 @@ static void init_node_list(struct zk_event *ev)
 	size_t node_nr = ev->nr_nodes;
 	int i;
 
-	sd_dprintf("%zu\n", node_nr);
+	sd_dprintf("%zu", node_nr);
 	for (i = 0; i < node_nr; i++) {
 		struct zk_node zk;
 		mempcpy(&zk.node, p, sizeof(struct sd_node));
@@ -585,9 +656,9 @@ static void init_node_list(struct zk_event *ev)
 
 static void zk_handle_join_response(struct zk_event *ev)
 {
-	char path[256];
+	char path[MAX_NODE_STR_LEN];
 
-	sd_dprintf("JOIN RESPONSE\n");
+	sd_dprintf("JOIN RESPONSE");
 	if (node_eq(&ev->sender.node, &this_node.node))
 		/* newly joined node */
 		init_node_list(ev);
@@ -600,14 +671,15 @@ static void zk_handle_join_response(struct zk_event *ev)
 		 */
 		zk_tree_destroy();
 
-	sd_dprintf("%s, %d\n", node_to_str(&ev->sender.node), ev->join_result);
+	sd_dprintf("%s, %d", node_to_str(&ev->sender.node), ev->join_result);
 	switch (ev->join_result) {
 	case CJ_RES_SUCCESS:
 	case CJ_RES_JOIN_LATER:
 	case CJ_RES_MASTER_TRANSFER:
-		sprintf(path, MEMBER_ZNODE"/%s", node_to_str(&ev->sender.node));
+		snprintf(path, sizeof(path), MEMBER_ZNODE"/%s",
+			 node_to_str(&ev->sender.node));
 		if (node_eq(&ev->sender.node, &this_node.node)) {
-			sd_dprintf("create path:%s\n", path);
+			sd_dprintf("create path:%s", path);
 			zk_create_node(path, (char *)&ev->sender,
 				       sizeof(ev->sender), &ZOO_OPEN_ACL_UNSAFE,
 				       ZOO_EPHEMERAL, NULL, 0);
@@ -653,8 +725,8 @@ static void zk_handle_leave(struct zk_event *ev)
 	struct zk_node *n = zk_tree_search(&ev->sender.node.nid);
 
 	if (!n) {
-		sd_dprintf("can't find this leave node:%s, ignore it.\n",
-			node_to_str(&ev->sender.node));
+		sd_dprintf("can't find this leave node:%s, ignore it.",
+			   node_to_str(&ev->sender.node));
 		return;
 	}
 	block_event_list_del(n);
@@ -667,7 +739,7 @@ static void zk_handle_block(struct zk_event *ev)
 {
 	struct zk_node *block = xzalloc(sizeof(*block));
 
-	sd_dprintf("BLOCK\n");
+	sd_dprintf("BLOCK");
 	block->node = ev->sender.node;
 	list_add_tail(&block->list, &zk_block_list);
 	block = list_first_entry(&zk_block_list, typeof(*block), list);
@@ -679,7 +751,7 @@ static void zk_handle_unblock(struct zk_event *ev)
 {
 	struct zk_node *block;
 
-	sd_dprintf("UNBLOCK\n");
+	sd_dprintf("UNBLOCK");
 	if (list_empty(&zk_block_list))
 		return;
 	block = list_first_entry(&zk_block_list, typeof(*block), list);
@@ -692,7 +764,7 @@ static void zk_handle_unblock(struct zk_event *ev)
 
 static void zk_handle_notify(struct zk_event *ev)
 {
-	sd_dprintf("NOTIFY\n");
+	sd_dprintf("NOTIFY");
 	sd_notify_handler(&ev->sender.node, ev->buf, ev->buf_len);
 }
 
@@ -712,16 +784,16 @@ static void zk_event_handler(int listen_fd, int events, void *data)
 	eventfd_t value;
 	struct zk_event ev;
 
-	sd_dprintf("%d, %d\n", events, queue_pos);
+	sd_dprintf("%d, %d", events, queue_pos);
 	if (events & EPOLLHUP) {
 		sd_eprintf("zookeeper driver received EPOLLHUP event,"
-			" exiting.\n");
+			   " exiting.");
 		log_close();
 		exit(1);
 	}
 
 	if (eventfd_read(efd, &value) < 0) {
-		sd_eprintf("%m\n");
+		sd_eprintf("%m");
 		return;
 	}
 
@@ -732,7 +804,7 @@ static void zk_event_handler(int listen_fd, int events, void *data)
 	if (ev.type < zk_max_event_handlers && zk_event_handlers[ev.type])
 		zk_event_handlers[ev.type](&ev);
 	else
-		panic("unhandled type %d\n", ev.type);
+		panic("unhandled type %d", ev.type);
 
 	 /* Someone has created next event, go kick event handler. */
 	if (zk_queue_peek()) {
@@ -758,25 +830,25 @@ static int zk_init(const char *option)
 	int ret, timeout = SESSION_TIMEOUT;
 
 	if (!option) {
-		sd_eprintf("You must specify zookeeper servers.\n");
+		sd_eprintf("You must specify zookeeper servers.");
 		return -1;
 	}
 
 	hosts = strtok((char *)option, "=");
 	if ((to = strtok(NULL, "="))) {
 		if (sscanf(to, "%u", &timeout) != 1) {
-			sd_eprintf("Invalid paramter for timeout\n");
+			sd_eprintf("Invalid paramter for timeout");
 			return -1;
 		}
 		p = strstr(hosts, "timeout");
 		*--p = '\0';
 	}
-	sd_dprintf("version %d.%d.%d, address %s, timeout %d\n",
-		ZOO_MAJOR_VERSION, ZOO_MINOR_VERSION, ZOO_PATCH_VERSION,
-		hosts, timeout);
+	sd_dprintf("version %d.%d.%d, address %s, timeout %d",
+		   ZOO_MAJOR_VERSION, ZOO_MINOR_VERSION, ZOO_PATCH_VERSION,
+		   hosts, timeout);
 	zhandle = zookeeper_init(hosts, zk_watcher, timeout, NULL, NULL, 0);
 	if (!zhandle) {
-		sd_eprintf("failed to connect to zk server %s\n", option);
+		sd_eprintf("failed to connect to zk server %s", option);
 		return -1;
 	}
 
@@ -784,20 +856,38 @@ static int zk_init(const char *option)
 
 	efd = eventfd(0, EFD_NONBLOCK);
 	if (efd < 0) {
-		sd_eprintf("failed to create an event fd: %m\n");
+		sd_eprintf("failed to create an event fd: %m");
 		return -1;
 	}
 
 	ret = register_event(efd, zk_event_handler, NULL);
 	if (ret) {
-		sd_eprintf("failed to register zookeeper event handler (%d)\n",
-			ret);
+		sd_eprintf("failed to register zookeeper event handler (%d)",
+			   ret);
 		return -1;
 	}
 
 	return 0;
 }
 
+static void zk_update_node(struct sd_node *node)
+{
+	struct zk_node n = {
+		.node = *node,
+	};
+	struct zk_node *t;
+
+	sd_dprintf("%s", node_to_str(&n.node));
+
+	pthread_rwlock_rdlock(&zk_tree_lock);
+	t = zk_tree_search_nolock(&n.node.nid);
+	if (t) {
+		t->node = n.node;
+		build_node_list();
+	}
+	pthread_rwlock_unlock(&zk_tree_lock);
+}
+
 static struct cluster_driver cdrv_zookeeper = {
 	.name       = "zookeeper",
 
@@ -807,6 +897,7 @@ static struct cluster_driver cdrv_zookeeper = {
 	.notify     = zk_notify,
 	.block      = zk_block,
 	.unblock    = zk_unblock,
+	.update_node = zk_update_node,
 };
 
 cdrv_register(cdrv_zookeeper);
