@@ -19,6 +19,7 @@
 #include <pthread.h>
 #include <urcu/uatomic.h>
 #include <math.h>
+#include <time.h>
 
 #include "sheepdog_proto.h"
 #include "sheep_priv.h"
@@ -36,6 +37,7 @@ struct node {
 struct get_vdis_work {
 	struct work work;
 	DECLARE_BITMAP(vdi_inuse, SD_NR_VDIS);
+	struct sd_node joined;
 	size_t nr_members;
 	struct sd_node members[];
 };
@@ -44,7 +46,21 @@ static pthread_mutex_t wait_vdis_lock = PTHREAD_MUTEX_INITIALIZER;
 static pthread_cond_t wait_vdis_cond = PTHREAD_COND_INITIALIZER;
 static bool is_vdi_list_ready = true;
 
-static struct vnode_info *current_vnode_info;
+static main_thread(struct vnode_info *) current_vnode_info;
+static main_thread(struct list_head *) pending_block_list;
+static main_thread(struct list_head *) pending_notify_list;
+
+/*
+ * List of nodes that were part of the last epoch before a shutdown,
+ * but failed to join.
+ */
+static main_thread(struct list_head *) failed_nodes;
+
+/*
+ * List of nodes that weren't part of the last epoch, but joined
+ * before restarting the cluster.
+ */
+static main_thread(struct list_head *) delayed_nodes;
 
 static size_t get_join_message_size(struct join_message *jm)
 {
@@ -86,26 +102,27 @@ static int get_zones_nr_from(const struct sd_node *nodes, int nr_nodes)
 bool have_enough_zones(void)
 {
 	int max_copies;
+	struct vnode_info *cur_vinfo = main_thread_get(current_vnode_info);
 
 	if (sys->flags & SD_FLAG_NOHALT)
 		return true;
 
-	if (!current_vnode_info)
+	if (!cur_vinfo)
 		return false;
 
 	max_copies = get_max_copy_number();
 
-	sd_dprintf("flags %d, nr_zones %d, min copies %d\n",
-		sys->flags, current_vnode_info->nr_zones, max_copies);
+	sd_dprintf("flags %d, nr_zones %d, min copies %d",
+		   sys->flags, cur_vinfo->nr_zones, max_copies);
 
-	if (!current_vnode_info->nr_zones)
+	if (!cur_vinfo->nr_zones)
 		return false;
 
 	if (sys->flags & SD_FLAG_QUORUM) {
-		if (current_vnode_info->nr_zones > (max_copies/2))
+		if (cur_vinfo->nr_zones > (max_copies/2))
 			return true;
 	} else {
-		if (current_vnode_info->nr_zones >= max_copies)
+		if (cur_vinfo->nr_zones >= max_copies)
 			return true;
 	}
 	return false;
@@ -113,8 +130,7 @@ bool have_enough_zones(void)
 
 static int get_node_idx(struct vnode_info *vnode_info, struct sd_node *ent)
 {
-	ent = bsearch(ent, vnode_info->nodes, vnode_info->nr_nodes,
-		      sizeof(*ent), node_id_cmp);
+	ent = xbsearch(ent, vnode_info->nodes, vnode_info->nr_nodes, node_cmp);
 	if (!ent)
 		return -1;
 
@@ -139,19 +155,19 @@ struct vnode_info *grab_vnode_info(struct vnode_info *vnode_info)
 /*
  * Get a reference to the currently active vnode information structure,
  * this must only be called from the main thread.
+ * This can return NULL if cluster is not started yet.
  */
 struct vnode_info *get_vnode_info(void)
 {
-	assert(current_vnode_info);
+	struct vnode_info *cur_vinfo = main_thread_get(current_vnode_info);
 
-	return grab_vnode_info(current_vnode_info);
+	if (cur_vinfo == NULL)
+		return NULL;
+
+	return grab_vnode_info(cur_vinfo);
 }
 
-/*
- * Release a reference to the current vnode information.
- *
- * Must be called from the main thread.
- */
+/* Release a reference to the current vnode information. */
 void put_vnode_info(struct vnode_info *vnode_info)
 {
 	if (vnode_info) {
@@ -162,8 +178,8 @@ void put_vnode_info(struct vnode_info *vnode_info)
 	}
 }
 
-static struct vnode_info *alloc_vnode_info(const struct sd_node *nodes,
-					   size_t nr_nodes)
+struct vnode_info *alloc_vnode_info(const struct sd_node *nodes,
+				    size_t nr_nodes)
 {
 	struct vnode_info *vnode_info;
 
@@ -171,7 +187,7 @@ static struct vnode_info *alloc_vnode_info(const struct sd_node *nodes,
 
 	vnode_info->nr_nodes = nr_nodes;
 	memcpy(vnode_info->nodes, nodes, sizeof(*nodes) * nr_nodes);
-	qsort(vnode_info->nodes, nr_nodes, sizeof(*nodes), node_id_cmp);
+	xqsort(vnode_info->nodes, nr_nodes, node_cmp);
 
 	recalculate_vnodes(vnode_info->nodes, nr_nodes);
 
@@ -182,14 +198,16 @@ static struct vnode_info *alloc_vnode_info(const struct sd_node *nodes,
 	return vnode_info;
 }
 
-struct vnode_info *get_vnode_info_epoch(uint32_t epoch)
+struct vnode_info *get_vnode_info_epoch(uint32_t epoch,
+					struct vnode_info *cur_vinfo)
 {
 	struct sd_node nodes[SD_MAX_NODES];
 	int nr_nodes;
 
 	nr_nodes = epoch_log_read(epoch, nodes, sizeof(nodes));
 	if (nr_nodes < 0) {
-		nr_nodes = epoch_log_read_remote(epoch, nodes, sizeof(nodes));
+		nr_nodes = epoch_log_read_remote(epoch, nodes, sizeof(nodes),
+						 NULL, cur_vinfo);
 		if (nr_nodes == 0)
 			return NULL;
 	}
@@ -200,29 +218,26 @@ struct vnode_info *get_vnode_info_epoch(uint32_t epoch)
 int local_get_node_list(const struct sd_req *req, struct sd_rsp *rsp,
 			       void *data)
 {
-	struct sd_node_rsp *node_rsp = (struct sd_node_rsp *)rsp;
 	int nr_nodes;
+	struct vnode_info *cur_vinfo = main_thread_get(current_vnode_info);
 
-	if (current_vnode_info) {
-		nr_nodes = current_vnode_info->nr_nodes;
-		memcpy(data, current_vnode_info->nodes,
+	if (cur_vinfo) {
+		nr_nodes = cur_vinfo->nr_nodes;
+		memcpy(data, cur_vinfo->nodes,
 			sizeof(struct sd_node) * nr_nodes);
-		node_rsp->data_length = nr_nodes * sizeof(struct sd_node);
-		node_rsp->nr_nodes = nr_nodes;
-		node_rsp->local_idx = get_node_idx(current_vnode_info,
-						   &sys->this_node);
+		rsp->data_length = nr_nodes * sizeof(struct sd_node);
+		rsp->node.nr_nodes = nr_nodes;
+		rsp->node.local_idx = get_node_idx(cur_vinfo, &sys->this_node);
 	} else {
-		node_rsp->nr_nodes = 0;
-		node_rsp->local_idx = 0;
+		rsp->node.nr_nodes = 0;
+		rsp->node.local_idx = 0;
 	}
 
-	node_rsp->master_idx = -1;
+	rsp->node.master_idx = -1;
 	return SD_RES_SUCCESS;
 }
 
-/*
- * Indicator if a cluster operation is currently running.
- */
+/* Indicator if a cluster operation is currently running. */
 static bool cluster_op_running;
 
 static struct vdi_op_message *prepare_cluster_msg(struct request *req,
@@ -232,9 +247,11 @@ static struct vdi_op_message *prepare_cluster_msg(struct request *req,
 	size_t size;
 
 	if (has_process_main(req->op) && req->rq.flags & SD_FLAG_CMD_WRITE)
+		/* notify data that was received from the sender */
 		size = sizeof(*msg) + req->rq.data_length;
 	else
-		size = sizeof(*msg);
+		/* notify data that was set in process_work */
+		size = sizeof(*msg) + req->rp.data_length;
 
 	assert(size <= SD_MAX_EVENT_BUF_SIZE);
 
@@ -242,8 +259,8 @@ static struct vdi_op_message *prepare_cluster_msg(struct request *req,
 	memcpy(&msg->req, &req->rq, sizeof(struct sd_req));
 	memcpy(&msg->rsp, &req->rp, sizeof(struct sd_rsp));
 
-	if (has_process_main(req->op) && req->rq.flags & SD_FLAG_CMD_WRITE)
-		memcpy(msg->data, req->data, req->rq.data_length);
+	if (has_process_main(req->op) && size > sizeof(*msg))
+		memcpy(msg->data, req->data, size - sizeof(*msg));
 
 	*sizep = size;
 	return msg;
@@ -255,9 +272,7 @@ static void cluster_op_done(struct work *work)
 	struct vdi_op_message *msg;
 	size_t size;
 
-	cluster_op_running = false;
-
-	sd_dprintf("%s (%p)\n", op_name(req->op), req);
+	sd_dprintf("%s (%p)", op_name(req->op), req);
 
 	msg = prepare_cluster_msg(req, &size);
 	sys->cdrv->unblock(msg, size);
@@ -279,14 +294,14 @@ bool sd_block_handler(const struct sd_node *sender)
 {
 	struct request *req;
 
-	if (!node_eq(sender, &sys->this_node))
+	if (!node_is_local(sender))
 		return false;
 	if (cluster_op_running)
 		return false;
 
 	cluster_op_running = true;
 
-	req = list_first_entry(&sys->pending_block_list,
+	req = list_first_entry(main_thread_get(pending_block_list),
 				struct request, pending_list);
 	req->work.fn = do_process_work;
 	req->work.done = cluster_op_done;
@@ -304,17 +319,19 @@ bool sd_block_handler(const struct sd_node *sender)
  */
 void queue_cluster_request(struct request *req)
 {
-	sd_eprintf("%s (%p)\n", op_name(req->op), req);
+	sd_dprintf("%s (%p)", op_name(req->op), req);
 
 	if (has_process_work(req->op)) {
-		list_add_tail(&req->pending_list, &sys->pending_block_list);
+		list_add_tail(&req->pending_list,
+			      main_thread_get(pending_block_list));
 		sys->cdrv->block();
 	} else {
 		struct vdi_op_message *msg;
 		size_t size;
 
 		msg = prepare_cluster_msg(req, &size);
-		list_add_tail(&req->pending_list, &sys->pending_notify_list);
+		list_add_tail(&req->pending_list,
+			      main_thread_get(pending_notify_list));
 
 		msg->rsp.result = SD_RES_SUCCESS;
 		sys->cdrv->notify(msg, size);
@@ -380,13 +397,13 @@ static bool add_delayed_node(uint32_t epoch, const struct sd_node *node)
 {
 	struct node *n;
 
-	if (find_entry_list(node, &sys->delayed_nodes))
+	if (find_entry_list(node, main_thread_get(delayed_nodes)))
 		return false;
 	assert(!find_entry_epoch(node, epoch));
 
 	n = xmalloc(sizeof(*n));
 	n->ent = *node;
-	list_add_tail(&n->list, &sys->delayed_nodes);
+	list_add_tail(&n->list, main_thread_get(delayed_nodes));
 	return true;
 }
 
@@ -399,14 +416,14 @@ static bool add_failed_node(uint32_t epoch, const struct sd_node *node)
 {
 	struct node *n;
 
-	if (find_entry_list(node, &sys->failed_nodes))
+	if (find_entry_list(node, main_thread_get(failed_nodes)))
 		return false;
 	if (!find_entry_epoch(node, epoch))
 		return false;
 
 	n = xmalloc(sizeof(*n));
 	n->ent = *node;
-	list_add_tail(&n->list, &sys->failed_nodes);
+	list_add_tail(&n->list, main_thread_get(failed_nodes));
 	return true;
 }
 
@@ -425,16 +442,14 @@ static void update_exceptional_node_list(uint32_t epoch,
 		add_delayed_node(epoch, &jm->nodes[i]);
 }
 
-/*
- * Format the lists of failed or delayed nodes into the join message.
- */
+/* Format the lists of failed or delayed nodes into the join message. */
 static void format_exceptional_node_list(struct join_message *jm)
 {
 	struct node *n;
 
-	list_for_each_entry(n, &sys->failed_nodes, list)
+	list_for_each_entry(n, main_thread_get(failed_nodes), list)
 		jm->nodes[jm->nr_failed_nodes++] = n->ent;
-	list_for_each_entry(n, &sys->delayed_nodes, list)
+	list_for_each_entry(n, main_thread_get(delayed_nodes), list)
 		jm->nodes[jm->nr_failed_nodes + jm->nr_delayed_nodes++] = n->ent;
 }
 
@@ -442,10 +457,53 @@ static void clear_exceptional_node_lists(void)
 {
 	struct node *n, *t;
 
-	list_for_each_entry_safe(n, t, &sys->failed_nodes, list)
+	list_for_each_entry_safe(n, t, main_thread_get(failed_nodes), list) {
 		list_del(&n->list);
-	list_for_each_entry_safe(n, t, &sys->delayed_nodes, list)
+		free(n);
+	}
+	list_for_each_entry_safe(n, t, main_thread_get(delayed_nodes), list) {
 		list_del(&n->list);
+		free(n);
+	}
+}
+
+int epoch_log_read_remote(uint32_t epoch, struct sd_node *nodes, int len,
+			  time_t *timestamp, struct vnode_info *vinfo)
+{
+	int i, nr, ret;
+	char buf[SD_MAX_NODES * sizeof(struct sd_node) + sizeof(time_t)];
+
+	nr = vinfo->nr_nodes;
+	for (i = 0; i < nr; i++) {
+		struct sd_req hdr;
+		struct sd_rsp *rsp = (struct sd_rsp *)&hdr;
+		const struct sd_node *node = vinfo->nodes + i;
+		int nodes_len;
+
+		if (node_is_local(node))
+			continue;
+
+		sd_init_req(&hdr, SD_OP_GET_EPOCH);
+		hdr.data_length = len;
+		hdr.obj.tgt_epoch = epoch;
+		hdr.epoch = sys_epoch();
+		ret = sheep_exec_req(&node->nid, &hdr, buf);
+		if (ret != SD_RES_SUCCESS)
+			continue;
+
+		nodes_len = rsp->data_length - sizeof(timestamp);
+		memcpy((void *)nodes, buf, nodes_len);
+		if (timestamp)
+			memcpy(timestamp, buf + nodes_len, sizeof(timestamp));
+
+		return nodes_len / sizeof(struct sd_node);
+	}
+
+	/*
+	 * If no node has targeted epoch log, return 0 here to at least
+	 * allow reading older epoch logs.
+	 */
+	return 0;
 }
 
 static int cluster_sanity_check(struct join_message *jm)
@@ -455,33 +513,31 @@ static int cluster_sanity_check(struct join_message *jm)
 	uint8_t local_nr_copies;
 
 	if (get_cluster_copies(&local_nr_copies)) {
-		sd_eprintf("failed to get nr_copies\n");
+		sd_eprintf("failed to get nr_copies");
 		return CJ_RES_FAIL;
 	}
 
 	if (jm->ctime != local_ctime) {
 		sd_eprintf("joining node ctime doesn't match: %"
-			PRIu64 " vs %" PRIu64 "\n",
-			jm->ctime, local_ctime);
+			   PRIu64 " vs %" PRIu64, jm->ctime, local_ctime);
 		return CJ_RES_FAIL;
 	}
 
 	if (jm->epoch > local_epoch) {
 		sd_eprintf("joining node epoch too large: %"
-			PRIu32 " vs %" PRIu32 "\n",
-			jm->epoch, local_epoch);
+			   PRIu32 " vs %" PRIu32, jm->epoch, local_epoch);
 		return CJ_RES_FAIL;
 	}
 
 	if (jm->nr_copies != local_nr_copies) {
-		sd_eprintf("joining node nr_copies doesn't match: %u vs %u\n",
-			jm->nr_copies, local_nr_copies);
+		sd_eprintf("joining node nr_copies doesn't match: %u vs %u",
+			   jm->nr_copies, local_nr_copies);
 		return CJ_RES_FAIL;
 	}
 
 	if (jm->cluster_flags != sys->flags) {
-		sd_eprintf("joining node cluster_flags don't match: %u vs %u\n",
-			jm->cluster_flags, sys->flags);
+		sd_eprintf("joining node cluster_flags don't match: %u vs %u",
+			   jm->cluster_flags, sys->flags);
 		return CJ_RES_FAIL;
 	}
 
@@ -495,6 +551,7 @@ static int cluster_wait_for_join_check(const struct sd_node *joined,
 	int nr, nr_local_entries, nr_failed_entries, nr_delayed_nodes;
 	uint32_t local_epoch = get_latest_epoch();
 	int ret;
+	struct vnode_info *cur_vinfo;
 
 	if (jm->nr_nodes == 0)
 		return CJ_RES_JOIN_LATER;
@@ -502,8 +559,8 @@ static int cluster_wait_for_join_check(const struct sd_node *joined,
 	ret = cluster_sanity_check(jm);
 	if (ret != CJ_RES_SUCCESS)  {
 		if (jm->epoch > sys->epoch) {
-			sd_eprintf("transfer mastership (%d, %d)\n",
-				jm->epoch, sys->epoch);
+			sd_eprintf("transfer mastership (%d, %d)", jm->epoch,
+				   sys->epoch);
 			return CJ_RES_MASTER_TRANSFER;
 		}
 		return ret;
@@ -516,34 +573,33 @@ static int cluster_wait_for_join_check(const struct sd_node *joined,
 
 	if (jm->epoch < local_epoch) {
 		sd_eprintf("joining node epoch too small: %"
-			PRIu32 " vs %" PRIu32 "\n",
-			jm->epoch, local_epoch);
+			   PRIu32 " vs %" PRIu32, jm->epoch, local_epoch);
 
-		if (bsearch(joined, local_entries, nr_local_entries,
-			    sizeof(struct sd_node), node_id_cmp))
+		if (xbsearch(joined, local_entries, nr_local_entries, node_cmp))
 			return CJ_RES_FAIL;
 		return CJ_RES_JOIN_LATER;
 	}
 
 	if (jm->nr_nodes != nr_local_entries) {
-		sd_eprintf("epoch log entries do not match: %d vs %d\n",
-			jm->nr_nodes, nr_local_entries);
+		sd_eprintf("epoch log entries do not match: %d vs %d",
+			   jm->nr_nodes, nr_local_entries);
 		return CJ_RES_FAIL;
 	}
 
 
 	if (memcmp(jm->nodes, local_entries,
 		   sizeof(jm->nodes[0]) * jm->nr_nodes) != 0) {
-		sd_eprintf("epoch log entries does not match\n");
+		sd_eprintf("epoch log entries does not match");
 		return CJ_RES_FAIL;
 	}
 
-	if (!current_vnode_info)
+	cur_vinfo = main_thread_get(current_vnode_info);
+	if (!cur_vinfo)
 		nr = 1;
 	else
-		nr = current_vnode_info->nr_nodes + 1;
+		nr = cur_vinfo->nr_nodes + 1;
 
-	nr_delayed_nodes = get_nodes_nr_from(&sys->delayed_nodes);
+	nr_delayed_nodes = get_nodes_nr_from(main_thread_get(delayed_nodes));
 
 	/*
 	 * If we have all members from the last epoch log in the in-memory
@@ -559,7 +615,7 @@ static int cluster_wait_for_join_check(const struct sd_node *joined,
 	 * If we reach the old node count, but some node failed we have to
 	 * update the epoch before setting the cluster live.
 	 */
-	nr_failed_entries = get_nodes_nr_from(&sys->failed_nodes);
+	nr_failed_entries = get_nodes_nr_from(main_thread_get(failed_nodes));
 	if (nr_local_entries == nr + nr_failed_entries - nr_delayed_nodes) {
 		jm->inc_epoch = 1;
 		jm->cluster_status = SD_STATUS_OK;
@@ -595,7 +651,7 @@ static int get_vdis_from(struct sd_node *node)
 {
 	struct sd_req hdr;
 	struct sd_rsp *rsp = (struct sd_rsp *)&hdr;
-	struct vdi_copy *vc = NULL;
+	struct vdi_state *vs = NULL;
 	int i, ret = SD_RES_SUCCESS;
 	unsigned int rlen;
 	int count;
@@ -604,21 +660,21 @@ static int get_vdis_from(struct sd_node *node)
 		goto out;
 
 	rlen = SD_DATA_OBJ_SIZE; /* FIXME */
-	vc = xzalloc(rlen);
+	vs = xzalloc(rlen);
 	sd_init_req(&hdr, SD_OP_GET_VDI_COPIES);
 	hdr.data_length = rlen;
 	hdr.epoch = sys_epoch();
-	ret = sheep_exec_req(&node->nid, &hdr, (char *)vc);
+	ret = sheep_exec_req(&node->nid, &hdr, (char *)vs);
 	if (ret != SD_RES_SUCCESS)
 		goto out;
 
-	count = rsp->data_length / sizeof(*vc);
+	count = rsp->data_length / sizeof(*vs);
 	for (i = 0; i < count; i++) {
-		set_bit(vc[i].vid, sys->vdi_inuse);
-		add_vdi_copy_number(vc[i].vid, vc[i].nr_copies);
+		set_bit(vs[i].vid, sys->vdi_inuse);
+		add_vdi_state(vs[i].vid, vs[i].nr_copies, vs[i].snapshot);
 	}
 out:
-	free(vc);
+	free(vs);
 	return ret;
 }
 
@@ -628,9 +684,18 @@ static void do_get_vdis(struct work *work)
 		container_of(work, struct get_vdis_work, work);
 	int i, ret;
 
+	if (!node_is_local(&w->joined)) {
+		switch (sys->status) {
+		case SD_STATUS_OK:
+		case SD_STATUS_HALT:
+			get_vdis_from(&w->joined);
+			return;
+		}
+	}
+
 	for (i = 0; i < w->nr_members; i++) {
 		/* We should not fetch vdi_bitmap and copy list from myself */
-		if (node_eq(&w->members[i], &sys->this_node))
+		if (node_is_local(&w->members[i]))
 			continue;
 
 		ret = get_vdis_from(&w->members[i]);
@@ -662,10 +727,12 @@ static void get_vdis_done(struct work *work)
 
 int log_current_epoch(void)
 {
-	if (!current_vnode_info)
+	struct vnode_info *cur_vinfo = main_thread_get(current_vnode_info);
+
+	if (!cur_vinfo)
 		return update_epoch_log(sys->epoch, NULL, 0);
-	return update_epoch_log(sys->epoch, current_vnode_info->nodes,
-				current_vnode_info->nr_nodes);
+	return update_epoch_log(sys->epoch, cur_vinfo->nodes,
+				cur_vinfo->nr_nodes);
 }
 
 static struct vnode_info *alloc_old_vnode_info(const struct sd_node *joined,
@@ -683,47 +750,73 @@ static struct vnode_info *alloc_old_vnode_info(const struct sd_node *joined,
 	return alloc_vnode_info(old_nodes, count);
 }
 
+static void setup_backend_store(const char *store, bool need_purge)
+{
+	int ret;
+
+	if (!sd_store) {
+		sd_store = find_store_driver(store);
+		if (!sd_store)
+			panic("backend store %s not supported", store);
+
+		ret = sd_store->init();
+		if (ret != SD_RES_SUCCESS)
+			panic("failed to initialize store");
+
+		ret = set_cluster_store(sd_store->name);
+		if (ret != SD_RES_SUCCESS)
+			panic("failed to store into config file");
+	}
+
+	/*
+	 * We need to purge the stale objects for sheep joining back
+	 * after crash
+	 */
+	if (need_purge && sd_store->purge_obj) {
+		ret = sd_store->purge_obj();
+		if (ret != SD_RES_SUCCESS)
+			panic("can't remove stale objects");
+	}
+}
+
 static void finish_join(const struct join_message *msg,
 			const struct sd_node *joined,
 			const struct sd_node *nodes, size_t nr_nodes)
 {
+	int ret;
+
 	sys->join_finished = true;
 	sys->epoch = msg->epoch;
 
 	if (msg->cluster_status != SD_STATUS_OK)
 		update_exceptional_node_list(get_latest_epoch(), msg);
 
-	/* We don't need backend for gateway-only node */
-	if (!sys->gateway_only && !sd_store && strlen((char *)msg->store)) {
-		sd_store = find_store_driver((char *)msg->store);
-		if (sd_store) {
-			if (sd_store->init(obj_path) != SD_RES_SUCCESS)
-				panic("failed to initialize store\n");
-			if (set_cluster_store(sd_store->name) != SD_RES_SUCCESS)
-				panic("failed to store into config file\n");
+	if (msg->store[0]) {
+		/*
+		 * We don't need backend for gateway-only node, but need to save
+		 * store name.  Otherwise, the node cannot notify the store name
+		 * when it become master
+		 */
+		if (sys->gateway_only) {
+			ret = set_cluster_store((char *)msg->store);
+			if (ret != SD_RES_SUCCESS)
+				panic("failed to store into config file");
 		} else
-				panic("backend store %s not supported\n",
-				      msg->store);
+			setup_backend_store((char *)msg->store,
+					    !!msg->inc_epoch);
 	}
 
-	/* We need to purge the stale objects for sheep joining back
-	 * after crash
-	 */
-	if (msg->inc_epoch)
-		if (!sys->gateway_only &&
-		    sd_store->purge_obj &&
-		    sd_store->purge_obj() != SD_RES_SUCCESS)
-			panic("can't remove stale objects\n");
-
 	sockfd_cache_add_group(nodes, nr_nodes);
 }
 
-static void get_vdis(const struct sd_node *nodes, size_t nr_nodes)
+static void get_vdis(const struct sd_node *nodes, size_t nr_nodes,
+		     const struct sd_node *joined)
 {
 	int array_len = nr_nodes * sizeof(struct sd_node);
 	struct get_vdis_work *w;
 
 	w = xmalloc(sizeof(*w) + array_len);
+	w->joined = *joined;
 	w->nr_members = nr_nodes;
 	memcpy(w->members, nodes, array_len);
 
@@ -736,14 +829,14 @@ static void get_vdis(const struct sd_node *nodes, size_t nr_nodes)
 
 void wait_get_vdis_done(void)
 {
-	sd_dprintf("waiting for vdi list\n");
+	sd_dprintf("waiting for vdi list");
 
 	pthread_mutex_lock(&wait_vdis_lock);
 	while (!is_vdi_list_ready)
 		pthread_cond_wait(&wait_vdis_cond, &wait_vdis_lock);
 	pthread_mutex_unlock(&wait_vdis_lock);
 
-	sd_dprintf("vdi list ready\n");
+	sd_dprintf("vdi list ready");
 }
 
 void recalculate_vnodes(struct sd_node *nodes, int nr_nodes)
@@ -767,8 +860,9 @@ void recalculate_vnodes(struct sd_node *nodes, int nr_nodes)
 	for (i = 0; i < nr_nodes; i++) {
 		factor = (float)nodes[i].space / (float)avg_size;
 		nodes[i].nr_vnodes = rintf(SD_DEFAULT_VNODES * factor);
-		sd_dprintf("node %d has %d vnodes, free space %" PRIu64 "\n",
-			nodes[i].nid.port, nodes[i].nr_vnodes, nodes[i].space);
+		sd_dprintf("node %d has %d vnodes, free space %" PRIu64,
+			   nodes[i].nid.port, nodes[i].nr_vnodes,
+			   nodes[i].space);
 	}
 }
 
@@ -779,14 +873,15 @@ static void update_cluster_info(const struct join_message *msg,
 {
 	struct vnode_info *old_vnode_info;
 
-	sd_eprintf("status = %d, epoch = %d, finished: %d\n",
-		msg->cluster_status, msg->epoch, sys->join_finished);
+	sd_eprintf("status = %d, epoch = %d, finished: %d",
+		   msg->cluster_status, msg->epoch, sys->join_finished);
 
 	if (!sys->join_finished)
 		finish_join(msg, joined, nodes, nr_nodes);
 
-	old_vnode_info = current_vnode_info;
-	current_vnode_info = alloc_vnode_info(nodes, nr_nodes);
+	old_vnode_info = main_thread_get(current_vnode_info);
+	main_thread_set(current_vnode_info,
+			  alloc_vnode_info(nodes, nr_nodes));
 
 	switch (msg->cluster_status) {
 	case SD_STATUS_OK:
@@ -802,12 +897,13 @@ static void update_cluster_info(const struct join_message *msg,
 			/*FALLTHROUGH*/
 		case SD_STATUS_WAIT_FOR_JOIN:
 			sys->disable_recovery = msg->disable_recovery;
-			get_vdis(nodes, nr_nodes);
 			break;
 		default:
 			break;
 		}
 
+		get_vdis(nodes, nr_nodes, joined);
+
 		sys->status = msg->cluster_status;
 
 		if (msg->inc_epoch) {
@@ -820,7 +916,8 @@ static void update_cluster_info(const struct join_message *msg,
 						nodes, nr_nodes);
 			}
 
-			start_recovery(current_vnode_info, old_vnode_info);
+			start_recovery(main_thread_get(current_vnode_info),
+				       old_vnode_info, true);
 		}
 
 		if (have_enough_zones())
@@ -850,16 +947,18 @@ void sd_notify_handler(const struct sd_node *sender, void *data,
 	int ret = msg->rsp.result;
 	struct request *req = NULL;
 
-	sd_dprintf("op %s, size: %zd, from: %s\n",
-		op_name(op), data_len, node_to_str(sender));
+	sd_dprintf("op %s, size: %zu, from: %s", op_name(op), data_len,
+		   node_to_str(sender));
 
 	if (node_is_local(sender)) {
 		if (has_process_work(op))
-			req = list_first_entry(&sys->pending_block_list,
-					       struct request, pending_list);
+			req = list_first_entry(
+				main_thread_get(pending_block_list),
+				struct request, pending_list);
 		else
-			req = list_first_entry(&sys->pending_notify_list,
-					       struct request, pending_list);
+			req = list_first_entry(
+				main_thread_get(pending_notify_list),
+				struct request, pending_list);
 		list_del(&req->pending_list);
 	}
 
@@ -875,32 +974,36 @@ void sd_notify_handler(const struct sd_node *sender, void *data,
 
 		put_request(req);
 	}
+
+	if (has_process_work(op))
+		cluster_op_running = false;
 }
 
 enum cluster_join_result sd_check_join_cb(const struct sd_node *joining,
 					  void *opaque)
 {
 	struct join_message *jm = opaque;
-	char str[256];
+	char str[MAX_NODE_STR_LEN];
 	int ret;
 
+	sd_dprintf("check %s, %d", node_to_str(joining), sys->status);
+
 	if (jm->proto_ver != SD_SHEEP_PROTO_VER) {
-		sd_eprintf("%s: invalid protocol version: %d\n", __func__,
-			jm->proto_ver);
+		sd_eprintf("invalid protocol version: %d", jm->proto_ver);
 		return CJ_RES_FAIL;
 	}
 
-	if (node_eq(joining, &sys->this_node)) {
+	if (node_is_local(joining)) {
 		struct sd_node entries[SD_MAX_NODES];
 		int nr_entries;
 		uint32_t epoch;
 
 		/*
-		 * If I'm the first sheep joins in colosync, I
+		 * If I'm the first sheep joins in corosync, I
 		 * becomes the master without sending JOIN.
 		 */
 
-		sd_printf(SDOG_DEBUG, "%s\n", node_to_str(&sys->this_node));
+		sd_printf(SDOG_DEBUG, "%s", node_to_str(&sys->this_node));
 
 		jm->cluster_status = sys->status;
 
@@ -909,8 +1012,8 @@ enum cluster_join_result sd_check_join_cb(const struct sd_node *joining,
 			return CJ_RES_SUCCESS;
 
 		if (sys->status != SD_STATUS_WAIT_FOR_JOIN) {
-			sd_eprintf("unexpected cluster status 0x%x\n",
-				sys->status);
+			sd_eprintf("unexpected cluster status 0x%x",
+				   sys->status);
 			return CJ_RES_FAIL;
 		}
 
@@ -949,13 +1052,13 @@ enum cluster_join_result sd_check_join_cb(const struct sd_node *joining,
 		ret = cluster_running_check(jm);
 		break;
 	default:
-		sd_eprintf("invalid system status: 0x%x\n", sys->status);
-		abort();
+		panic("invalid system status: 0x%x", sys->status);
 	}
 
-	sd_eprintf("%s: ret = 0x%x, cluster_status = 0x%x\n",
-		addr_to_str(str, sizeof(str), joining->nid.addr, joining->nid.port),
-		ret, jm->cluster_status);
+	sd_eprintf("%s: ret = 0x%x, cluster_status = 0x%x",
+		   addr_to_str(str, sizeof(str), joining->nid.addr,
+			       joining->nid.port),
+		   ret, jm->cluster_status);
 
 	jm->nr_copies = sys->nr_copies;
 	jm->cluster_flags = sys->flags;
@@ -995,7 +1098,7 @@ static int send_join_request(struct sd_node *ent)
 
 	ret = sys->cdrv->join(ent, msg, get_join_message_size(msg));
 
-	sd_printf(SDOG_INFO, "%s\n", node_to_str(&sys->this_node));
+	sd_printf(SDOG_INFO, "%s", node_to_str(&sys->this_node));
 
 	free(msg);
 
@@ -1008,13 +1111,13 @@ void sd_join_handler(const struct sd_node *joined,
 		     const void *opaque)
 {
 	int i;
-	int nr, nr_local, nr_failed, nr_delayed_nodes;
+	int nr, nr_local, nr_failed, nr_delayed;
 	const struct join_message *jm = opaque;
 	uint32_t le = get_latest_epoch();
 
-	if (node_eq(joined, &sys->this_node)) {
+	if (node_is_local(joined)) {
 		if (result == CJ_RES_FAIL) {
-			sd_eprintf("Failed to join, exiting.\n");
+			sd_eprintf("Failed to join, exiting.");
 			sys->cdrv->leave();
 			exit(1);
 		}
@@ -1025,18 +1128,18 @@ void sd_join_handler(const struct sd_node *joined,
 		add_delayed_node(le, joined);
 		/*FALLTHRU*/
 	case CJ_RES_SUCCESS:
-		sd_dprintf("join %s\n", node_to_str(joined));
+		sd_dprintf("join %s", node_to_str(joined));
 		for (i = 0; i < nr_members; i++)
-			sd_dprintf("[%x] %s\n", i, node_to_str(members + i));
+			sd_dprintf("[%x] %s", i, node_to_str(members + i));
 
 		if (sys->status == SD_STATUS_SHUTDOWN)
 			break;
 
 		update_cluster_info(jm, joined, members, nr_members);
 
-		if (node_eq(joined, &sys->this_node))
+		if (node_is_local(joined))
 			/* this output is used for testing */
-			sd_printf(SDOG_DEBUG, "join Sheepdog cluster\n");
+			sd_printf(SDOG_DEBUG, "join Sheepdog cluster");
 		break;
 	case CJ_RES_FAIL:
 		if (sys->status != SD_STATUS_WAIT_FOR_JOIN)
@@ -1047,11 +1150,11 @@ void sd_join_handler(const struct sd_node *joined,
 
 		nr_local = get_nodes_nr_epoch(sys->epoch);
 		nr = nr_members;
-		nr_failed = get_nodes_nr_from(&sys->failed_nodes);
-		nr_delayed_nodes = get_nodes_nr_from(&sys->delayed_nodes);
+		nr_failed = get_nodes_nr_from(main_thread_get(failed_nodes));
+		nr_delayed = get_nodes_nr_from(main_thread_get(delayed_nodes));
 
-		sd_dprintf("%d == %d + %d\n", nr_local, nr, nr_failed);
-		if (nr_local == nr + nr_failed - nr_delayed_nodes) {
+		sd_dprintf("%d == %d + %d", nr_local, nr, nr_failed);
+		if (nr_local == nr + nr_failed - nr_delayed) {
 			sys->status = SD_STATUS_OK;
 			log_current_epoch();
 		}
@@ -1059,31 +1162,37 @@ void sd_join_handler(const struct sd_node *joined,
 	case CJ_RES_MASTER_TRANSFER:
 		update_exceptional_node_list(le, jm);
 
-		/* Sheep needs this to identify itself as master.
+		/*
+		 * Sheep needs this to identify itself as master.
 		 * Now mastership transfer is done.
 		 */
 		if (!sys->join_finished) {
 			sys->join_finished = true;
 			sys->epoch = get_latest_epoch();
 
-			put_vnode_info(current_vnode_info);
-			current_vnode_info = alloc_vnode_info(&sys->this_node, 1);
+			put_vnode_info(main_thread_get(current_vnode_info));
+			main_thread_set(current_vnode_info,
+					  alloc_vnode_info(&sys->this_node, 1));
 		}
 
 		nr_local = get_nodes_nr_epoch(sys->epoch);
 		nr = nr_members;
-		nr_failed = get_nodes_nr_from(&sys->failed_nodes);
-		nr_delayed_nodes = get_nodes_nr_from(&sys->delayed_nodes);
+		nr_failed = get_nodes_nr_from(main_thread_get(failed_nodes));
+		nr_delayed = get_nodes_nr_from(main_thread_get(delayed_nodes));
 
-		sd_dprintf("%d == %d + %d\n", nr_local, nr, nr_failed);
-		if (nr_local == nr + nr_failed - nr_delayed_nodes) {
+		sd_dprintf("%d == %d + %d", nr_local, nr, nr_failed);
+		if (nr_local == nr + nr_failed - nr_delayed) {
 			sys->status = SD_STATUS_OK;
 			log_current_epoch();
 		}
 
-		if (node_eq(joined, &sys->this_node))
+		if (node_is_local(joined))
 			/* this output is used for testing */
-			sd_printf(SDOG_DEBUG, "join Sheepdog cluster\n");
+			sd_printf(SDOG_DEBUG, "join Sheepdog cluster");
+		break;
+	default:
+		/* this means sd_check_join_cb() is buggy */
+		panic("unknown cluster join result: %d", result);
 		break;
 	}
 }
@@ -1094,23 +1203,27 @@ void sd_leave_handler(const struct sd_node *left, const struct sd_node *members,
 	struct vnode_info *old_vnode_info;
 	int i;
 
-	sd_dprintf("leave %s\n", node_to_str(left));
+	sd_dprintf("leave %s", node_to_str(left));
 	for (i = 0; i < nr_members; i++)
-		sd_dprintf("[%x] %s\n", i, node_to_str(members + i));
+		sd_dprintf("[%x] %s", i, node_to_str(members + i));
 
 	if (sys->status == SD_STATUS_SHUTDOWN)
 		return;
 
-	old_vnode_info = current_vnode_info;
-	current_vnode_info = alloc_vnode_info(members, nr_members);
+	if (node_is_local(left))
+		/* Mark leave node as gateway only node */
+		sys->this_node.nr_vnodes = 0;
 
+	old_vnode_info = main_thread_get(current_vnode_info);
+	main_thread_set(current_vnode_info,
+			  alloc_vnode_info(members, nr_members));
 	switch (sys->status) {
 	case SD_STATUS_HALT:
 	case SD_STATUS_OK:
 		uatomic_inc(&sys->epoch);
 		log_current_epoch();
-		start_recovery(current_vnode_info, old_vnode_info);
-
+		start_recovery(main_thread_get(current_vnode_info),
+			       old_vnode_info, true);
 		if (!have_enough_zones())
 			sys->status = SD_STATUS_HALT;
 		break;
@@ -1123,6 +1236,27 @@ void sd_leave_handler(const struct sd_node *left, const struct sd_node *members,
 	sockfd_cache_del(&left->nid);
 }
 
+void update_node_size(struct sd_node *node)
+{
+	struct vnode_info *cur_vinfo = main_thread_get(current_vnode_info);
+	int idx = get_node_idx(cur_vinfo, node);
+	assert(idx != -1);
+	cur_vinfo->nodes[idx].space = node->space;
+	sys->cdrv->update_node(node);
+}
+
+void kick_node_recover(void)
+{
+	struct vnode_info *old = main_thread_get(current_vnode_info);
+
+	main_thread_set(current_vnode_info,
+			alloc_vnode_info(old->nodes, old->nr_nodes));
+	uatomic_inc(&sys->epoch);
+	log_current_epoch();
+	start_recovery(main_thread_get(current_vnode_info), old, true);
+	put_vnode_info(old);
+}
+
 int create_cluster(int port, int64_t zone, int nr_vnodes,
 		   bool explicit_addr)
 {
@@ -1131,11 +1265,11 @@ int create_cluster(int port, int64_t zone, int nr_vnodes,
 	if (!sys->cdrv) {
 		sys->cdrv = find_cdrv("corosync");
 		if (sys->cdrv)
-			sd_dprintf("use corosync cluster driver as default\n");
+			sd_dprintf("use corosync cluster driver as default");
 		else {
 			/* corosync cluster driver is not compiled */
 			sys->cdrv = find_cdrv("local");
-			sd_dprintf("use local cluster driver as default\n");
+			sd_dprintf("use local cluster driver as default");
 		}
 	}
 
@@ -1160,29 +1294,33 @@ int create_cluster(int port, int64_t zone, int nr_vnodes,
 		sys->this_node.zone = b[0] | b[1] << 8 | b[2] << 16 | b[3] << 24;
 	} else
 		sys->this_node.zone = zone;
-	sd_dprintf("zone id = %u\n", sys->this_node.zone);
+	sd_dprintf("zone id = %u", sys->this_node.zone);
 
 	sys->this_node.space = sys->disk_space;
 
-	if (get_latest_epoch() > 0) {
+	sys->epoch = get_latest_epoch();
+	if (sys->epoch) {
 		sys->status = SD_STATUS_WAIT_FOR_JOIN;
-
 		get_cluster_copies(&sys->nr_copies);
 		get_cluster_flags(&sys->flags);
 
-		sys->epoch = get_latest_epoch();
 	} else {
 		sys->status = SD_STATUS_WAIT_FOR_FORMAT;
 	}
 
-	INIT_LIST_HEAD(&sys->pending_block_list);
-	INIT_LIST_HEAD(&sys->pending_notify_list);
-	INIT_LIST_HEAD(&sys->failed_nodes);
-	INIT_LIST_HEAD(&sys->delayed_nodes);
+	main_thread_set(pending_block_list,
+			  xzalloc(sizeof(struct list_head)));
+	INIT_LIST_HEAD(main_thread_get(pending_block_list));
+	main_thread_set(pending_notify_list,
+			  xzalloc(sizeof(struct list_head)));
+	INIT_LIST_HEAD(main_thread_get(pending_notify_list));
+	main_thread_set(failed_nodes, xzalloc(sizeof(struct list_head)));
+	INIT_LIST_HEAD(main_thread_get(failed_nodes));
+	main_thread_set(delayed_nodes, xzalloc(sizeof(struct list_head)));
+	INIT_LIST_HEAD(main_thread_get(delayed_nodes));
 
-	INIT_LIST_HEAD(&sys->wait_req_queue);
-	INIT_LIST_HEAD(&sys->wait_rw_queue);
-	INIT_LIST_HEAD(&sys->wait_obj_queue);
+	INIT_LIST_HEAD(&sys->local_req_queue);
+	INIT_LIST_HEAD(&sys->req_wait_queue);
 
 	ret = send_join_request(&sys->this_node);
 	if (ret != 0)
@@ -1191,7 +1329,8 @@ int create_cluster(int port, int64_t zone, int nr_vnodes,
 	return 0;
 }
 
-/* We will call this function for two reason:
+/*
+ * We will call this function for two reason:
  * 1) make this node working as a gateway, or
  * 2) the program is going to shutdown itself.
  */
