@@ -22,6 +22,7 @@
 #include <sys/file.h>
 #include <dirent.h>
 #include <urcu/uatomic.h>
+#include <sys/eventfd.h>
 
 #include "sheep_priv.h"
 #include "util.h"
@@ -48,29 +49,32 @@
 #define CACHE_OBJECT_SIZE (SD_DATA_OBJ_SIZE / 1024 / 1024) /* M */
 
 struct global_cache {
-	uint32_t capacity;
-	uatomic_bool in_reclaim;
+	uint32_t capacity; /* The real capacity of object cache of this node */
+	uatomic_bool in_reclaim; /* If the relcaimer is working */
 };
 
 struct object_cache_entry {
-	uint32_t idx;
-	int refcnt;
-	uint64_t bmap; /* each bit represents one dirty block in object */
-	struct object_cache *oc;
-	struct rb_node node;
-	struct list_head dirty_list;
-	struct list_head lru_list;
+	uint32_t idx; /* Index of this entry */
+	int refcnt; /* Reference count of this entry */
+	uint64_t bmap; /* Each bit represents one dirty block in object */
+	struct object_cache *oc; /* Object cache this entry belongs to */
+	struct rb_node node; /* For lru tree of object cache */
+	struct list_head dirty_list; /* For dirty list of object cache */
+	struct list_head lru_list; /* For lru list of object cache */
+
+	pthread_rwlock_t lock; /* Entry lock */
 };
 
 struct object_cache {
-	uint32_t vid;
-	struct hlist_node hash;
-
-	struct rb_root lru_tree;
+	uint32_t vid; /* The VID of this VDI */
+	uint32_t push_count; /* How many push threads queued in push phase. */
+	struct hlist_node hash; /* VDI is linked to the global hash lists */
+	struct rb_root lru_tree; /* For faster object search */
 	struct list_head lru_head; /* Per VDI LRU list for reclaimer */
-	struct list_head dirty_head;
+	struct list_head dirty_head; /* Dirty objects linked to this list */
+	int push_efd; /* Used to synchronize between pusher and push threads */
 
-	pthread_rwlock_t lock;
+	pthread_rwlock_t lock; /* Cache lock */
 };
 
 static struct global_cache gcache;
@@ -130,6 +134,63 @@ static uint64_t calc_object_bmap(size_t len, off_t offset)
 	return (uint64_t)bmap;
 }
 
+static inline void get_cache_entry(struct object_cache_entry *entry)
+{
+	uatomic_inc(&entry->refcnt);
+}
+
+static inline void put_cache_entry(struct object_cache_entry *entry)
+{
+	uatomic_dec(&entry->refcnt);
+}
+
+static inline bool entry_in_use(struct object_cache_entry *entry)
+{
+	return uatomic_read(&entry->refcnt) > 0;
+}
+
+/*
+ * Mutual exclusive protection strategy:
+ *
+ * reader and writer:          no need to project since it is okay to read
+ *                             unacked stale data.
+ * reader, writer and pusher:    cache lock and entry lock and refcnt.
+ * reader, writer and reclaimer: cache lock and entry refcnt.
+ * pusher and reclaimer:       cache lock and entry refcnt.
+ *
+ * entry->bmap is projected by mostly entry lock, sometimes cache lock.
+ * dirty list is projected by cache lock.
+ */
+static inline void read_lock_cache(struct object_cache *oc)
+{
+	pthread_rwlock_rdlock(&oc->lock);
+}
+
+static inline void write_lock_cache(struct object_cache *oc)
+{
+	pthread_rwlock_wrlock(&oc->lock);
+}
+
+static inline void unlock_cache(struct object_cache *oc)
+{
+	pthread_rwlock_unlock(&oc->lock);
+}
+
+static inline void read_lock_entry(struct object_cache_entry *entry)
+{
+	pthread_rwlock_rdlock(&entry->lock);
+}
+
+static inline void write_lock_entry(struct object_cache_entry *entry)
+{
+	pthread_rwlock_wrlock(&entry->lock);
+}
+
+static inline void unlock_entry(struct object_cache_entry *entry)
+{
+	pthread_rwlock_unlock(&entry->lock);
+}
+
 static struct object_cache_entry *
 lru_tree_insert(struct rb_root *root, struct object_cache_entry *new)
 {
@@ -158,7 +219,7 @@ lru_tree_insert(struct rb_root *root, struct object_cache_entry *new)
 }
 
 static struct object_cache_entry *lru_tree_search(struct rb_root *root,
-						     uint32_t idx)
+						  uint32_t idx)
 {
 	struct rb_node *n = root->rb_node;
 	struct object_cache_entry *t;
@@ -178,13 +239,15 @@ static struct object_cache_entry *lru_tree_search(struct rb_root *root,
 }
 
 static inline void
-free_cache_entry(struct object_cache_entry *entry,
-	      struct rb_root *lru_tree)
+free_cache_entry(struct object_cache_entry *entry)
 {
-	rb_erase(&entry->node, lru_tree);
+	struct object_cache *oc = entry->oc;
+
+	rb_erase(&entry->node, &oc->lru_tree);
 	list_del_init(&entry->lru_list);
 	if (!list_empty(&entry->dirty_list))
 		list_del_init(&entry->dirty_list);
+	pthread_rwlock_destroy(&entry->lock);
 	free(entry);
 }
 
@@ -201,12 +264,11 @@ static int remove_cache_object(struct object_cache *oc, uint32_t idx)
 	int ret = SD_RES_SUCCESS;
 	char path[PATH_MAX];
 
-	sprintf(path, "%s/%06"PRIx32"/%08"PRIx32, object_cache_dir,
-		oc->vid, idx);
-	sd_dprintf("removing cache object %"PRIx64"\n",
-		idx_to_oid(oc->vid, idx));
+	snprintf(path, sizeof(path), "%s/%06"PRIx32"/%08"PRIx32,
+		 object_cache_dir, oc->vid, idx);
+	sd_dprintf("%"PRIx64, idx_to_oid(oc->vid, idx));
 	if (unlink(path) < 0) {
-		sd_eprintf("failed to remove cached object %m\n");
+		sd_eprintf("failed to remove cached object %m");
 		if (errno == ENOENT)
 			return SD_RES_SUCCESS;
 		ret = SD_RES_EIO;
@@ -223,14 +285,17 @@ static int read_cache_object_noupdate(uint32_t vid, uint32_t idx, void *buf,
 	int fd, flags = def_open_flags, ret = SD_RES_SUCCESS;
 	char p[PATH_MAX];
 
-	sprintf(p, "%s/%06"PRIx32"/%08"PRIx32, object_cache_dir, vid, idx);
+	snprintf(p, sizeof(p), "%s/%06"PRIx32"/%08"PRIx32, object_cache_dir,
+		 vid, idx);
 
-	if (sys->object_cache_directio && !idx_has_vdi_bit(idx))
+	if (sys->object_cache_directio && !idx_has_vdi_bit(idx)) {
+		assert(is_aligned_to_pagesize(buf));
 		flags |= O_DIRECT;
+	}
 
-	fd = open(p, flags, def_fmode);
+	fd = open(p, flags, sd_def_fmode);
 	if (fd < 0) {
-		sd_eprintf("%m\n");
+		sd_eprintf("%m");
 		ret = SD_RES_EIO;
 		goto out;
 	}
@@ -238,8 +303,8 @@ static int read_cache_object_noupdate(uint32_t vid, uint32_t idx, void *buf,
 	size = xpread(fd, buf, count, offset);
 
 	if (size != count) {
-		sd_eprintf("size %zu, count:%zu, offset %jd %m\n",
-			size, count, (intmax_t)offset);
+		sd_eprintf("size %zu, count:%zu, offset %jd %m", size, count,
+			   (intmax_t)offset);
 		ret = SD_RES_EIO;
 		goto out_close;
 	}
@@ -257,13 +322,16 @@ static int write_cache_object_noupdate(uint32_t vid, uint32_t idx, void *buf,
 	int fd, flags = def_open_flags, ret = SD_RES_SUCCESS;
 	char p[PATH_MAX];
 
-	sprintf(p, "%s/%06"PRIx32"/%08"PRIx32, object_cache_dir, vid, idx);
-	if (sys->object_cache_directio && !idx_has_vdi_bit(idx))
+	snprintf(p, sizeof(p), "%s/%06"PRIx32"/%08"PRIx32, object_cache_dir,
+		 vid, idx);
+	if (sys->object_cache_directio && !idx_has_vdi_bit(idx)) {
+		assert(is_aligned_to_pagesize(buf));
 		flags |= O_DIRECT;
+	}
 
-	fd = open(p, flags, def_fmode);
+	fd = open(p, flags, sd_def_fmode);
 	if (fd < 0) {
-		sd_eprintf("%m\n");
+		sd_eprintf("%m");
 		ret = SD_RES_EIO;
 		goto out;
 	}
@@ -271,8 +339,8 @@ static int write_cache_object_noupdate(uint32_t vid, uint32_t idx, void *buf,
 	size = xpwrite(fd, buf, count, offset);
 
 	if (size != count) {
-		sd_eprintf("size %zu, count:%zu, offset %jd %m\n",
-			size, count, (intmax_t)offset);
+		sd_eprintf("size %zu, count:%zu, offset %jd %m", size, count,
+			   (intmax_t)offset);
 		ret = SD_RES_EIO;
 		goto out_close;
 	}
@@ -293,9 +361,9 @@ static int read_cache_object(struct object_cache_entry *entry, void *buf,
 	ret = read_cache_object_noupdate(vid, idx, buf, count, offset);
 
 	if (ret == SD_RES_SUCCESS) {
-		pthread_rwlock_wrlock(&oc->lock);
+		write_lock_cache(oc);
 		list_move_tail(&entry->lru_list, &oc->lru_head);
-		pthread_rwlock_unlock(&oc->lock);
+		unlock_cache(oc);
 	}
 	return ret;
 }
@@ -310,19 +378,23 @@ static int write_cache_object(struct object_cache_entry *entry, void *buf,
 	struct sd_req hdr;
 	int ret;
 
-	ret = write_cache_object_noupdate(vid, idx, buf, count, offset);
-	if (ret != SD_RES_SUCCESS)
-		return ret;
+	write_lock_entry(entry);
 
-	pthread_rwlock_wrlock(&oc->lock);
+	ret = write_cache_object_noupdate(vid, idx, buf, count, offset);
+	if (ret != SD_RES_SUCCESS) {
+		unlock_entry(entry);
+		return ret;
+	}
+	write_lock_cache(oc);
 	if (writeback) {
 		entry->bmap |= calc_object_bmap(count, offset);
 		if (list_empty(&entry->dirty_list))
-			list_add_tail(&entry->dirty_list,
-				      &oc->dirty_head);
+			list_add_tail(&entry->dirty_list, &oc->dirty_head);
 	}
 	list_move_tail(&entry->lru_list, &oc->lru_head);
-	pthread_rwlock_unlock(&oc->lock);
+	unlock_cache(oc);
+
+	unlock_entry(entry);
 
 	if (writeback)
 		goto out;
@@ -339,8 +411,8 @@ static int write_cache_object(struct object_cache_entry *entry, void *buf,
 
 	ret = exec_local_req(&hdr, buf);
 	if (ret != SD_RES_SUCCESS) {
-		sd_eprintf("failed to write object %" PRIx64 ", %x\n",
-			oid, ret);
+		sd_eprintf("failed to write object %" PRIx64 ", %s", oid,
+			   sd_strerror(ret));
 		return ret;
 	}
 out:
@@ -358,18 +430,18 @@ static int push_cache_object(uint32_t vid, uint32_t idx, uint64_t bmap,
 	uint64_t oid = idx_to_oid(vid, idx);
 	int first_bit, last_bit;
 
-	sd_dprintf("%"PRIx64", create %d\n", oid, create);
+	sd_dprintf("%"PRIx64", create %d", oid, create);
 
 	if (!bmap) {
-		sd_dprintf("WARN: nothing to flush\n");
+		sd_dprintf("WARN: nothing to flush");
 		return SD_RES_SUCCESS;
 	}
 
 	first_bit = ffsll(bmap) - 1;
 	last_bit = fls64(bmap) - 1;
 
-	sd_dprintf("bmap:0x%"PRIx64", first_bit:%d, last_bit:%d\n",
-		bmap, first_bit, last_bit);
+	sd_dprintf("bmap:0x%"PRIx64", first_bit:%d, last_bit:%d", bmap,
+		   first_bit, last_bit);
 	offset = first_bit * CACHE_BLOCK_SIZE;
 	data_length = (last_bit - first_bit + 1) * CACHE_BLOCK_SIZE;
 
@@ -396,7 +468,7 @@ static int push_cache_object(uint32_t vid, uint32_t idx, uint64_t bmap,
 
 	ret = exec_local_req(&hdr, buf);
 	if (ret != SD_RES_SUCCESS)
-		sd_eprintf("failed to push object %x\n", ret);
+		sd_eprintf("failed to push object %s", sd_strerror(ret));
 out:
 	free(buf);
 	return ret;
@@ -422,34 +494,26 @@ static void do_reclaim_object(struct object_cache *oc)
 	uint64_t oid;
 	uint32_t cap;
 
-	pthread_rwlock_wrlock(&oc->lock);
+	write_lock_cache(oc);
 	list_for_each_entry_safe(entry, t, &oc->lru_head, lru_list) {
 		oid = idx_to_oid(oc->vid, entry_idx(entry));
-		if (uatomic_read(&entry->refcnt) > 0) {
-			sd_dprintf("%"PRIx64" is in operation, skip...\n", oid);
+		if (entry_in_use(entry)) {
+			sd_dprintf("%"PRIx64" is in use, skip...", oid);
 			continue;
 		}
 		if (entry_is_dirty(entry)) {
-			sd_dprintf("%"PRIx64" is dirty, skip...\n", oid);
+			sd_dprintf("%"PRIx64" is dirty, skip...", oid);
 			continue;
 		}
-		if (remove_cache_object(oc, entry_idx(entry))
-		    != SD_RES_SUCCESS)
+		if (remove_cache_object(oc, entry_idx(entry)) != SD_RES_SUCCESS)
 			continue;
-		free_cache_entry(entry, &oc->lru_tree);
+		free_cache_entry(entry);
 		cap = uatomic_sub_return(&gcache.capacity, CACHE_OBJECT_SIZE);
-		sd_dprintf("%"PRIx64" reclaimed. capacity:%"PRId32"\n",
-			oid, cap);
+		sd_dprintf("%"PRIx64" reclaimed. capacity:%"PRId32, oid, cap);
 		if (cap <= HIGH_WATERMARK)
 			break;
 	}
-	pthread_rwlock_unlock(&oc->lock);
-	/*
-	 * Reclaimer grabs a write lock, which will blocks all the IO thread of
-	 * this VDI. We call pthread_yield() to expect that other threads can
-	 * grab the lock more often.
-	 */
-	pthread_yield();
+	unlock_cache(oc);
 }
 
 struct reclaim_work {
@@ -480,14 +544,13 @@ static void do_reclaim(struct work *work)
 			cap = uatomic_read(&gcache.capacity);
 			if (cap <= HIGH_WATERMARK) {
 				pthread_rwlock_unlock(&hashtable_lock[idx]);
-				sd_dprintf("complete, capacity %"PRIu32"\n",
-					cap);
+				sd_dprintf("complete, capacity %"PRIu32, cap);
 				return;
 			}
 		}
 		pthread_rwlock_unlock(&hashtable_lock[idx]);
 	}
-	sd_dprintf("finished\n");
+	sd_dprintf("finished");
 }
 
 static void reclaim_done(struct work *work)
@@ -500,18 +563,13 @@ static void reclaim_done(struct work *work)
 static int create_dir_for(uint32_t vid)
 {
 	int ret = 0;
-	struct strbuf buf = STRBUF_INIT;
+	char p[PATH_MAX];
 
-	strbuf_addstr(&buf, object_cache_dir);
-	strbuf_addf(&buf, "/%06"PRIx32, vid);
-	if (mkdir(buf.buf, def_dmode) < 0)
-		if (errno != EEXIST) {
-			sd_eprintf("%s, %m\n", buf.buf);
-			ret = -1;
-			goto err;
-		}
-err:
-	strbuf_release(&buf);
+	snprintf(p, sizeof(p), "%s/%06"PRIx32, object_cache_dir, vid);
+	if (xmkdir(p, sd_def_dmode) < 0) {
+		sd_eprintf("%s, %m", p);
+		ret = -1;
+	}
 	return ret;
 }
 
@@ -538,8 +596,9 @@ not_found:
 	if (create) {
 		cache = xzalloc(sizeof(*cache));
 		cache->vid = vid;
-		cache->lru_tree = RB_ROOT;
+		INIT_RB_ROOT(&cache->lru_tree);
 		create_dir_for(vid);
+		cache->push_efd = eventfd(0, 0);
 
 		INIT_LIST_HEAD(&cache->dirty_head);
 		INIT_LIST_HEAD(&cache->lru_head);
@@ -561,7 +620,7 @@ void object_cache_try_to_reclaim(int delay)
 	if (!sys->object_cache_size)
 		return;
 
-	if (uatomic_read(&gcache.capacity) < sys->object_cache_size)
+	if (uatomic_read(&gcache.capacity) < HIGH_WATERMARK)
 		return;
 
 	if (!uatomic_set_true(&gcache.in_reclaim))
@@ -572,7 +631,7 @@ void object_cache_try_to_reclaim(int delay)
 	rw->delay = delay;
 	rw->work.fn = do_reclaim;
 	rw->work.done = reclaim_done;
-	queue_work(sys->reclaim_wqueue, &rw->work);
+	queue_work(sys->oc_reclaim_wqueue, &rw->work);
 }
 
 static inline struct object_cache_entry *
@@ -583,6 +642,7 @@ alloc_cache_entry(struct object_cache *oc, uint32_t idx)
 	entry = xzalloc(sizeof(*entry));
 	entry->oc = oc;
 	entry->idx = idx;
+	pthread_rwlock_init(&entry->lock, NULL);
 	INIT_LIST_HEAD(&entry->dirty_list);
 	INIT_LIST_HEAD(&entry->lru_list);
 
@@ -593,18 +653,20 @@ static void add_to_lru_cache(struct object_cache *oc, uint32_t idx, bool create)
 {
 	struct object_cache_entry *entry = alloc_cache_entry(oc, idx);
 
-	sd_dprintf("oid %"PRIx64" added\n", idx_to_oid(oc->vid, idx));
+	sd_dprintf("oid %"PRIx64" added", idx_to_oid(oc->vid, idx));
 
-	pthread_rwlock_wrlock(&oc->lock);
-	assert(!lru_tree_insert(&oc->lru_tree, entry));
+	write_lock_cache(oc);
+	if (lru_tree_insert(&oc->lru_tree, entry))
+		panic("the object already exist");
 	uatomic_add(&gcache.capacity, CACHE_OBJECT_SIZE);
 	list_add_tail(&entry->lru_list, &oc->lru_head);
 	if (create) {
+		/* Cache lock assure it is not raced with pusher */
 		entry->bmap = UINT64_MAX;
 		entry->idx |= CACHE_CREATE_BIT;
-		list_add(&entry->dirty_list, &oc->dirty_head);
+		list_add_tail(&entry->dirty_list, &oc->dirty_head);
 	}
-	pthread_rwlock_unlock(&oc->lock);
+	unlock_cache(oc);
 }
 
 static inline int lookup_path(char *path)
@@ -613,7 +675,7 @@ static inline int lookup_path(char *path)
 
 	if (access(path, R_OK | W_OK) < 0) {
 		if (errno != ENOENT) {
-			sd_dprintf("%m\n");
+			sd_dprintf("%m");
 			ret = SD_RES_EIO;
 		} else {
 			ret = SD_RES_NO_CACHE;
@@ -628,15 +690,15 @@ static int object_cache_lookup(struct object_cache *oc, uint32_t idx,
 	int fd, ret, flags = def_open_flags;
 	char path[PATH_MAX];
 
-	sprintf(path, "%s/%06"PRIx32"/%08"PRIx32, object_cache_dir,
-		oc->vid, idx);
+	snprintf(path, sizeof(path), "%s/%06"PRIx32"/%08"PRIx32,
+		 object_cache_dir, oc->vid, idx);
 	if (!create)
 		return lookup_path(path);
 
 	flags |= O_CREAT | O_TRUNC;
-	fd = open(path, flags, def_fmode);
+	fd = open(path, flags, sd_def_fmode);
 	if (fd < 0) {
-		sd_dprintf("%s, %m\n", path);
+		sd_dprintf("%s, %m", path);
 		ret = SD_RES_EIO;
 		goto out;
 	}
@@ -661,15 +723,15 @@ static int create_cache_object(struct object_cache *oc, uint32_t idx,
 	int ret = SD_RES_OID_EXIST;
 	char path[PATH_MAX], tmp_path[PATH_MAX];
 
-	sprintf(tmp_path, "%s/%06"PRIx32"/%08"PRIx32".tmp", object_cache_dir,
-		oc->vid, idx);
-	fd = open(tmp_path, flags, def_fmode);
+	snprintf(tmp_path, sizeof(tmp_path), "%s/%06"PRIx32"/%08"PRIx32".tmp",
+		object_cache_dir, oc->vid, idx);
+	fd = open(tmp_path, flags, sd_def_fmode);
 	if (fd < 0) {
 		if (errno == EEXIST) {
-			sd_dprintf("%08"PRIx32" already created\n", idx);
+			sd_dprintf("%08"PRIx32" already created", idx);
 			goto out;
 		}
-		sd_dprintf("%m\n");
+		sd_dprintf("%m");
 		ret = SD_RES_EIO;
 		goto out;
 	}
@@ -679,7 +741,7 @@ static int create_cache_object(struct object_cache *oc, uint32_t idx,
 		ret = prealloc(fd, obj_size);
 		if (ret < 0) {
 			ret = SD_RES_EIO;
-			sd_eprintf("%m\n");
+			sd_eprintf("%m");
 			goto out_close;
 		}
 	}
@@ -687,25 +749,25 @@ static int create_cache_object(struct object_cache *oc, uint32_t idx,
 	ret = xpwrite(fd, buffer, buf_size, offset);
 	if (ret != buf_size) {
 		ret = SD_RES_EIO;
-		sd_eprintf("failed, vid %"PRIx32", idx %"PRIx32"\n",
-			oc->vid, idx);
+		sd_eprintf("failed, vid %"PRIx32", idx %"PRIx32, oc->vid, idx);
 		goto out_close;
 	}
 	/* This is intended to take care of partial write due to crash */
-	sprintf(path, "%s/%06"PRIx32"/%08"PRIx32, object_cache_dir,
-		oc->vid, idx);
+	snprintf(path, sizeof(path), "%s/%06"PRIx32"/%08"PRIx32,
+		 object_cache_dir, oc->vid, idx);
 	ret = link(tmp_path, path);
 	if (ret < 0) {
 		if (errno == EEXIST) {
 			ret = SD_RES_OID_EXIST;
 			goto out_close;
 		}
-		sd_dprintf("failed to link %s to %s: %m\n", tmp_path, path);
-		ret = err_to_sderr(idx_to_oid(oc->vid, idx), errno);
+		sd_dprintf("failed to link %s to %s: %m", tmp_path, path);
+		/* FIXME: teach object cache handle EIO gracefully */
+		ret = SD_RES_EIO;
 		goto out_close;
 	}
 	ret = SD_RES_SUCCESS;
-	sd_dprintf("%08"PRIx32" size %zu\n", idx, obj_size);
+	sd_dprintf("%08"PRIx32" size %zu", idx, obj_size);
 out_close:
 	close(fd);
 	unlink(tmp_path);
@@ -732,7 +794,7 @@ static int object_cache_pull(struct object_cache *oc, uint32_t idx)
 	if (ret != SD_RES_SUCCESS)
 		goto err;
 
-	sd_dprintf("oid %"PRIx64" pulled successfully\n", oid);
+	sd_dprintf("oid %"PRIx64" pulled successfully", oid);
 	ret = create_cache_object(oc, idx, buf, rsp->data_length,
 				  rsp->obj.offset, data_length);
 	/*
@@ -758,26 +820,80 @@ err:
 	return ret;
 }
 
-/* Push back all the dirty objects to sheep cluster storage */
+struct push_work {
+	struct work work;
+	struct object_cache_entry *entry;
+};
+
+static void do_push_object(struct work *work)
+{
+	struct push_work *pw = container_of(work, struct push_work, work);
+	struct object_cache_entry *entry = pw->entry;
+	struct object_cache *oc = entry->oc;
+	uint64_t oid = idx_to_oid(oc->vid, entry_idx(entry));
+
+	sd_dprintf("%"PRIx64, oid);
+
+	read_lock_entry(entry);
+	if (push_cache_object(oc->vid, entry_idx(entry), entry->bmap,
+			      !!(entry->idx & CACHE_CREATE_BIT))
+	    != SD_RES_SUCCESS)
+		panic("push failed but should never fail");
+	if (uatomic_sub_return(&oc->push_count, 1) == 0)
+		eventfd_write(oc->push_efd, 1);
+	entry->idx &= ~CACHE_CREATE_BIT;
+	entry->bmap = 0;
+	unlock_entry(entry);
+
+	sd_dprintf("%"PRIx64" done", oid);
+	put_cache_entry(entry);
+}
+
+static void push_object_done(struct work *work)
+{
+	struct push_work *pw = container_of(work, struct push_work, work);
+	free(pw);
+}
+
+/*
+ * Push back all the dirty objects before the FLUSH request to sheep replicated
+ * storage synchronously.
+ *
+ * 1. Don't grab cache lock tight so we can serve RW requests while pushing.
+ *    It is okay for allow subsequent RW after FLUSH because we only need to
+ *    garantee the dirty objects before FLUSH to be pushed.
+ * 2. Use threaded AIO to boost push performance, such as fsync(2) from VM.
+ */
 static int object_cache_push(struct object_cache *oc)
 {
 	struct object_cache_entry *entry, *t;
+	eventfd_t value;
 
-	int ret = SD_RES_SUCCESS;
-
-	pthread_rwlock_wrlock(&oc->lock);
+	write_lock_cache(oc);
+	if (list_empty(&oc->dirty_head)) {
+		unlock_cache(oc);
+		return SD_RES_SUCCESS;
+	}
 	list_for_each_entry_safe(entry, t, &oc->dirty_head, dirty_list) {
-		ret = push_cache_object(oc->vid, entry_idx(entry), entry->bmap,
-					!!(entry->idx & CACHE_CREATE_BIT));
-		if (ret != SD_RES_SUCCESS)
-			goto push_failed;
-		entry->idx &= ~CACHE_CREATE_BIT;
-		entry->bmap = 0;
+		struct push_work *pw;
+
+		get_cache_entry(entry);
+		uatomic_inc(&oc->push_count);
+		pw = xzalloc(sizeof(struct push_work));
+		pw->work.fn = do_push_object;
+		pw->work.done = push_object_done;
+		pw->entry = entry;
+		queue_work(sys->oc_push_wqueue, &pw->work);
 		list_del_init(&entry->dirty_list);
 	}
-push_failed:
-	pthread_rwlock_unlock(&oc->lock);
-	return ret;
+	unlock_cache(oc);
+reread:
+	if (eventfd_read(oc->push_efd, &value) < 0) {
+		sd_eprintf("eventfd read failed, %m");
+		goto reread;
+	}
+	sd_dprintf("%"PRIx32" completed", oc->vid);
+	return SD_RES_SUCCESS;
 }
 
 bool object_is_cached(uint64_t oid)
@@ -809,42 +925,38 @@ void object_cache_delete(uint32_t vid)
 	hlist_del(&cache->hash);
 	pthread_rwlock_unlock(&hashtable_lock[h]);
 
-	pthread_rwlock_wrlock(&cache->lock);
+	write_lock_cache(cache);
 	list_for_each_entry_safe(entry, t, &cache->lru_head, lru_list) {
-		free_cache_entry(entry, &cache->lru_tree);
+		free_cache_entry(entry);
 		uatomic_sub(&gcache.capacity, CACHE_OBJECT_SIZE);
 	}
-	pthread_rwlock_unlock(&cache->lock);
+	unlock_cache(cache);
 	pthread_rwlock_destroy(&cache->lock);
+	close(cache->push_efd);
 	free(cache);
 
 	/* Then we free disk */
-	sprintf(path, "%s/%06"PRIx32, object_cache_dir, vid);
+	snprintf(path, sizeof(path), "%s/%06"PRIx32, object_cache_dir, vid);
 	rmdir_r(path);
 }
 
 static struct object_cache_entry *
-get_cache_entry(struct object_cache *cache, uint32_t idx)
+get_cache_entry_from(struct object_cache *cache, uint32_t idx)
 {
 	struct object_cache_entry *entry;
 
-	pthread_rwlock_rdlock(&cache->lock);
+	read_lock_cache(cache);
 	entry = lru_tree_search(&cache->lru_tree, idx);
 	if (!entry) {
 		/* The cache entry may be reclaimed, so try again. */
-		pthread_rwlock_unlock(&cache->lock);
+		unlock_cache(cache);
 		return NULL;
 	}
-	uatomic_inc(&entry->refcnt);
-	pthread_rwlock_unlock(&cache->lock);
+	get_cache_entry(entry);
+	unlock_cache(cache);
 	return entry;
 }
 
-static void put_cache_entry(struct object_cache_entry *entry)
-{
-	uatomic_dec(&entry->refcnt);
-}
-
 static int object_cache_flush_and_delete(struct object_cache *oc)
 {
 	DIR *dir;
@@ -855,11 +967,11 @@ static int object_cache_flush_and_delete(struct object_cache *oc)
 	int ret = 0;
 	char p[PATH_MAX];
 
-	sd_dprintf("%"PRIx32"\n", vid);
-	sprintf(p, "%s/%06"PRIx32, object_cache_dir, vid);
+	sd_dprintf("%"PRIx32, vid);
+	snprintf(p, sizeof(p), "%s/%06"PRIx32, object_cache_dir, vid);
 	dir = opendir(p);
 	if (!dir) {
-		sd_dprintf("%m\n");
+		sd_dprintf("%m");
 		ret = -1;
 		goto out;
 	}
@@ -868,9 +980,9 @@ static int object_cache_flush_and_delete(struct object_cache *oc)
 		if (!strncmp(d->d_name, ".", 1))
 			continue;
 		if (strcmp(d->d_name + 8, ".tmp") == 0) {
-			sd_dprintf("try to del %s\n", d->d_name);
+			sd_dprintf("try to del %s", d->d_name);
 			if (unlinkat(dirfd(dir), d->d_name, 0) < 0)
-				sd_eprintf("%m\n");
+				sd_eprintf("%m");
 			continue;
 		}
 
@@ -878,9 +990,9 @@ static int object_cache_flush_and_delete(struct object_cache *oc)
 		if (idx == ULLONG_MAX)
 			continue;
 		if (push_cache_object(vid, idx, all, true) !=
-				SD_RES_SUCCESS) {
-			sd_dprintf("failed to push %"PRIx64"\n",
-				idx_to_oid(vid, idx));
+		    SD_RES_SUCCESS) {
+			sd_dprintf("failed to push %"PRIx64,
+				   idx_to_oid(vid, idx));
 			ret = -1;
 			goto out_close_dir;
 		}
@@ -897,6 +1009,9 @@ bool bypass_object_cache(const struct request *req)
 {
 	uint64_t oid = req->rq.obj.oid;
 
+	if (!sys->enable_object_cache || req->local)
+		return true;
+
 	if (req->rq.flags & SD_FLAG_CMD_DIRECT) {
 		uint32_t vid = oid_to_vid(oid);
 		struct object_cache *cache;
@@ -918,9 +1033,7 @@ bool bypass_object_cache(const struct request *req)
 		}
 	}
 
-	/*
-	 * For vmstate && vdi_attr object, we don't do caching
-	 */
+	/* For vmstate && vdi_attr object, we don't do caching */
 	if (is_vmstate_obj(oid) || is_vdi_attr_obj(oid) ||
 	    req->rq.flags & SD_FLAG_CMD_COW)
 		return true;
@@ -938,8 +1051,8 @@ int object_cache_handle_request(struct request *req)
 	int ret;
 	bool create = false;
 
-	sd_dprintf("%08"PRIx32", len %"PRIu32", off %"PRIu64"\n", idx,
-		hdr->data_length, hdr->obj.offset);
+	sd_dprintf("%08"PRIx32", len %"PRIu32", off %"PRIu64, idx,
+		   hdr->data_length, hdr->obj.offset);
 
 	cache = find_object_cache(vid, true);
 
@@ -958,9 +1071,9 @@ retry:
 		return ret;
 	}
 
-	entry = get_cache_entry(cache, idx);
+	entry = get_cache_entry_from(cache, idx);
 	if (!entry) {
-		sd_dprintf("retry oid %"PRIx64"\n", oid);
+		sd_dprintf("retry oid %"PRIx64, oid);
 		/*
 		 * For the case that object exists but isn't added to object
 		 * list yet, we call pthread_yield() to expect other thread can
@@ -989,7 +1102,7 @@ err:
 }
 
 int object_cache_write(uint64_t oid, char *data, unsigned int datalen,
-		       uint64_t offset, uint16_t flags, bool create)
+		       uint64_t offset, bool create)
 {
 	uint32_t vid = oid_to_vid(oid);
 	uint32_t idx = object_cache_oid_to_idx(oid);
@@ -997,11 +1110,11 @@ int object_cache_write(uint64_t oid, char *data, unsigned int datalen,
 	struct object_cache_entry *entry;
 	int ret;
 
-	sd_dprintf("%" PRIx64 "\n", oid);
+	sd_dprintf("%" PRIx64, oid);
 	cache = find_object_cache(vid, false);
-	entry = get_cache_entry(cache, idx);
+	entry = get_cache_entry_from(cache, idx);
 	if (!entry) {
-		sd_dprintf("%" PRIx64 " doesn't exist\n", oid);
+		sd_dprintf("%" PRIx64 " doesn't exist", oid);
 		return SD_RES_NO_CACHE;
 	}
 
@@ -1019,11 +1132,11 @@ int object_cache_read(uint64_t oid, char *data, unsigned int datalen,
 	struct object_cache_entry *entry;
 	int ret;
 
-	sd_dprintf("%" PRIx64 "\n", oid);
+	sd_dprintf("%" PRIx64, oid);
 	cache = find_object_cache(vid, false);
-	entry = get_cache_entry(cache, idx);
+	entry = get_cache_entry_from(cache, idx);
 	if (!entry) {
-		sd_dprintf("%" PRIx64 " doesn't exist\n", oid);
+		sd_dprintf("%" PRIx64 " doesn't exist", oid);
 		return SD_RES_NO_CACHE;
 	}
 
@@ -1038,7 +1151,7 @@ int object_cache_flush_vdi(uint32_t vid)
 
 	cache = find_object_cache(vid, false);
 	if (!cache) {
-		sd_dprintf("%"PRIx32" not found\n", vid);
+		sd_dprintf("%"PRIx32" not found", vid);
 		return SD_RES_SUCCESS;
 	}
 
@@ -1058,27 +1171,6 @@ int object_cache_flush_and_del(const struct request *req)
 	return SD_RES_SUCCESS;
 }
 
-void object_cache_remove(uint64_t oid)
-{
-	uint32_t vid = oid_to_vid(oid);
-	uint32_t idx = object_cache_oid_to_idx(oid);
-	struct object_cache *oc;
-	struct object_cache_entry *entry;
-
-	oc = find_object_cache(vid, false);
-	if (!oc)
-		return;
-
-	pthread_rwlock_wrlock(&oc->lock);
-	entry = lru_tree_search(&oc->lru_tree, idx);
-	if (!entry)
-		goto out;
-	free_cache_entry(entry, &oc->lru_tree);
-out:
-	pthread_rwlock_unlock(&oc->lock);
-	return;
-}
-
 static int load_cache_object(struct object_cache *cache)
 {
 	DIR *dir;
@@ -1087,10 +1179,11 @@ static int load_cache_object(struct object_cache *cache)
 	char path[PATH_MAX];
 	int ret = 0;
 
-	sprintf(path, "%s/%06"PRIx32, object_cache_dir, cache->vid);
+	snprintf(path, sizeof(path), "%s/%06"PRIx32, object_cache_dir,
+		 cache->vid);
 	dir = opendir(path);
 	if (!dir) {
-		sd_dprintf("%m\n");
+		sd_dprintf("%m");
 		ret = -1;
 		goto out;
 	}
@@ -1100,9 +1193,9 @@ static int load_cache_object(struct object_cache *cache)
 			continue;
 
 		if (strcmp(d->d_name + 8, ".tmp") == 0) {
-			sd_dprintf("try to del %s\n", d->d_name);
+			sd_dprintf("try to del %s", d->d_name);
 			if (unlinkat(dirfd(dir), d->d_name, 0) < 0)
-				sd_eprintf("%m\n");
+				sd_eprintf("%m");
 			continue;
 		}
 
@@ -1117,7 +1210,7 @@ static int load_cache_object(struct object_cache *cache)
 		 * cluster isn't fully working.
 		 */
 		add_to_lru_cache(cache, idx, true);
-		sd_dprintf("%"PRIx64"\n", idx_to_oid(cache->vid, idx));
+		sd_dprintf("%"PRIx64, idx_to_oid(cache->vid, idx));
 	}
 
 	closedir(dir);
@@ -1133,10 +1226,10 @@ static int load_cache(void)
 	char path[PATH_MAX];
 	int ret = 0;
 
-	sprintf(path, "%s", object_cache_dir);
+	snprintf(path, sizeof(path), "%s", object_cache_dir);
 	dir = opendir(path);
 	if (!dir) {
-		sd_dprintf("%m\n");
+		sd_dprintf("%m");
 		ret = -1;
 		goto out;
 	}
@@ -1162,25 +1255,21 @@ int object_cache_init(const char *p)
 	struct strbuf buf = STRBUF_INIT;
 
 	strbuf_addstr(&buf, p);
-	if (mkdir(buf.buf, def_dmode) < 0) {
-		if (errno != EEXIST) {
-			sd_eprintf("%s %m\n", buf.buf);
-			ret = -1;
-			goto err;
-		}
+	if (xmkdir(buf.buf, sd_def_dmode) < 0) {
+		sd_eprintf("%s %m", buf.buf);
+		ret = -1;
+		goto err;
 	}
 	strbuf_addstr(&buf, "/cache");
-	if (mkdir(buf.buf, def_dmode) < 0) {
-		if (errno != EEXIST) {
-			sd_eprintf("%s %m\n", buf.buf);
-			ret = -1;
-			goto err;
-		}
+	if (xmkdir(buf.buf, sd_def_dmode) < 0) {
+		sd_eprintf("%s %m", buf.buf);
+		ret = -1;
+		goto err;
 	}
 	strbuf_copyout(&buf, object_cache_dir, sizeof(object_cache_dir));
 
 	uatomic_set(&gcache.capacity, 0);
-	gcache.in_reclaim = false;
+	uatomic_set_false(&gcache.in_reclaim);
 
 	ret = load_cache();
 err:
