@@ -61,68 +61,40 @@ struct sd_op_template {
 	int (*process_main)(const struct sd_req *req, struct sd_rsp *rsp, void *data);
 };
 
-static int stat_sheep(uint64_t *store_size, uint64_t *store_free, uint32_t epoch)
+static int stat_sheep(uint64_t *store_size, uint64_t *store_free,
+		      uint32_t epoch)
 {
-	int ret;
-	DIR *dir;
-	struct dirent *d;
-	uint64_t used = 0;
-	struct stat s;
-	char path[1024];
-	struct strbuf store_dir = STRBUF_INIT;
+	uint64_t used;
 
-	strbuf_addf(&store_dir, "%s", obj_path);
-	dir = opendir(store_dir.buf);
-	if (!dir) {
-		ret = SD_RES_EIO;
-		goto out;
-	}
-
-	while ((d = readdir(dir))) {
-		if (!strcmp(d->d_name, ".") || !strcmp(d->d_name, ".."))
-			continue;
-
-		snprintf(path, sizeof(path), "%s/%s", store_dir.buf, d->d_name);
-
-		ret = stat(path, &s);
-		if (ret)
-			continue;
-
-		used += s.st_size;
-	}
-
-	closedir(dir);
-	ret = SD_RES_SUCCESS;
-
-	*store_size = sys->disk_space;
-	if (sys->gateway_only)
+	if (sys->gateway_only) {
+		*store_size = 0;
 		*store_free = 0;
-	else
-		*store_free = sys->disk_space - used;
-out:
-	strbuf_release(&store_dir);
-	return ret;
+	} else {
+		*store_size = md_get_size(&used);
+		*store_free = *store_size - used;
+	}
+	return SD_RES_SUCCESS;
 }
 
 static int cluster_new_vdi(struct request *req)
 {
 	const struct sd_req *hdr = &req->rq;
 	struct sd_rsp *rsp = &req->rp;
-	uint32_t vid = 0;
-	struct vdi_iocb iocb;
+	uint32_t vid;
 	int ret;
+	struct vdi_iocb iocb = {
+		.name = req->data,
+		.data_len = hdr->data_length,
+		.size = hdr->vdi.vdi_size,
+		.base_vid = hdr->vdi.base_vdi_id,
+		.create_snapshot = !!hdr->vdi.snapid,
+		.nr_copies = hdr->vdi.copies ? hdr->vdi.copies : sys->nr_copies,
+	};
 
-	iocb.name = req->data;
-	iocb.data_len = hdr->data_length;
-	iocb.size = hdr->vdi.vdi_size;
-	iocb.base_vid = hdr->vdi.base_vdi_id;
-	iocb.create_snapshot = !!hdr->vdi.snapid;
-	iocb.nr_copies = hdr->vdi.copies;
+	if (hdr->data_length != SD_MAX_VDI_LEN)
+		return SD_RES_INVALID_PARMS;
 
-	if (!iocb.nr_copies)
-		iocb.nr_copies = sys->nr_copies;
-
-	ret = add_vdi(&iocb, &vid);
+	ret = vdi_create(&iocb, &vid);
 
 	rsp->vdi.vdi_id = vid;
 	rsp->vdi.copies = iocb.nr_copies;
@@ -136,27 +108,39 @@ static int post_cluster_new_vdi(const struct sd_req *req, struct sd_rsp *rsp,
 	unsigned long nr = rsp->vdi.vdi_id;
 	int ret = rsp->result;
 
-	sd_printf(SDOG_INFO, "done %d %ld\n", ret, nr);
+	sd_dprintf("done %d %lx", ret, nr);
 	if (ret == SD_RES_SUCCESS)
 		set_bit(nr, sys->vdi_inuse);
 
 	return ret;
 }
 
+static int vdi_init_tag(const char **tag, const char *buf, uint32_t len)
+{
+	if (len == SD_MAX_VDI_LEN + SD_MAX_VDI_TAG_LEN)
+		*tag = buf + SD_MAX_VDI_LEN;
+	else if (len == SD_MAX_VDI_LEN)
+		*tag = NULL;
+	else
+		return -1;
+
+	return 0;
+}
+
 static int cluster_del_vdi(struct request *req)
 {
 	const struct sd_req *hdr = &req->rq;
-	struct sd_rsp *rsp = &req->rp;
-	uint32_t vid = 0, nr_copies = sys->nr_copies;
-	int ret;
+	uint32_t data_len = hdr->data_length;
+	struct vdi_iocb iocb = {
+		.name = req->data,
+		.data_len = data_len,
+		.snapid = hdr->vdi.snapid,
+	};
 
-	ret = del_vdi(req, req->data, hdr->data_length,
-		      &vid, hdr->vdi.snapid, &nr_copies);
+	if (vdi_init_tag(&iocb.tag, req->data, data_len) < 0)
+		return SD_RES_INVALID_PARMS;
 
-	rsp->vdi.vdi_id = vid;
-	rsp->vdi.copies = nr_copies;
-
-	return ret;
+	return vdi_delete(&iocb, req);
 }
 
 struct cache_deletion_work {
@@ -187,7 +171,7 @@ static int post_cluster_del_vdi(const struct sd_req *req, struct sd_rsp *rsp,
 	struct cache_deletion_work *dw;
 	int ret = rsp->result;
 
-	if (!is_object_cache_enabled())
+	if (!sys->enable_object_cache)
 		return ret;
 
 	dw = xzalloc(sizeof(*dw));
@@ -200,28 +184,36 @@ static int post_cluster_del_vdi(const struct sd_req *req, struct sd_rsp *rsp,
 	return ret;
 }
 
-static int local_get_vdi_info(struct request *req)
+/*
+ * Look up vid and copy number from vdi name
+ *
+ * This must be a cluster operation.  If QEMU reads the vdi object
+ * while sheep snapshots the vdi, sheep can return SD_RES_NO_VDI.  To
+ * avoid this problem, SD_OP_GET_INFO must be ordered with
+ * SD_OP_NEW_VDI.
+ */
+static int cluster_get_vdi_info(struct request *req)
 {
 	const struct sd_req *hdr = &req->rq;
 	struct sd_rsp *rsp = &req->rp;
-	uint32_t vid = 0, nr_copies = sys->nr_copies;
-	void *tag;
+	uint32_t data_len = hdr->data_length;
 	int ret;
+	struct vdi_info info = {};
+	struct vdi_iocb iocb = {
+		.name = req->data,
+		.data_len = data_len,
+		.snapid = hdr->vdi.snapid,
+	};
 
-	if (hdr->data_length == SD_MAX_VDI_LEN + SD_MAX_VDI_TAG_LEN)
-		tag = (char *)req->data + SD_MAX_VDI_LEN;
-	else if (hdr->data_length == SD_MAX_VDI_LEN)
-		tag = NULL;
-	else
+	if (vdi_init_tag(&iocb.tag, req->data, data_len) < 0)
 		return SD_RES_INVALID_PARMS;
 
-	ret = lookup_vdi(req->data, tag, &vid,
-			 hdr->vdi.snapid, &nr_copies, NULL);
+	ret = vdi_lookup(&iocb, &info);
 	if (ret != SD_RES_SUCCESS)
 		return ret;
 
-	rsp->vdi.vdi_id = vid;
-	rsp->vdi.copies = nr_copies;
+	rsp->vdi.vdi_id = info.vid;
+	rsp->vdi.copies = get_vdi_copy_number(info.vid);
 
 	return ret;
 }
@@ -231,27 +223,20 @@ static int remove_epoch(uint32_t epoch)
 	int ret;
 	char path[PATH_MAX];
 
-	sd_dprintf("remove epoch %"PRIu32"\n", epoch);
+	sd_dprintf("remove epoch %"PRIu32, epoch);
 	snprintf(path, sizeof(path), "%s%08u", epoch_path, epoch);
 	ret = unlink(path);
 	if (ret && ret != -ENOENT) {
-		sd_eprintf("failed to remove %s: %s\n", path, strerror(-ret));
+		sd_eprintf("failed to remove %s: %s", path, strerror(-ret));
 		return SD_RES_EIO;
 	}
 
-	snprintf(path, sizeof(path), "%s%08u/", jrnl_path, epoch);
-	ret = rmdir_r(path);
-	if (ret && ret != -ENOENT) {
-		sd_eprintf("failed to remove %s: %s\n", path, strerror(-ret));
-		return SD_RES_EIO;
-	}
-	return 0;
+	return SD_RES_EIO;
 }
 
 static int cluster_make_fs(const struct sd_req *req, struct sd_rsp *rsp,
 			   void *data)
 {
-	const struct sd_so_req *hdr = (const struct sd_so_req *)req;
 	int i, ret;
 	uint32_t latest_epoch;
 	uint64_t created_time;
@@ -271,16 +256,16 @@ static int cluster_make_fs(const struct sd_req *req, struct sd_rsp *rsp,
 	if (set_cluster_store(store_name) < 0)
 		return SD_RES_EIO;
 
-	ret = sd_store->init(obj_path);
+	ret = sd_store->init();
 	if (ret != SD_RES_SUCCESS)
 		return ret;
 
-	sys->nr_copies = hdr->copies;
-	sys->flags = hdr->flags;
+	sys->nr_copies = req->cluster.copies;
+	sys->flags = req->flags;
 	if (!sys->nr_copies)
 		sys->nr_copies = SD_DEFAULT_COPIES;
 
-	created_time = hdr->ctime;
+	created_time = req->cluster.ctime;
 	set_cluster_ctime(created_time);
 	set_cluster_copies(sys->nr_copies);
 	set_cluster_flags(sys->flags);
@@ -289,9 +274,9 @@ static int cluster_make_fs(const struct sd_req *req, struct sd_rsp *rsp,
 		remove_epoch(i);
 
 	memset(sys->vdi_inuse, 0, sizeof(sys->vdi_inuse));
+	clean_vdi_state();
 
 	sys->epoch = 1;
-	sys->recovered_epoch = 1;
 
 	ret = log_current_epoch();
 	if (ret)
@@ -331,30 +316,34 @@ static int cluster_get_vdi_attr(struct request *req)
 {
 	const struct sd_req *hdr = &req->rq;
 	struct sd_rsp *rsp = &req->rp;
-	uint32_t vid = 0, attrid = 0, nr_copies = sys->nr_copies;
-	uint64_t created_time = 0;
-	int ret;
+	uint32_t vid, attrid = 0;
 	struct sheepdog_vdi_attr *vattr;
+	struct vdi_iocb iocb = {};
+	struct vdi_info info = {};
+	int ret;
 
 	vattr = req->data;
-	ret = lookup_vdi(vattr->name, vattr->tag, &vid,
-			 hdr->vdi.snapid, &nr_copies, &created_time);
+	iocb.name = vattr->name;
+	iocb.tag = vattr->tag;
+	iocb.snapid = hdr->vdi.snapid;
+	ret = vdi_lookup(&iocb, &info);
 	if (ret != SD_RES_SUCCESS)
 		return ret;
-
-	/* the current VDI id can change if we take a snapshot,
-	   so we use the hash value of the VDI name as the VDI id */
+	/*
+	 * the current VDI id can change if we take a snapshot,
+	 * so we use the hash value of the VDI name as the VDI id
+	 */
 	vid = fnv_64a_buf(vattr->name, strlen(vattr->name), FNV1A_64_INIT);
 	vid &= SD_NR_VDIS - 1;
 	ret = get_vdi_attr(req->data, hdr->data_length,
-			   vid, &attrid, created_time,
+			   vid, &attrid, info.create_time,
 			   !!(hdr->flags & SD_FLAG_CMD_CREAT),
 			   !!(hdr->flags & SD_FLAG_CMD_EXCL),
 			   !!(hdr->flags & SD_FLAG_CMD_DEL));
 
 	rsp->vdi.vdi_id = vid;
 	rsp->vdi.attr_id = attrid;
-	rsp->vdi.copies = nr_copies;
+	rsp->vdi.copies = get_vdi_copy_number(vid);
 
 	return ret;
 }
@@ -365,7 +354,7 @@ static int local_release_vdi(struct request *req)
 
 	if (!vid) {
 		sd_iprintf("Some VDI failed to release the object cache. "
-			   "Probably you are running old QEMU.\n");
+			   "Probably you are running old QEMU.");
 		return SD_RES_SUCCESS;
 	}
 
@@ -398,17 +387,17 @@ static int local_read_vdis(const struct sd_req *req, struct sd_rsp *rsp,
 static int local_get_vdi_copies(const struct sd_req *req, struct sd_rsp *rsp,
 			   void *data)
 {
-	rsp->data_length = fill_vdi_copy_list(data);
+	rsp->data_length = fill_vdi_state_list(data);
 
 	return SD_RES_SUCCESS;
 }
 
 static int local_stat_sheep(struct request *req)
 {
-	struct sd_node_rsp *node_rsp = (struct sd_node_rsp *)&req->rp;
+	struct sd_rsp *rsp = &req->rp;
 	uint32_t epoch = req->rq.epoch;
 
-	return stat_sheep(&node_rsp->store_size, &node_rsp->store_free, epoch);
+	return stat_sheep(&rsp->node.store_size, &rsp->node.store_free, epoch);
 }
 
 static int local_stat_recovery(const struct sd_req *req, struct sd_rsp *rsp,
@@ -427,9 +416,16 @@ static int local_stat_cluster(struct request *req)
 	int i, max_logs;
 	uint32_t epoch;
 
+	if (req->vinfo == NULL) {
+		sd_dprintf("cluster is not started up");
+		goto out;
+	}
+
 	max_logs = req->rq.data_length / sizeof(*log);
 	epoch = get_latest_epoch();
 	for (i = 0; i < max_logs; i++) {
+		size_t nr_nodes;
+
 		if (epoch <= 0)
 			break;
 
@@ -437,20 +433,24 @@ static int local_stat_cluster(struct request *req)
 		memset(log, 0, sizeof(*log));
 		log->epoch = epoch;
 		log->ctime = get_cluster_ctime();
-		log->nr_nodes = epoch_log_read(epoch, log->nodes,
-					       sizeof(log->nodes));
-		if (log->nr_nodes == -1)
-			log->nr_nodes = epoch_log_read_remote(epoch, log->nodes,
-							      sizeof(log->nodes));
+		nr_nodes = epoch_log_read_with_timestamp(epoch, log->nodes,
+							 sizeof(log->nodes),
+							 (time_t *)&log->time);
+		if (nr_nodes == -1)
+			nr_nodes = epoch_log_read_remote(epoch, log->nodes,
+							 sizeof(log->nodes),
+							 (time_t *)&log->time,
+							 req->vinfo);
+		assert(nr_nodes >= 0);
+		assert(nr_nodes <= SD_MAX_NODES);
+		log->nr_nodes = nr_nodes;
 
-		log->nr_copies = sys->nr_copies;
+		log->disable_recovery = sys->disable_recovery;
 
 		rsp->data_length += sizeof(*log);
-		/* FIXME: this hack would require sizeof(time_t) < sizeof(log->nodes[0]) */
-		log->time = *(uint64_t *)(&log->nodes[log->nr_nodes]);
 		epoch--;
 	}
-
+out:
 	switch (sys->status) {
 	case SD_STATUS_OK:
 		return SD_RES_SUCCESS;
@@ -469,63 +469,104 @@ static int local_stat_cluster(struct request *req)
 
 static int local_get_obj_list(struct request *req)
 {
-	return get_obj_list((const struct sd_list_req *)&req->rq,
-			    (struct sd_list_rsp *)&req->rp, req->data);
+	return get_obj_list(&req->rq, &req->rp, req->data);
 }
 
 static int local_get_epoch(struct request *req)
 {
 	uint32_t epoch = req->rq.obj.tgt_epoch;
-	int nr_nodes;
+	int nr_nodes, nodes_len;
+	time_t timestamp;
 
-	sd_dprintf("%d\n", epoch);
+	sd_dprintf("%d", epoch);
 
-	nr_nodes = epoch_log_read(epoch, req->data, req->rq.data_length);
+	nr_nodes =
+		epoch_log_read_with_timestamp(epoch, req->data,
+					req->rq.data_length - sizeof(timestamp),
+					&timestamp);
 	if (nr_nodes == -1)
 		return SD_RES_NO_TAG;
 
-	req->rp.data_length = nr_nodes * sizeof(struct sd_node) + sizeof(time_t);
+	nodes_len = nr_nodes * sizeof(struct sd_node);
+	memcpy((void *)((char *)req->data + nodes_len), &timestamp,
+		sizeof(timestamp));
+	req->rp.data_length = nodes_len + sizeof(time_t);
 	return SD_RES_SUCCESS;
 }
 
-static int cluster_force_recover(const struct sd_req *req, struct sd_rsp *rsp,
-				void *data)
+static int cluster_force_recover_work(struct request *req)
+{
+	struct vnode_info *old_vnode_info;
+	uint32_t epoch = sys_epoch();
+
+	/*
+	 * We should manually recover the cluster when
+	 * 1) the master is physically down (different epoch condition).
+	 * 2) some nodes are physically down (same epoch condition).
+	 * In both case, the nodes(s) stat is WAIT_FOR_JOIN.
+	 */
+	if (sys->status != SD_STATUS_WAIT_FOR_JOIN || req->vinfo == NULL)
+		return SD_RES_FORCE_RECOVER;
+
+	old_vnode_info = get_vnode_info_epoch(epoch, req->vinfo);
+	if (!old_vnode_info) {
+		sd_printf(SDOG_EMERG, "cannot get vnode info for epoch %d",
+			  epoch);
+		put_vnode_info(old_vnode_info);
+		return SD_RES_FORCE_RECOVER;
+	}
+
+	if (req->rq.data_length <
+	    sizeof(*old_vnode_info->nodes) * old_vnode_info->nr_nodes) {
+		sd_eprintf("too small buffer size, %d", req->rq.data_length);
+		return SD_RES_INVALID_PARMS;
+	}
+
+	req->rp.epoch = epoch;
+	req->rp.data_length = sizeof(*old_vnode_info->nodes) *
+		old_vnode_info->nr_nodes;
+	memcpy(req->data, old_vnode_info->nodes, req->rp.data_length);
+
+	put_vnode_info(old_vnode_info);
+
+	return SD_RES_SUCCESS;
+}
+
+static int cluster_force_recover_main(const struct sd_req *req,
+				      struct sd_rsp *rsp,
+				      void *data)
 {
 	struct vnode_info *old_vnode_info, *vnode_info;
 	int ret = SD_RES_SUCCESS;
 	uint8_t c;
 	uint16_t f;
+	struct sd_node *nodes = data;
+	size_t nr_nodes = rsp->data_length / sizeof(*nodes);
 
-	/* We should manually recover the cluster when
-	 * 1) the master is physically down (different epoch condition).
-	 * 2) some nodes are physically down (same epoch condition).
-	 * In both case, the nodes(s) stat is WAIT_FOR_JOIN.
-	 */
-	if (sys->status != SD_STATUS_WAIT_FOR_JOIN)
+	if (rsp->epoch != sys->epoch) {
+		sd_eprintf("epoch was incremented while cluster_force_recover");
 		return SD_RES_FORCE_RECOVER;
+	}
 
 	ret = get_cluster_copies(&c);
-	if (ret)
-		return ret;
+	if (ret) {
+		sd_printf(SDOG_EMERG, "cannot get cluster copies");
+		goto err;
+	}
 	ret = get_cluster_flags(&f);
-	if (ret)
-		return ret;
+	if (ret) {
+		sd_printf(SDOG_EMERG, "cannot get cluster flags");
+		goto err;
+	}
 
 	sys->nr_copies = c;
 	sys->flags = f;
 
-	old_vnode_info = get_vnode_info_epoch(sys->epoch);
-	if (!old_vnode_info) {
-		sd_eprintf("cannot get vnode info for epoch %d\n", sys->epoch);
-		return SD_RES_EIO;
-	}
-
 	sys->epoch++; /* some nodes are left, so we get a new epoch */
 	ret = log_current_epoch();
 	if (ret) {
-		ret = SD_RES_EIO;
-		sys->epoch--;
-		goto out;
+		sd_printf(SDOG_EMERG, "cannot update epoch log");
+		goto err;
 	}
 
 	if (have_enough_zones())
@@ -534,25 +575,13 @@ static int cluster_force_recover(const struct sd_req *req, struct sd_rsp *rsp,
 		sys->status = SD_STATUS_HALT;
 
 	vnode_info = get_vnode_info();
-	start_recovery(vnode_info, old_vnode_info);
+	old_vnode_info = alloc_vnode_info(nodes, nr_nodes);
+	start_recovery(vnode_info, old_vnode_info, true);
 	put_vnode_info(vnode_info);
-out:
 	put_vnode_info(old_vnode_info);
 	return ret;
-}
-
-static int cluster_snapshot(const struct sd_req *req, struct sd_rsp *rsp,
-			    void *data)
-{
-	int ret;
-	struct siocb iocb = { 0 };
-
-	if (sd_store->snapshot)
-		ret = sd_store->snapshot(&iocb);
-	else
-		ret = SD_RES_NO_SUPPORT;
-
-	return ret;
+err:
+	panic("failed in force recovery");
 }
 
 static int cluster_cleanup(const struct sd_req *req, struct sd_rsp *rsp,
@@ -577,10 +606,16 @@ static int cluster_cleanup(const struct sd_req *req, struct sd_rsp *rsp,
 static int cluster_notify_vdi_add(const struct sd_req *req, struct sd_rsp *rsp,
 				  void *data)
 {
-	uint32_t vid = *(uint32_t *)data;
-	uint32_t nr_copies = *(uint32_t *)((char *)data + sizeof(vid));
+	if (req->vdi_state.old_vid)
+		/* make the previous working vdi a snapshot */
+		add_vdi_state(req->vdi_state.old_vid,
+			      get_vdi_copy_number(req->vdi_state.old_vid),
+			      true);
 
-	add_vdi_copy_number(vid, nr_copies);
+	if (req->vdi_state.set_bitmap)
+		set_bit(req->vdi_state.new_vid, sys->vdi_inuse);
+
+	add_vdi_state(req->vdi_state.new_vid, req->vdi_state.copies, false);
 
 	return SD_RES_SUCCESS;
 }
@@ -598,7 +633,8 @@ static int cluster_delete_cache(const struct sd_req *req, struct sd_rsp *rsp,
 {
 	uint32_t vid = oid_to_vid(req->obj.oid);
 
-	object_cache_delete(vid);
+	if (sys->enable_object_cache)
+		object_cache_delete(vid);
 
 	return SD_RES_SUCCESS;
 }
@@ -620,29 +656,33 @@ static int cluster_recovery_completion(const struct sd_req *req,
 		return SD_RES_SUCCESS;
 
 	if (latest_epoch < epoch) {
-		sd_dprintf("new epoch %d\n", epoch);
+		sd_dprintf("new epoch %d", epoch);
 		latest_epoch = epoch;
 		nr_recovereds = 0;
 	}
 
-	recovereds[nr_recovereds++] = *(struct sd_node *)node;
-	qsort(recovereds, nr_recovereds, sizeof(*recovereds), node_id_cmp);
+	recovereds[nr_recovereds++] = *node;
+	xqsort(recovereds, nr_recovereds, node_cmp);
 
-	sd_dprintf("%s is recovered at epoch %d\n", node_to_str(node), epoch);
+	sd_dprintf("%s is recovered at epoch %d", node_to_str(node), epoch);
 	for (i = 0; i < nr_recovereds; i++)
-		sd_dprintf("[%x] %s\n", i, node_to_str(recovereds + i));
+		sd_dprintf("[%x] %s", i, node_to_str(recovereds + i));
 
 	if (sys->epoch != latest_epoch)
 		return SD_RES_SUCCESS;
 
 	vnode_info = get_vnode_info();
 
-	if (vnode_info->nr_nodes == nr_recovereds &&
-	    memcmp(vnode_info->nodes, recovereds,
-		   sizeof(*recovereds) * nr_recovereds) == 0) {
-		sd_dprintf("all nodes are recovered at epoch %d\n", epoch);
-		if (sd_store->cleanup)
-			sd_store->cleanup();
+	if (vnode_info->nr_nodes == nr_recovereds) {
+		for (i = 0; i < nr_recovereds; ++i) {
+			if (!node_eq(vnode_info->nodes + i, recovereds + i))
+				break;
+		}
+		if (i == nr_recovereds) {
+			sd_dprintf("all nodes are recovered, epoch %d", epoch);
+			if (sd_store->cleanup)
+				sd_store->cleanup();
+		}
 	}
 
 	put_vnode_info(vnode_info);
@@ -650,44 +690,120 @@ static int cluster_recovery_completion(const struct sd_req *req,
 	return SD_RES_SUCCESS;
 }
 
-static int local_set_cache_size(const struct sd_req *req, struct sd_rsp *rsp,
-				  void *data)
+static void do_reweight(struct work *work)
 {
-	uint32_t cache_size = *(uint32_t *)data;
+	struct sd_req hdr;
+	int ret;
 
-	uatomic_set(&sys->object_cache_size, cache_size);
-	sd_dprintf("Max cache size set to %dM\n", cache_size);
+	sd_init_req(&hdr, SD_OP_UPDATE_SIZE);
+	hdr.flags = SD_FLAG_CMD_WRITE;
+	hdr.data_length = sizeof(sys->this_node);
 
-	object_cache_try_to_reclaim(0);
+	ret = exec_local_req(&hdr, &sys->this_node);
+	if (ret != SD_RES_SUCCESS)
+		sd_eprintf("failed to update node size");
+}
+
+static void reweight_done(struct work *work)
+{
+	free(work);
+}
+
+static void reweight_node(void)
+{
+	struct work *rw = xzalloc(sizeof(*rw));
+
+	rw->fn = do_reweight;
+	rw->done = reweight_done;
+
+	queue_work(sys->recovery_wqueue, rw);
+}
+
+static bool node_size_varied(void)
+{
+	uint64_t new, used, old = sys->this_node.space;
+	double diff;
+
+	if (sys->gateway_only)
+		return false;
+
+	new = md_get_size(&used);
+	/* If !old, it is forced-out-gateway. Not supported by current node */
+	if (!old) {
+		if (new)
+			return true;
+		else
+			return false;
+	}
+
+	diff = new > old ? (double)(new - old) : (double)(old - new);
+	sd_dprintf("new %"PRIu64 ", old %"PRIu64", ratio %f", new, old,
+		   diff / (double)old);
+	if (diff / (double)old < 0.01)
+		return false;
+
+	sys->this_node.space = new;
+	set_node_space(new);
+
+	return true;
+}
+
+static int cluster_reweight(const struct sd_req *req, struct sd_rsp *rsp,
+			    void *data)
+{
+	if (node_size_varied())
+		reweight_node();
 
 	return SD_RES_SUCCESS;
 }
 
-static int cluster_restore(const struct sd_req *req, struct sd_rsp *rsp,
-			   void *data)
+static int cluster_update_size(const struct sd_req *req, struct sd_rsp *rsp,
+			       void *data)
 {
-	int ret;
-	struct siocb iocb = { .epoch = req->obj.tgt_epoch };
+	struct sd_node *node = (struct sd_node *)data;
 
-	if (sd_store->restore)
-		ret = sd_store->restore(&iocb);
-	else
-		ret = SD_RES_NO_SUPPORT;
-	return ret;
+	update_node_size(node);
+	kick_node_recover();
+
+	return SD_RES_SUCCESS;
 }
 
-static int local_get_snap_file(struct request *req)
+static int local_md_info(struct request *request)
 {
-	int ret;
-	struct siocb iocb = { .buf = req->data };
+	struct sd_rsp *rsp = &request->rp;
 
-	if (sd_store->get_snap_file) {
-		ret = sd_store->get_snap_file(&iocb);
-		req->rp.data_length = iocb.length;
-	} else
-		ret = SD_RES_NO_SUPPORT;
+	assert(request->rq.data_length == sizeof(struct sd_md_info));
+	rsp->data_length = md_get_info((struct sd_md_info *)request->data);
 
-	return ret;
+	return rsp->data_length ? SD_RES_SUCCESS : SD_RES_UNKNOWN;
+}
+
+static int local_md_plug(const struct sd_req *req, struct sd_rsp *rsp,
+			 void *data)
+{
+	char *disks = (char *)data;
+
+	return md_plug_disks(disks);
+}
+
+static int local_md_unplug(const struct sd_req *req, struct sd_rsp *rsp,
+			   void *data)
+{
+	char *disks = (char *)data;
+
+	return md_unplug_disks(disks);
+}
+
+static int local_get_hash(struct request *request)
+{
+	struct sd_req *req = &request->rq;
+	struct sd_rsp *rsp = &request->rp;
+
+	if (!sd_store->get_hash)
+		return SD_RES_NO_SUPPORT;
+
+	return sd_store->get_hash(req->obj.oid, req->obj.tgt_epoch,
+				  rsp->hash.digest);
 }
 
 /* Return SD_RES_INVALID_PARMS to ask client not to send flush req again */
@@ -695,26 +811,39 @@ static int local_flush_vdi(struct request *req)
 {
 	int ret = SD_RES_INVALID_PARMS;
 
-	if (is_object_cache_enabled()) {
+	if (sys->enable_object_cache) {
 		uint32_t vid = oid_to_vid(req->rq.obj.oid);
 		ret = object_cache_flush_vdi(vid);
 		if (ret != SD_RES_SUCCESS)
 			return ret;
 	}
 
-	if (is_disk_cache_enabled()) {
-		struct sd_req hdr;
-
-		sd_init_req(&hdr, SD_OP_FLUSH_NODES);
-		return exec_local_req(&hdr, NULL);
-	}
-
 	return ret;
 }
 
+static int local_discard_obj(struct request *req)
+{
+	uint64_t oid = req->rq.obj.oid;
+	uint32_t vid = oid_to_vid(oid), zero = 0;
+	int ret, idx = data_oid_to_idx(oid);
+
+	sd_dprintf("%"PRIx64, oid);
+	ret = write_object(vid_to_vdi_oid(vid), (char *)&zero, sizeof(zero),
+			   SD_INODE_HEADER_SIZE + sizeof(vid) * idx, false);
+	if (ret != SD_RES_SUCCESS)
+		return ret;
+	if (remove_object(oid) != SD_RES_SUCCESS)
+		sd_eprintf("failed to remove %"PRIx64, oid);
+	/*
+	 * Return success even if remove_object fails because we have updated
+	 * inode successfully.
+	 */
+	return SD_RES_SUCCESS;
+}
+
 static int local_flush_and_del(struct request *req)
 {
-	if (!is_object_cache_enabled())
+	if (!sys->enable_object_cache)
 		return SD_RES_SUCCESS;
 	return object_cache_flush_and_del(req);
 }
@@ -742,7 +871,7 @@ static int local_trace_read_buf(struct request *request)
 		return SD_RES_AGAIN;
 
 	rsp->data_length = ret;
-	sd_dprintf("%u\n", rsp->data_length);
+	sd_dprintf("%u", rsp->data_length);
 	return SD_RES_SUCCESS;
 }
 
@@ -789,7 +918,6 @@ int peer_remove_obj(struct request *req)
 	uint64_t oid = req->rq.obj.oid;
 
 	objlist_cache_remove(oid);
-	object_cache_remove(oid);
 
 	return sd_store->remove_object(oid);
 }
@@ -807,7 +935,6 @@ int peer_read_obj(struct request *req)
 
 	memset(&iocb, 0, sizeof(iocb));
 	iocb.epoch = epoch;
-	iocb.flags = hdr->flags;
 	iocb.buf = req->data;
 	iocb.length = hdr->data_length;
 	iocb.offset = hdr->obj.offset;
@@ -845,7 +972,6 @@ int peer_write_obj(struct request *req)
 	uint64_t oid = hdr->obj.oid;
 
 	iocb.epoch = hdr->epoch;
-	iocb.flags = hdr->flags;
 	iocb.buf = req->data;
 	iocb.length = hdr->data_length;
 	iocb.offset = hdr->obj.offset;
@@ -865,17 +991,16 @@ int peer_create_and_write_obj(struct request *req)
 
 	memset(&iocb, 0, sizeof(iocb));
 	iocb.epoch = epoch;
-	iocb.flags = hdr->flags;
 	iocb.length = get_objsize(oid);
 	if (hdr->flags & SD_FLAG_CMD_COW) {
-		sd_dprintf("%" PRIx64 ", %" PRIx64 "\n", oid, hdr->obj.cow_oid);
+		sd_dprintf("%" PRIx64 ", %" PRIx64, oid, hdr->obj.cow_oid);
 
 		buf = xvalloc(SD_DATA_OBJ_SIZE);
 		if (hdr->data_length != SD_DATA_OBJ_SIZE) {
 			ret = read_copy_from_replica(req, hdr->epoch,
 						     hdr->obj.cow_oid, buf);
 			if (ret != SD_RES_SUCCESS) {
-				sd_eprintf("failed to read cow object\n");
+				sd_eprintf("failed to read cow object");
 				goto out;
 			}
 		}
@@ -899,14 +1024,6 @@ out:
 	return ret;
 }
 
-int peer_flush(struct request *req)
-{
-	if (sys->gateway_only)
-		return SD_RES_SUCCESS;
-
-	return sd_store->flush();
-}
-
 static struct sd_op_template sd_ops[] = {
 
 	/* cluster operations */
@@ -934,6 +1051,7 @@ static struct sd_op_template sd_ops[] = {
 	[SD_OP_SHUTDOWN] = {
 		.name = "SHUTDOWN",
 		.type = SD_OP_TYPE_CLUSTER,
+		.force = true,
 		.process_main = cluster_shutdown,
 	},
 
@@ -947,21 +1065,8 @@ static struct sd_op_template sd_ops[] = {
 		.name = "FORCE_RECOVER",
 		.type = SD_OP_TYPE_CLUSTER,
 		.force = true,
-		.process_main = cluster_force_recover,
-	},
-
-	[SD_OP_SNAPSHOT] = {
-		.name = "SNAPSHOT",
-		.type = SD_OP_TYPE_CLUSTER,
-		.force = true,
-		.process_main = cluster_snapshot,
-	},
-
-	[SD_OP_RESTORE] = {
-		.name = "RESTORE",
-		.type = SD_OP_TYPE_CLUSTER,
-		.force = true,
-		.process_main = cluster_restore,
+		.process_work = cluster_force_recover_work,
+		.process_main = cluster_force_recover_main,
 	},
 
 	[SD_OP_CLEANUP] = {
@@ -998,19 +1103,43 @@ static struct sd_op_template sd_ops[] = {
 		.process_main = cluster_recovery_completion,
 	},
 
-	/* local operations */
 	[SD_OP_GET_VDI_INFO] = {
 		.name = "GET_VDI_INFO",
-		.type = SD_OP_TYPE_LOCAL,
-		.process_work = local_get_vdi_info,
+		.type = SD_OP_TYPE_CLUSTER,
+		.process_work = cluster_get_vdi_info,
 	},
 
 	[SD_OP_LOCK_VDI] = {
 		.name = "LOCK_VDI",
-		.type = SD_OP_TYPE_LOCAL,
-		.process_work = local_get_vdi_info,
+		.type = SD_OP_TYPE_CLUSTER,
+		.process_work = cluster_get_vdi_info,
 	},
 
+	[SD_OP_REWEIGHT] = {
+		.name = "REWEIGHT",
+		.type = SD_OP_TYPE_CLUSTER,
+		.process_main = cluster_reweight,
+	},
+
+	[SD_OP_UPDATE_SIZE] = {
+		.name = "UPDATE_SIZE",
+		.type = SD_OP_TYPE_CLUSTER,
+		.process_main = cluster_update_size,
+	},
+
+	[SD_OP_ENABLE_RECOVER] = {
+		.name = "ENABLE_RECOVER",
+		.type = SD_OP_TYPE_CLUSTER,
+		.process_main = cluster_enable_recover,
+	},
+
+	[SD_OP_DISABLE_RECOVER] = {
+		.name = "DISABLE_RECOVER",
+		.type = SD_OP_TYPE_CLUSTER,
+		.process_main = cluster_disable_recover,
+	},
+
+	/* local operations */
 	[SD_OP_RELEASE_VDI] = {
 		.name = "RELEASE_VDI",
 		.type = SD_OP_TYPE_LOCAL,
@@ -1076,19 +1205,18 @@ static struct sd_op_template sd_ops[] = {
 		.process_work = local_get_epoch,
 	},
 
-	[SD_OP_GET_SNAP_FILE] = {
-		.name = "GET_SNAP_FILE",
-		.type = SD_OP_TYPE_LOCAL,
-		.force = true,
-		.process_work = local_get_snap_file,
-	},
-
 	[SD_OP_FLUSH_VDI] = {
 		.name = "FLUSH_VDI",
 		.type = SD_OP_TYPE_LOCAL,
 		.process_work = local_flush_vdi,
 	},
 
+	[SD_OP_DISCARD_OBJ] = {
+		.name = "DISCARD_OBJ",
+		.type = SD_OP_TYPE_LOCAL,
+		.process_work = local_discard_obj,
+	},
+
 	[SD_OP_FLUSH_DEL_CACHE] = {
 		.name = "DEL_CACHE",
 		.type = SD_OP_TYPE_LOCAL,
@@ -1116,10 +1244,28 @@ static struct sd_op_template sd_ops[] = {
 		.process_main = local_kill_node,
 	},
 
-	[SD_OP_SET_CACHE_SIZE] = {
-		.name = "SET_CACHE_SIZE",
+	[SD_OP_MD_INFO] = {
+		.name = "MD_INFO",
 		.type = SD_OP_TYPE_LOCAL,
-		.process_main = local_set_cache_size,
+		.process_work = local_md_info,
+	},
+
+	[SD_OP_MD_PLUG] = {
+		.name = "MD_PLUG_DISKS",
+		.type = SD_OP_TYPE_LOCAL,
+		.process_main = local_md_plug,
+	},
+
+	[SD_OP_MD_UNPLUG] = {
+		.name = "MD_UNPLUG_DISKS",
+		.type = SD_OP_TYPE_LOCAL,
+		.process_main = local_md_unplug,
+	},
+
+	[SD_OP_GET_HASH] = {
+		.name = "GET_HASH",
+		.type = SD_OP_TYPE_LOCAL,
+		.process_work = local_get_hash,
 	},
 
 	/* gateway I/O operations */
@@ -1171,29 +1317,6 @@ static struct sd_op_template sd_ops[] = {
 		.type = SD_OP_TYPE_PEER,
 		.process_work = peer_remove_obj,
 	},
-
-	[SD_OP_ENABLE_RECOVER] = {
-		.name = "ENABLE_RECOVER",
-		.type = SD_OP_TYPE_CLUSTER,
-		.process_main = cluster_enable_recover,
-	},
-
-	[SD_OP_DISABLE_RECOVER] = {
-		.name = "DISABLE_RECOVER",
-		.type = SD_OP_TYPE_CLUSTER,
-		.process_main = cluster_disable_recover,
-	},
-
-	[SD_OP_FLUSH_PEER] = {
-		.name = "FLUSH_PEER",
-		.type = SD_OP_TYPE_PEER,
-		.process_work = peer_flush,
-	},
-	[SD_OP_FLUSH_NODES] = {
-		.name = "FLUSH_NODES",
-		.type = SD_OP_TYPE_GATEWAY,
-		.process_work = gateway_flush_nodes,
-	},
 };
 
 const struct sd_op_template *get_sd_op(uint8_t opcode)
@@ -1249,15 +1372,16 @@ void do_process_work(struct work *work)
 	struct request *req = container_of(work, struct request, work);
 	int ret = SD_RES_SUCCESS;
 
-	sd_dprintf("%x, %" PRIx64", %"PRIu32"\n",
-		req->rq.opcode, req->rq.obj.oid, req->rq.epoch);
+	sd_dprintf("%x, %" PRIx64", %"PRIu32, req->rq.opcode, req->rq.obj.oid,
+		   req->rq.epoch);
 
 	if (req->op->process_work)
 		ret = req->op->process_work(req);
 
 	if (ret != SD_RES_SUCCESS) {
-		sd_dprintf("failed: %x, %" PRIx64" , %u, %"PRIx32"\n",
-			req->rq.opcode, req->rq.obj.oid, req->rq.epoch, ret);
+		sd_dprintf("failed: %x, %" PRIx64" , %u, %s",
+			   req->rq.opcode, req->rq.obj.oid, req->rq.epoch,
+			   sd_strerror(ret));
 	}
 
 	req->rp.result = ret;
@@ -1279,7 +1403,6 @@ static int map_table[] = {
 	[SD_OP_READ_OBJ] = SD_OP_READ_PEER,
 	[SD_OP_WRITE_OBJ] = SD_OP_WRITE_PEER,
 	[SD_OP_REMOVE_OBJ] = SD_OP_REMOVE_PEER,
-	[SD_OP_FLUSH_NODES] = SD_OP_FLUSH_PEER,
 };
 
 int gateway_to_peer_opcode(int opcode)
