@@ -25,6 +25,12 @@
 
 static void requeue_request(struct request *req);
 
+static void del_requeue_request(struct request *req)
+{
+	list_del(&req->request_list);
+	requeue_request(req);
+}
+
 static bool is_access_local(struct request *req, uint64_t oid)
 {
 	const struct sd_vnode *obj_vnodes[SD_MAX_COPIES];
@@ -51,21 +57,48 @@ static void io_op_done(struct work *work)
 	case SD_RES_EIO:
 		req->rp.result = SD_RES_NETWORK_ERROR;
 
-		sd_eprintf("leaving sheepdog cluster\n");
+		sd_eprintf("leaving sheepdog cluster");
 		leave_cluster();
 		break;
 	case SD_RES_SUCCESS:
+	case SD_RES_NETWORK_ERROR:
 		break;
 	default:
-		sd_dprintf("unhandled error %d\n", req->rp.result);
+		sd_dprintf("unhandled error %s", sd_strerror(req->rp.result));
 		break;
-
 	}
 
 	put_request(req);
 	return;
 }
 
+/*
+ * There are 4 cases that a request needs to sleep on wait queues for requeue:
+ *
+ *	1. Epoch of request sender is older than system epoch of receiver
+ *	   In this case, we response the sender with SD_RES_OLD_NODE_VER to
+ *	   sender so sender would put the request into its own wait queue to
+ *	   wait its system epoch get lifted and resend the request.
+ *
+ *      2. Epoch of request sender is newer than system epoch of receiver
+ *         In this case, we put the request into wait queue of receiver, to wait
+ *         system epoch of receiver to get lifted, then retry this request on
+ *         its own.
+ *
+ *      3. Object requested doesn't exist and recovery work is at RW_INIT state
+ *         In this case, we check whether the requested object exists, if so,
+ *         go process the request directly, if not put the request into wait
+ *         queue of the receiver to wait for the finish of this oid recovery.
+ *
+ *      4. Object requested doesn't exist and is being recoverred
+ *         In this case, we put the request into wait queue of receiver and when
+ *         we recover an object we try to wake up the request on this oid.
+ */
+static inline void sleep_on_wait_queue(struct request *req)
+{
+	list_add_tail(&req->request_list, &sys->req_wait_queue);
+}
+
 static void gateway_op_done(struct work *work)
 {
 	struct request *req = container_of(work, struct request, work);
@@ -74,12 +107,11 @@ static void gateway_op_done(struct work *work)
 	switch (req->rp.result) {
 	case SD_RES_OLD_NODE_VER:
 		if (req->rp.epoch > sys->epoch) {
-			list_add_tail(&req->request_list,
-				      &sys->wait_rw_queue);
 			/*
 			 * Gateway of this node is expected to process this
 			 * request later when epoch is lifted.
 			 */
+			sleep_on_wait_queue(req);
 			return;
 		}
 		/*FALLTHRU*/
@@ -88,16 +120,13 @@ static void gateway_op_done(struct work *work)
 	case SD_RES_WAIT_FOR_JOIN:
 	case SD_RES_WAIT_FOR_FORMAT:
 	case SD_RES_KILLED:
-		sd_dprintf("retrying failed I/O request "
-			"op %s result %x epoch %"PRIu32", sys epoch %"PRIu32"\n",
-			op_name(req->op),
-			req->rp.result,
-			req->rq.epoch,
-			sys->epoch);
+		sd_dprintf("retrying failed I/O request op %s result %x epoch %"
+			   PRIu32", sys epoch %"PRIu32, op_name(req->op),
+			   req->rp.result, req->rq.epoch, sys->epoch);
 		goto retry;
 	case SD_RES_EIO:
 		if (is_access_local(req, hdr->obj.oid)) {
-			sd_eprintf("leaving sheepdog cluster\n");
+			sd_eprintf("leaving sheepdog cluster");
 			leave_cluster();
 			goto retry;
 		}
@@ -105,7 +134,7 @@ static void gateway_op_done(struct work *work)
 	case SD_RES_SUCCESS:
 		break;
 	default:
-		sd_dprintf("unhandled error %d\n", req->rp.result);
+		sd_dprintf("unhandled error %s", sd_strerror(req->rp.result));
 		break;
 	}
 
@@ -130,21 +159,19 @@ static void local_op_done(struct work *work)
 static int check_request_epoch(struct request *req)
 {
 	if (before(req->rq.epoch, sys->epoch)) {
-		sd_eprintf("old node version %u, %u (%s)\n",
-			sys->epoch, req->rq.epoch, op_name(req->op));
-		/* ask gateway to retry. */
+		sd_eprintf("old node version %u, %u (%s)",
+			   sys->epoch, req->rq.epoch, op_name(req->op));
+		/* Ask for sleeping req on requester's wait queue */
 		req->rp.result = SD_RES_OLD_NODE_VER;
 		req->rp.epoch = sys->epoch;
 		put_request(req);
 		return -1;
 	} else if (after(req->rq.epoch, sys->epoch)) {
-		sd_eprintf("new node version %u, %u (%s)\n",
-			sys->epoch, req->rq.epoch, op_name(req->op));
-
-		/* put on local wait queue, waiting for local epoch
-		   to be lifted */
+		sd_eprintf("new node version %u, %u (%s)",
+			   sys->epoch, req->rq.epoch, op_name(req->op));
+		/* Wait for local epoch to be lifted */
 		req->rp.result = SD_RES_NEW_NODE_VER;
-		list_add_tail(&req->request_list, &sys->wait_rw_queue);
+		sleep_on_wait_queue(req);
 		return -1;
 	}
 
@@ -170,30 +197,20 @@ static bool request_in_recovery(struct request *req)
 	 */
 	if (oid_in_recovery(req->local_oid) &&
 	    !(req->rq.flags & SD_FLAG_CMD_RECOVERY)) {
-		/*
-		 * Put request on wait queues of local node
-		 */
-		if (is_recovery_init()) {
-			sd_dprintf("%"PRIx64" on rw_queue\n", req->local_oid);
-			req->rp.result = SD_RES_OBJ_RECOVERING;
-			list_add_tail(&req->request_list,
-				      &sys->wait_rw_queue);
-		} else {
-			sd_dprintf("%"PRIx64" on obj_queue\n", req->local_oid);
-			list_add_tail(&req->request_list,
-				      &sys->wait_obj_queue);
-		}
+		sd_dprintf("%"PRIx64" wait on oid", req->local_oid);
+		sleep_on_wait_queue(req);
 		return true;
 	}
 	return false;
 }
 
-void resume_wait_epoch_requests(void)
+/* Wakeup requests because of epoch mismatch */
+void wakeup_requests_on_epoch(void)
 {
 	struct request *req, *t;
 	LIST_HEAD(pending_list);
 
-	list_splice_init(&sys->wait_rw_queue, &pending_list);
+	list_splice_init(&sys->req_wait_queue, &pending_list);
 
 	list_for_each_entry_safe(req, t, &pending_list, request_list) {
 		switch (req->rp.result) {
@@ -203,73 +220,54 @@ void resume_wait_epoch_requests(void)
 			 * its epoch changes.
 			 */
 			assert(is_gateway_op(req->op));
+			sd_dprintf("gateway %"PRIx64, req->rq.obj.oid);
 			req->rq.epoch = sys->epoch;
-			list_del(&req->request_list);
-			requeue_request(req);
+			del_requeue_request(req);
 			break;
 		case SD_RES_NEW_NODE_VER:
-			/* Peer retries the request locally when its epoch changes. */
+			/*
+			 * Peer retries the request locally when its epoch
+			 * changes.
+			 */
 			assert(!is_gateway_op(req->op));
-			list_del(&req->request_list);
-			requeue_request(req);
+			sd_dprintf("peer %"PRIx64, req->rq.obj.oid);
+			del_requeue_request(req);
 			break;
 		default:
 			break;
 		}
 	}
 
-	list_splice_init(&pending_list, &sys->wait_rw_queue);
+	list_splice_init(&pending_list, &sys->req_wait_queue);
 }
 
-void resume_wait_recovery_requests(void)
+/* Wakeup the requests on the oid that was previously being recoverred */
+void wakeup_requests_on_oid(uint64_t oid)
 {
 	struct request *req, *t;
 	LIST_HEAD(pending_list);
 
-	list_splice_init(&sys->wait_rw_queue, &pending_list);
-
-	list_for_each_entry_safe(req, t, &pending_list, request_list) {
-		if (req->rp.result != SD_RES_OBJ_RECOVERING)
-			continue;
-
-		sd_dprintf("resume wait oid %" PRIx64 "\n", req->local_oid);
-		list_del(&req->request_list);
-		requeue_request(req);
-	}
-
-	list_splice_init(&pending_list, &sys->wait_rw_queue);
-}
-
-void resume_wait_obj_requests(uint64_t oid)
-{
-	struct request *req, *t;
-	LIST_HEAD(pending_list);
-
-	list_splice_init(&sys->wait_obj_queue, &pending_list);
+	list_splice_init(&sys->req_wait_queue, &pending_list);
 
 	list_for_each_entry_safe(req, t, &pending_list, request_list) {
 		if (req->local_oid != oid)
 			continue;
-
-		/* the object requested by a pending request has been
-		 * recovered, notify the pending request. */
-		sd_dprintf("retry %" PRIx64 "\n", req->local_oid);
-		list_del(&req->request_list);
-		requeue_request(req);
+		sd_dprintf("retry %" PRIx64, req->local_oid);
+		del_requeue_request(req);
 	}
-	list_splice_init(&pending_list, &sys->wait_obj_queue);
+	list_splice_init(&pending_list, &sys->req_wait_queue);
 }
 
-void flush_wait_obj_requests(void)
+void wakeup_all_requests(void)
 {
 	struct request *req, *n;
 	LIST_HEAD(pending_list);
 
-	list_splice_init(&sys->wait_obj_queue, &pending_list);
+	list_splice_init(&sys->req_wait_queue, &pending_list);
 
 	list_for_each_entry_safe(req, n, &pending_list, request_list) {
-		list_del(&req->request_list);
-		requeue_request(req);
+		sd_dprintf("%"PRIx64, req->rq.obj.oid);
+		del_requeue_request(req);
 	}
 }
 
@@ -303,7 +301,7 @@ static void queue_gateway_request(struct request *req)
 	 * Even if it doesn't exist in cache, we'll rely on cache layer to pull
 	 * it.
 	 */
-	if (is_object_cache_enabled())
+	if (sys->enable_object_cache)
 		goto queue_work;
 
 	if (req->local_oid)
@@ -347,12 +345,12 @@ static void queue_request(struct request *req)
 
 	req->op = get_sd_op(hdr->opcode);
 	if (!req->op) {
-		sd_eprintf("invalid opcode %d\n", hdr->opcode);
+		sd_eprintf("invalid opcode %d", hdr->opcode);
 		rsp->result = SD_RES_INVALID_PARMS;
 		goto done;
 	}
 
-	sd_dprintf("%s, %d\n", op_name(req->op), sys->status);
+	sd_dprintf("%s, %d", op_name(req->op), sys->status);
 
 	switch (sys->status) {
 	case SD_STATUS_KILLED:
@@ -383,13 +381,7 @@ static void queue_request(struct request *req)
 		break;
 	}
 
-	/*
-	 * force operations shouldn't access req->vnodes in their
-	 * process_work() and process_main() because they can be
-	 * called before we set up current_vnode_info
-	 */
-	if (!is_force_op(req->op))
-		req->vinfo = get_vnode_info();
+	req->vinfo = get_vnode_info();
 
 	if (is_peer_op(req->op)) {
 		queue_peer_request(req);
@@ -403,7 +395,7 @@ static void queue_request(struct request *req)
 		hdr->epoch = sys->epoch;
 		queue_cluster_request(req);
 	} else {
-		sd_eprintf("unknown operation %d\n", hdr->opcode);
+		sd_eprintf("unknown operation %d", hdr->opcode);
 		rsp->result = SD_RES_SYSTEM_ERROR;
 		goto done;
 	}
@@ -459,21 +451,23 @@ int exec_local_req(struct sd_req *rq, void *data)
 	eventfd_t value = 1;
 	int ret;
 
+	assert(is_worker_thread());
+
 	req = alloc_local_request(data, rq->data_length);
 	req->rq = *rq;
-	req->wait_efd = eventfd(0, 0);
+	req->local_req_efd = eventfd(0, 0);
 
-	pthread_mutex_lock(&sys->wait_req_lock);
-	list_add_tail(&req->request_list, &sys->wait_req_queue);
-	pthread_mutex_unlock(&sys->wait_req_lock);
+	pthread_mutex_lock(&sys->local_req_lock);
+	list_add_tail(&req->request_list, &sys->local_req_queue);
+	pthread_mutex_unlock(&sys->local_req_lock);
 
-	eventfd_write(sys->req_efd, value);
+	eventfd_write(sys->local_req_efd, value);
 
 again:
 	/* In error case (for e.g, EINTR) just retry read */
-	ret = eventfd_read(req->wait_efd, &value);
+	ret = eventfd_read(req->local_req_efd, &value);
 	if (ret < 0) {
-		sd_eprintf("%m\n");
+		sd_eprintf("%m");
 		if (errno == EINTR)
 			goto again;
 		/* Fake the result to ask for retry */
@@ -483,7 +477,7 @@ again:
 	/* fill rq with response header as exec_req does */
 	memcpy(rq, &req->rp, sizeof(req->rp));
 
-	close(req->wait_efd);
+	close(req->local_req_efd);
 	ret = req->rp.result;
 	free_local_request(req);
 
@@ -536,7 +530,7 @@ void put_request(struct request *req)
 		return;
 
 	if (req->local)
-		eventfd_write(req->wait_efd, value);
+		eventfd_write(req->local_req_efd, value);
 	else {
 		if (conn_tx_on(&ci->conn)) {
 			clear_client_info(ci);
@@ -593,7 +587,7 @@ static inline int begin_rx(struct client_info *ci)
 		ret = rx(conn, C_IO_END);
 		break;
 	default:
-		sd_eprintf("bug: unknown state %d\n", conn->c_rx_state);
+		sd_eprintf("bug: unknown state %d", conn->c_rx_state);
 	}
 
 	if (is_conn_dead(conn)) {
@@ -615,7 +609,7 @@ static inline void finish_rx(struct client_info *ci)
 	req = ci->rx_req;
 	init_rx_hdr(ci);
 
-	sd_dprintf("%d, %s:%d\n", ci->conn.fd, ci->conn.ipstr, ci->conn.port);
+	sd_dprintf("%d, %s:%d", ci->conn.fd, ci->conn.ipstr, ci->conn.port);
 	queue_request(req);
 }
 
@@ -701,8 +695,8 @@ static inline int finish_tx(struct client_info *ci)
 {
 	/* Finish sending one response */
 	if (ci->conn.c_tx_state == C_IO_END) {
-		sd_dprintf("connection from: %d, %s:%d\n", ci->conn.fd,
-			ci->conn.ipstr, ci->conn.port);
+		sd_dprintf("connection from: %d, %s:%d", ci->conn.fd,
+			   ci->conn.ipstr, ci->conn.port);
 		free_request(ci->tx_req);
 		ci->tx_req = NULL;
 	}
@@ -732,7 +726,7 @@ static void do_client_tx(struct client_info *ci)
 
 static void destroy_client(struct client_info *ci)
 {
-	sd_dprintf("connection from: %s:%d\n", ci->conn.ipstr, ci->conn.port);
+	sd_dprintf("connection from: %s:%d", ci->conn.ipstr, ci->conn.port);
 	close(ci->conn.fd);
 	free(ci);
 }
@@ -741,7 +735,7 @@ static void clear_client_info(struct client_info *ci)
 {
 	struct request *req, *t;
 
-	sd_dprintf("connection seems to be dead\n");
+	sd_dprintf("connection seems to be dead");
 
 	if (ci->rx_req) {
 		free_request(ci->rx_req);
@@ -760,9 +754,8 @@ static void clear_client_info(struct client_info *ci)
 
 	unregister_event(ci->conn.fd);
 
-	sd_dprintf("refcnt:%d, fd:%d, %s:%d\n",
-		ci->refcnt, ci->conn.fd,
-		ci->conn.ipstr, ci->conn.port);
+	sd_dprintf("refcnt:%d, fd:%d, %s:%d", ci->refcnt, ci->conn.fd,
+		   ci->conn.ipstr, ci->conn.port);
 
 	if (ci->refcnt)
 		return;
@@ -811,8 +804,8 @@ static void client_handler(int fd, int events, void *data)
 {
 	struct client_info *ci = (struct client_info *)data;
 
-	sd_dprintf("%x, rx %d, tx %d\n", events, ci->conn.c_rx_state,
-		ci->conn.c_tx_state);
+	sd_dprintf("%x, rx %d, tx %d", events, ci->conn.c_rx_state,
+		   ci->conn.c_tx_state);
 
 	if (events & (EPOLLERR | EPOLLHUP) || is_conn_dead(&ci->conn))
 		return clear_client_info(ci);
@@ -833,7 +826,7 @@ static void listen_handler(int listen_fd, int events, void *data)
 	bool is_inet_socket = *(bool *)data;
 
 	if (sys->status == SD_STATUS_SHUTDOWN) {
-		sd_dprintf("unregistering connection %d\n", listen_fd);
+		sd_dprintf("unregistering connection %d", listen_fd);
 		unregister_event(listen_fd);
 		return;
 	}
@@ -841,7 +834,7 @@ static void listen_handler(int listen_fd, int events, void *data)
 	namesize = sizeof(from);
 	fd = accept(listen_fd, (struct sockaddr *)&from, &namesize);
 	if (fd < 0) {
-		sd_eprintf("failed to accept a new connection: %m\n");
+		sd_eprintf("failed to accept a new connection: %m");
 		return;
 	}
 
@@ -871,7 +864,7 @@ static void listen_handler(int listen_fd, int events, void *data)
 		return;
 	}
 
-	sd_dprintf("accepted a new connection: %d\n", fd);
+	sd_dprintf("accepted a new connection: %d", fd);
 }
 
 static int create_listen_port_fn(int fd, void *data)
@@ -889,7 +882,7 @@ int create_listen_port(char *bindaddr, int port)
 
 int init_unix_domain_socket(const char *dir)
 {
-	static bool is_inet_socket = false;
+	static bool is_inet_socket;
 	char unix_path[PATH_MAX];
 
 	snprintf(unix_path, sizeof(unix_path), "%s/sock", dir);
@@ -899,7 +892,7 @@ int init_unix_domain_socket(const char *dir)
 					 &is_inet_socket);
 }
 
-static void req_handler(int listen_fd, int events, void *data)
+static void local_req_handler(int listen_fd, int events, void *data)
 {
 	eventfd_t value;
 	struct request *req, *t;
@@ -907,15 +900,15 @@ static void req_handler(int listen_fd, int events, void *data)
 	int ret;
 
 	if (events & EPOLLERR)
-		sd_eprintf("request handler error\n");
+		sd_eprintf("request handler error");
 
 	ret = eventfd_read(listen_fd, &value);
 	if (ret < 0)
 		return;
 
-	pthread_mutex_lock(&sys->wait_req_lock);
-	list_splice_init(&sys->wait_req_queue, &pending_list);
-	pthread_mutex_unlock(&sys->wait_req_lock);
+	pthread_mutex_lock(&sys->local_req_lock);
+	list_splice_init(&sys->local_req_queue, &pending_list);
+	pthread_mutex_unlock(&sys->local_req_lock);
 
 	list_for_each_entry_safe(req, t, &pending_list, request_list) {
 		list_del(&req->request_list);
@@ -925,7 +918,9 @@ static void req_handler(int listen_fd, int events, void *data)
 
 void local_req_init(void)
 {
-	pthread_mutex_init(&sys->wait_req_lock, NULL);
-	sys->req_efd = eventfd(0, EFD_NONBLOCK);
-	register_event(sys->req_efd, req_handler, NULL);
+	pthread_mutex_init(&sys->local_req_lock, NULL);
+	sys->local_req_efd = eventfd(0, EFD_NONBLOCK);
+	if (sys->local_req_efd < 0)
+		panic("failed to init local req efd");
+	register_event(sys->local_req_efd, local_req_handler, NULL);
 }
