@@ -46,18 +46,19 @@ static struct sd_option sheep_options[] = {
 	{'d', "debug", false, "include debug messages in the log"},
 	{'D', "directio", false, "use direct IO for backend store"},
 	{'f', "foreground", false, "make the program run in the foreground"},
+	{'F', "log-format", true, "specify log format"},
 	{'g', "gateway", false, "make the progam run as a gateway mode"},
 	{'h', "help", false, "display this help and exit"},
 	{'i', "ioaddr", true, "use separate network card to handle IO requests"},
 	{'j', "journal", true, "use jouranl file to log all the write operations"},
 	{'l', "loglevel", true, "specify the level of logging detail"},
+	{'n', "nosync", false, "drop O_SYNC for write of backend"},
 	{'o', "stdout", false, "log to stdout instead of shared logger"},
 	{'p', "port", true, "specify the TCP port on which to listen"},
 	{'P', "pidfile", true, "create a pid file"},
-	{'s', "disk-space", true, "specify the free disk space in megabytes"},
 	{'u', "upgrade", false, "upgrade to the latest data layout"},
 	{'v', "version", false, "show the version"},
-	{'w', "write-cache", true, "specify the cache type"},
+	{'w', "enable-cache", true, "enable object cache"},
 	{'y', "myaddr", true, "specify the address advertised to other sheep"},
 	{'z', "zone", true, "specify the zone id"},
 	{ 0, NULL, false, NULL },
@@ -128,17 +129,17 @@ static int sigfd;
 static void signal_handler(int listen_fd, int events, void *data)
 {
 	struct signalfd_siginfo siginfo;
-	int ret;
+	int uninitialized_var(ret);
 
 	ret = read(sigfd, &siginfo, sizeof(siginfo));
 	assert(ret == sizeof(siginfo));
-	sd_dprintf("signal %d\n", siginfo.ssi_signo);
+	sd_dprintf("signal %d", siginfo.ssi_signo);
 	switch (siginfo.ssi_signo) {
 	case SIGTERM:
 		sys->status = SD_STATUS_KILLED;
 		break;
 	default:
-		sd_eprintf("signal %d unhandled\n", siginfo.ssi_signo);
+		sd_eprintf("signal %d unhandled", siginfo.ssi_signo);
 		break;
 	}
 }
@@ -158,21 +159,32 @@ static int init_signal(void)
 
 	sigfd = signalfd(-1, &mask, SFD_NONBLOCK);
 	if (sigfd < 0) {
-		sd_eprintf("failed to create a signal fd: %m\n");
+		sd_eprintf("failed to create a signal fd: %m");
 		return -1;
 	}
 
 	ret = register_event(sigfd, signal_handler, NULL);
 	if (ret) {
-		sd_eprintf("failed to register signal handler (%d)\n", ret);
+		sd_eprintf("failed to register signal handler (%d)", ret);
 		return -1;
 	}
 
-	sd_dprintf("register signal_handler for %d\n", sigfd);
+	sd_dprintf("register signal_handler for %d", sigfd);
 
 	return 0;
 }
 
+static void crash_handler(int signo)
+{
+	sd_printf(SDOG_EMERG, "sheep exits unexpectedly (%s).",
+		  strsignal(signo));
+
+	sd_backtrace();
+	sd_dump_variable(__sys);
+
+	reraise_crash_signal(signo, 1);
+}
+
 static struct cluster_info __sys;
 struct cluster_info *sys = &__sys;
 
@@ -223,13 +235,12 @@ static void object_cache_dir_set(char *s)
 	char *p = s;
 
 	p = p + strlen("dir=");
-	sprintf(ocpath, "%s", p);
+	snprintf(ocpath, sizeof(ocpath), "%s", p);
 }
 
 static void _object_cache_set(char *s)
 {
 	int i;
-	static bool first = true;
 
 	struct object_cache_arg {
 		const char *name;
@@ -243,12 +254,6 @@ static void _object_cache_set(char *s)
 		{ NULL, NULL },
 	};
 
-	if (first) {
-		assert(!strcmp(s, "object"));
-		first = false;
-		return;
-	}
-
 	for (i = 0; object_cache_args[i].name; i++) {
 		const char *n = object_cache_args[i].name;
 
@@ -262,52 +267,14 @@ static void _object_cache_set(char *s)
 	exit(1);
 }
 
-static void object_cache_set(char *s)
-{
-	sys->enabled_cache_type |= CACHE_TYPE_OBJECT;
-	parse_arg(s, ":", _object_cache_set);
-}
-
-static void disk_cache_set(char *s)
-{
-	assert(!strcmp(s, "disk"));
-	sys->enabled_cache_type |= CACHE_TYPE_DISK;
-}
-
-static void do_cache_type(char *s)
-{
-	int i;
-
-	struct cache_type {
-		const char *name;
-		void (*set)(char *);
-	};
-	struct cache_type cache_types[] = {
-		{ "object", object_cache_set },
-		{ "disk", disk_cache_set },
-		{ NULL, NULL },
-	};
-
-	for (i = 0; cache_types[i].name; i++) {
-		const char *n = cache_types[i].name;
-
-		if (!strncmp(s, n, strlen(n))) {
-			cache_types[i].set(s);
-			return;
-		}
-	}
-
-	fprintf(stderr, "invalid cache type: %s\n", s);
-	exit(1);
-}
-
-static void init_cache_type(char *arg)
+static void object_cache_set(char *arg)
 {
+	sys->enable_object_cache = true;
 	sys->object_cache_size = 0;
 
-	parse_arg(arg, ",", do_cache_type);
+	parse_arg(arg, ",", _object_cache_set);
 
-	if (is_object_cache_enabled() && sys->object_cache_size == 0) {
+	if (sys->object_cache_size == 0) {
 		fprintf(stderr, "object cache size is not set\n");
 		exit(1);
 	}
@@ -325,7 +292,7 @@ static void init_journal_arg(char *arg)
 
 	if (!strncmp(d, arg, dl)) {
 		arg += dl;
-		sprintf(jpath, "%s", arg);
+		snprintf(jpath, sizeof(jpath), "%s", arg);
 	} else if (!strncmp(sz, arg, szl)) {
 		arg += szl;
 		jsize = strtoll(arg, NULL, 10);
@@ -363,24 +330,42 @@ static void init_io_arg(char *arg)
 	}
 }
 
-static int init_work_queues(void)
+static size_t get_nr_nodes(void)
 {
-	if (init_wqueue_eventfd())
+	struct vnode_info *vinfo;
+	size_t nr = 1;
+
+	vinfo = get_vnode_info();
+	if (vinfo != NULL)
+		nr = vinfo->nr_nodes;
+	put_vnode_info(vinfo);
+
+	return nr;
+}
+
+static int create_work_queues(void)
+{
+	if (init_work_queue(get_nr_nodes, trace_register_thread,
+			    trace_unregister_thread))
 		return -1;
 
-	sys->gateway_wqueue = init_work_queue("gway", false);
-	sys->io_wqueue = init_work_queue("io", false);
-	sys->recovery_wqueue = init_work_queue("rw", false);
-	sys->deletion_wqueue = init_work_queue("deletion", true);
-	sys->block_wqueue = init_work_queue("block", true);
-	sys->sockfd_wqueue = init_work_queue("sockfd", true);
-	if (is_object_cache_enabled()) {
-		sys->reclaim_wqueue = init_work_queue("reclaim", true);
-		if (!sys->reclaim_wqueue)
+	sys->gateway_wqueue = create_work_queue("gway", WQ_UNLIMITED);
+	sys->io_wqueue = create_work_queue("io", WQ_UNLIMITED);
+	sys->recovery_wqueue = create_ordered_work_queue("rw");
+	sys->deletion_wqueue = create_ordered_work_queue("deletion");
+	sys->block_wqueue = create_ordered_work_queue("block");
+	sys->sockfd_wqueue = create_ordered_work_queue("sockfd");
+	sys->md_wqueue = create_ordered_work_queue("md");
+	if (sys->enable_object_cache) {
+		sys->oc_reclaim_wqueue =
+			create_ordered_work_queue("oc_reclaim");
+		sys->oc_push_wqueue = create_work_queue("oc_push", WQ_DYNAMIC);
+		if (!sys->oc_reclaim_wqueue || !sys->oc_push_wqueue)
 			return -1;
 	}
 	if (!sys->gateway_wqueue || !sys->io_wqueue || !sys->recovery_wqueue ||
-	    !sys->deletion_wqueue || !sys->block_wqueue || !sys->sockfd_wqueue)
+	    !sys->deletion_wqueue || !sys->block_wqueue ||
+	    !sys->sockfd_wqueue || !sys->md_wqueue)
 			return -1;
 	return 0;
 }
@@ -400,36 +385,125 @@ static void check_host_env(void)
 	struct rlimit r;
 
 	if (getrlimit(RLIMIT_NOFILE, &r) < 0)
-		sd_eprintf("failed to get nofile %m\n");
+		sd_eprintf("failed to get nofile %m");
 	/*
 	 * 1024 is default for NOFILE on most distributions, which is very
 	 * dangerous to run Sheepdog cluster.
 	 */
 	else if (r.rlim_cur == 1024)
 		sd_eprintf("WARN: Allowed open files 1024 too small, "
-			   "suggested %u\n", SD_RLIM_NOFILE);
+			   "suggested %u", SD_RLIM_NOFILE);
 	else if (r.rlim_cur < SD_RLIM_NOFILE)
-		sd_iprintf("Allowed open files %lu, suggested %u\n",
-			   r.rlim_cur, SD_RLIM_NOFILE);
+		sd_iprintf("Allowed open files %lu, suggested %u", r.rlim_cur,
+			   SD_RLIM_NOFILE);
 
 	if (getrlimit(RLIMIT_CORE, &r) < 0)
-		sd_eprintf("failed to get core %m\n");
+		sd_eprintf("failed to get core %m");
 	else if (r.rlim_cur < RLIM_INFINITY)
-		sd_iprintf("Allowed core file size %lu, suggested unlimited\n",
+		sd_iprintf("Allowed core file size %lu, suggested unlimited",
 			   r.rlim_cur);
 }
 
+static int lock_and_daemon(bool daemonize, const char *base_dir)
+{
+	int ret, devnull_fd = 0, status = 0;
+	int pipefd[2];
+
+	ret = pipe(pipefd);
+	if (ret < 0)
+		panic("pipe() for passing exit status failed: %m");
+
+	if (daemonize) {
+		switch (fork()) {
+		case 0:
+			break;
+		case -1:
+			panic("fork() failed during daemonize: %m");
+			break;
+		default:
+			ret = read(pipefd[0], &status, sizeof(status));
+			if (ret != sizeof(status))
+				panic("read exit status failed: %m");
+
+			exit(status);
+			break;
+		}
+
+		if (setsid() == -1) {
+			sd_eprintf("becoming a leader of a new session"
+				" failed: %m");
+			status = 1;
+			goto end;
+		}
+
+		switch (fork()) {
+		case 0:
+			break;
+		case -1:
+			sd_eprintf("fork() failed during daemonize: %m");
+			status = 1;
+			goto end;
+		default:
+			exit(0);
+			break;
+		}
+
+		if (chdir("/")) {
+			sd_eprintf("chdir to / failed: %m");
+			status = 1;
+			goto end;
+		}
+
+		devnull_fd = open("/dev/null", O_RDWR);
+		if (devnull_fd < 0) {
+			sd_eprintf("opening /dev/null failed: %m");
+			status = 1;
+			goto end;
+		}
+	}
+
+	ret = lock_base_dir(base_dir);
+	if (ret < 0) {
+		sd_eprintf("locking directory: %s failed", base_dir);
+		status = 1;
+		goto end;
+	}
+
+	if (daemonize) {
+		/*
+		 * now we can use base_dir/sheep.log for logging error messages,
+		 * we can close 0, 1, and 2 safely
+		 */
+		dup2(devnull_fd, 0);
+		dup2(devnull_fd, 1);
+		dup2(devnull_fd, 2);
+
+		close(devnull_fd);
+	}
+
+end:
+	ret = write(pipefd[1], &status, sizeof(status));
+	if (ret != sizeof(status))
+		panic("writing exit status failed: %m");
+
+	return status;
+}
+
 int main(int argc, char **argv)
 {
 	int ch, longindex, ret, port = SD_LISTEN_PORT, io_port = SD_LISTEN_PORT;
 	int log_level = SDOG_INFO, nr_vnodes = SD_DEFAULT_VNODES;
 	const char *dirp = DEFAULT_OBJECT_DIR, *short_options;
-	char *dir, *p, *pid_file = NULL, *bindaddr = NULL, path[PATH_MAX];
+	char *dir, *p, *pid_file = NULL, *bindaddr = NULL, path[PATH_MAX],
+	     *argp = NULL;
 	bool is_daemon = true, to_stdout = false, explicit_addr = false;
-	int64_t zone = -1, free_space = 0;
+	int64_t zone = -1;
 	struct cluster_driver *cdrv;
 	struct option *long_options;
+	const char *log_format = "default";
+	static struct logger_user_info sheep_info;
 
+	install_crash_handler(crash_handler);
 	signal(SIGPIPE, SIG_IGN);
 
 	long_options = build_long_options(sheep_options);
@@ -439,7 +513,8 @@ int main(int argc, char **argv)
 		switch (ch) {
 		case 'p':
 			port = strtol(optarg, &p, 10);
-			if (optarg == p || port < 1 || port > UINT16_MAX) {
+			if (optarg == p || port < 1 || UINT16_MAX < port
+				|| *p != '\0') {
 				fprintf(stderr, "Invalid port number '%s'\n",
 					optarg);
 				exit(1);
@@ -454,13 +529,16 @@ int main(int argc, char **argv)
 		case 'l':
 			log_level = strtol(optarg, &p, 10);
 			if (optarg == p || log_level < SDOG_EMERG ||
-			    log_level > SDOG_DEBUG) {
+			    SDOG_DEBUG < log_level || *p != '\0') {
 				fprintf(stderr, "Invalid log level '%s'\n",
 					optarg);
 				sdlog_help();
 				exit(1);
 			}
 			break;
+		case 'n':
+			sys->nosync = true;
+			break;
 		case 'y':
 			if (!str_to_addr(optarg, sys->this_node.nid.addr)) {
 				fprintf(stderr, "Invalid address: '%s'\n",
@@ -485,7 +563,8 @@ int main(int argc, char **argv)
 			break;
 		case 'z':
 			zone = strtol(optarg, &p, 10);
-			if (optarg == p || zone < 0 || UINT32_MAX < zone) {
+			if (optarg == p || zone < 0 || UINT32_MAX < zone
+				|| *p != '\0') {
 				fprintf(stderr, "Invalid zone id '%s': "
 					"must be an integer between 0 and %u\n",
 					optarg, UINT32_MAX);
@@ -493,17 +572,6 @@ int main(int argc, char **argv)
 			}
 			sys->this_node.zone = zone;
 			break;
-		case 's':
-			free_space = strtoll(optarg, &p, 10);
-			if (optarg == p || free_space <= 0 ||
-			    UINT64_MAX < free_space) {
-				fprintf(stderr, "Invalid free space size '%s': "
-					"must be an integer between 0 and "
-					"%"PRIu64"\n", optarg, UINT64_MAX);
-				exit(1);
-			}
-			sys->disk_space = free_space * 1024 * 1024;
-			break;
 		case 'u':
 			sys->upgrade = true;
 			break;
@@ -522,7 +590,7 @@ int main(int argc, char **argv)
 			sys->cdrv_option = get_cdrv_option(sys->cdrv, optarg);
 			break;
 		case 'w':
-			init_cache_type(optarg);
+			object_cache_set(optarg);
 			break;
 		case 'i':
 			parse_arg(optarg, ",", init_io_arg);
@@ -562,18 +630,31 @@ int main(int argc, char **argv)
 				PACKAGE_VERSION);
 			exit(0);
 			break;
+		case 'F':
+			log_format = optarg;
+			break;
 		default:
 			usage(1);
 			break;
 		}
 	}
+
+	/*
+	 * early_log_init() must be called before any calling of
+	 * sd_printf() series
+	 */
+	sheep_info.port = port;
+	early_log_init(log_format, &sheep_info);
+
 	if (nr_vnodes == 0) {
 		sys->gateway_only = true;
 		sys->disk_space = 0;
 	}
 
-	if (optind != argc)
-		dirp = argv[optind];
+	if (optind != argc) {
+		argp = strdup(argv[optind]);
+		dirp = strtok(argv[optind], ",");
+	}
 
 	ret = init_base_path(dirp);
 	if (ret)
@@ -589,10 +670,11 @@ int main(int argc, char **argv)
 
 	srandom(port);
 
-	if (is_daemon && daemon(0, 0))
+	if (lock_and_daemon(is_daemon, dir))
 		exit(1);
 
-	ret = log_init(program_name, LOG_SPACE_SIZE, to_stdout, log_level, path);
+	ret = log_init(program_name, LOG_SPACE_SIZE, to_stdout, log_level,
+		path);
 	if (ret)
 		exit(1);
 
@@ -600,7 +682,12 @@ int main(int argc, char **argv)
 	if (ret)
 		exit(1);
 
-	ret = init_global_pathnames(dir);
+	ret = init_global_pathnames(dir, argp);
+	free(argp);
+	if (ret)
+		exit(1);
+
+	ret = init_config_file();
 	if (ret)
 		exit(1);
 
@@ -628,7 +715,7 @@ int main(int argc, char **argv)
 
 	ret = create_cluster(port, zone, nr_vnodes, explicit_addr);
 	if (ret) {
-		sd_eprintf("failed to create sheepdog cluster\n");
+		sd_eprintf("failed to create sheepdog cluster");
 		exit(1);
 	}
 
@@ -637,7 +724,7 @@ int main(int argc, char **argv)
 		if (!strlen(jpath))
 			/* internal journal */
 			memcpy(jpath, dir, strlen(dir));
-		sd_dprintf("%s, %zu, %d\n", jpath, jsize, jskip);
+		sd_dprintf("%s, %zd, %d", jpath, jsize, jskip);
 		ret = journal_file_init(jpath, jsize, jskip);
 		if (ret)
 			exit(1);
@@ -650,17 +737,15 @@ int main(int argc, char **argv)
 	 * e.g, signal handling, above this call and those need multi-threaded
 	 * environment, for e.g, work queues below.
 	 */
-	ret = init_work_queues();
+	ret = create_work_queues();
 	if (ret)
 		exit(1);
 
-	if (!sys->gateway_only) {
-		ret = init_store_driver();
-		if (ret)
-			exit(1);
-	}
+	ret = init_store_driver(sys->gateway_only);
+	if (ret)
+		exit(1);
 
-	if (is_object_cache_enabled()) {
+	if (sys->enable_object_cache) {
 		if (!strlen(ocpath))
 			/* use object cache internally */
 			memcpy(ocpath, dir, strlen(dir));
@@ -674,8 +759,7 @@ int main(int argc, char **argv)
 		exit(1);
 
 	if (pid_file && (create_pidfile(pid_file) != 0)) {
-		fprintf(stderr, "failed to pid file '%s' - %s\n", pid_file,
-			strerror(errno));
+		fprintf(stderr, "failed to pid file '%s' - %m\n", pid_file);
 		exit(1);
 	}
 
@@ -686,17 +770,23 @@ int main(int argc, char **argv)
 
 	free(dir);
 	check_host_env();
-	sd_printf(SDOG_NOTICE, "sheepdog daemon (version %s) started\n",
-		PACKAGE_VERSION);
+	sd_printf(SDOG_NOTICE, "sheepdog daemon (version %s) started",
+		  PACKAGE_VERSION);
 
 	while (sys->nr_outstanding_reqs != 0 ||
 	       (sys->status != SD_STATUS_KILLED &&
 		sys->status != SD_STATUS_SHUTDOWN))
 		event_loop(-1);
 
-	sd_printf(SDOG_INFO, "shutdown\n");
+	sd_printf(SDOG_INFO, "shutdown");
 
 	leave_cluster();
+
+	if (uatomic_is_true(&sys->use_journal)) {
+		sd_iprintf("cleaning journal file");
+		clean_journal_file(jpath);
+	}
+
 	log_close();
 
 	if (pid_file)
