@@ -14,6 +14,7 @@
 #include <inttypes.h>
 #include <stdbool.h>
 #include <urcu/uatomic.h>
+#include <time.h>
 
 #include "sheepdog_proto.h"
 #include "event.h"
@@ -52,7 +53,7 @@ struct request {
 
 	int refcnt;
 	bool local;
-	int wait_efd;
+	int local_req_efd;
 
 	uint64_t local_oid;
 
@@ -75,36 +76,19 @@ struct cluster_info {
 
 	uint64_t disk_space;
 
-	/*
-	 * List of nodes that were past of the last epoch before a shutdown,
-	 * but failed to join.
-	 */
-	struct list_head failed_nodes;
-
-	/*
-	 * List of nodes that weren't part of the last epoch, but joined
-	 * before restarting the cluster.
-	 */
-	struct list_head delayed_nodes;
-
-	struct list_head pending_block_list;
-	struct list_head pending_notify_list;
-
 	DECLARE_BITMAP(vdi_inuse, SD_NR_VDIS);
 
 	uint8_t nr_copies;
-	int req_efd;
+	int local_req_efd;
 
-	pthread_mutex_t wait_req_lock;
-	struct list_head wait_req_queue;
-	struct list_head wait_rw_queue;
-	struct list_head wait_obj_queue;
+	pthread_mutex_t local_req_lock;
+	struct list_head local_req_queue;
+	struct list_head req_wait_queue;
 	int nr_outstanding_reqs;
 
-	uint32_t recovered_epoch;
-
 	bool gateway_only;
 	bool disable_recovery;
+	bool nosync;
 
 	struct work_queue *gateway_wqueue;
 	struct work_queue *io_wqueue;
@@ -113,23 +97,22 @@ struct cluster_info {
 	struct work_queue *recovery_notify_wqueue;
 	struct work_queue *block_wqueue;
 	struct work_queue *sockfd_wqueue;
-	struct work_queue *reclaim_wqueue;
+	struct work_queue *oc_reclaim_wqueue;
+	struct work_queue *oc_push_wqueue;
+	struct work_queue *md_wqueue;
 
-#define CACHE_TYPE_OBJECT 0x1
-#define CACHE_TYPE_DISK   0x2
-	int enabled_cache_type;
+	bool enable_object_cache;
 
 	uint32_t object_cache_size;
 	bool object_cache_directio;
 
 	uatomic_bool use_journal;
 	bool backend_dio;
-	bool upgrade; /* upgrade data layout before starting service
-		       * if necessary*/
+	/* upgrade data layout before starting service if necessary*/
+	bool upgrade;
 };
 
 struct siocb {
-	uint16_t flags;
 	uint32_t epoch;
 	void *buf;
 	uint32_t length;
@@ -138,17 +121,32 @@ struct siocb {
 
 struct vdi_iocb {
 	const char *name;
+	const char *tag;
 	uint32_t data_len;
 	uint64_t size;
 	uint32_t base_vid;
+	uint32_t snapid;
 	bool create_snapshot;
 	int nr_copies;
 };
 
+struct vdi_info {
+	uint32_t vid;
+	uint32_t free_bit;
+	uint64_t create_time;
+};
+
+struct vdi_state {
+	uint32_t vid;
+	uint8_t nr_copies;
+	uint8_t snapshot;
+	uint16_t _pad;
+};
+
 struct store_driver {
 	struct list_head list;
 	const char *name;
-	int (*init)(const char *path);
+	int (*init)(void);
 	bool (*exist)(uint64_t oid);
 	/* create_and_write must be an atomic operation*/
 	int (*create_and_write)(uint64_t oid, const struct siocb *);
@@ -156,40 +154,39 @@ struct store_driver {
 	int (*read)(uint64_t oid, const struct siocb *);
 	int (*format)(void);
 	int (*remove_object)(uint64_t oid);
+	int (*get_hash)(uint64_t oid, uint32_t epoch, uint8_t *sha1);
 	/* Operations in recovery */
 	int (*link)(uint64_t oid, uint32_t tgt_epoch);
-	int (*begin_recover)(const struct siocb *);
-	int (*end_recover)(uint32_t epoch,
-			   const struct vnode_info *old_vnode_info);
+	int (*update_epoch)(uint32_t epoch);
 	int (*purge_obj)(void);
 	/* Operations for snapshot */
-	int (*snapshot)(const struct siocb *);
 	int (*cleanup)(void);
-	int (*restore)(const struct siocb *);
-	int (*get_snap_file)(struct siocb *);
-	int (*flush)(void);
 };
 
-int default_init(const char *p);
+int default_init(void);
 bool default_exist(uint64_t oid);
 int default_create_and_write(uint64_t oid, const struct siocb *iocb);
 int default_write(uint64_t oid, const struct siocb *iocb);
 int default_read(uint64_t oid, const struct siocb *iocb);
 int default_link(uint64_t oid, uint32_t tgt_epoch);
-int default_end_recover(uint32_t old_epoch,
-			const struct vnode_info *old_vnode_info);
+int default_update_epoch(uint32_t epoch);
 int default_cleanup(void);
 int default_format(void);
 int default_remove_object(uint64_t oid);
+int default_get_hash(uint64_t oid, uint32_t epoch, uint8_t *sha1);
 int default_purge_obj(void);
-int for_each_object_in_wd(int (*func)(uint64_t oid, void *arg), bool cleanup,
-			  void *arg);
-int err_to_sderr(uint64_t oid, int err);
+int for_each_object_in_wd(int (*func)(uint64_t, char *, uint32_t, void *), bool,
+			  void *);
+int for_each_object_in_stale(int (*func)(uint64_t oid, char *path,
+					 uint32_t epoch, void *arg),
+			     void *arg);
+int for_each_obj_path(int (*func)(char *path));
 
-struct list_head store_drivers;
-#define add_store_driver(driver)                                 \
-static void __attribute__((constructor)) add_ ## driver(void) {  \
-	list_add(&driver.list, &store_drivers);                  \
+extern struct list_head store_drivers;
+#define add_store_driver(driver)				\
+static void __attribute__((constructor)) add_ ## driver(void)	\
+{								\
+	list_add(&driver.list, &store_drivers);			\
 }
 
 static inline struct store_driver *find_store_driver(const char *name)
@@ -206,10 +203,7 @@ static inline struct store_driver *find_store_driver(const char *name)
 extern struct cluster_info *sys;
 extern struct store_driver *sd_store;
 extern char *obj_path;
-extern char *jrnl_path;
 extern char *epoch_path;
-extern mode_t def_fmode;
-extern mode_t def_dmode;
 
 /* One should call this function to get sys->epoch outside main thread */
 static inline uint32_t sys_epoch(void)
@@ -217,29 +211,32 @@ static inline uint32_t sys_epoch(void)
 	return uatomic_read(&sys->epoch);
 }
 
+static inline bool is_aligned_to_pagesize(void *p)
+{
+	return ((uintptr_t)p & (getpagesize() - 1)) == 0;
+}
+
 int create_listen_port(char *bindaddr, int port);
 int init_unix_domain_socket(const char *dir);
 
-int init_store_driver(void);
-int init_global_pathnames(const char *d);
+int init_store_driver(bool is_gateway);
+int init_global_pathnames(const char *d, char *);
 int init_base_path(const char *dir);
-int init_obj_path(const char *d);
 int init_disk_space(const char *d);
+int lock_base_dir(const char *d);
 
-int fill_vdi_copy_list(void *data);
+int fill_vdi_state_list(void *data);
+bool oid_is_readonly(uint64_t oid);
 int get_vdi_copy_number(uint32_t vid);
 int get_obj_copy_number(uint64_t oid, int nr_zones);
 int get_max_copy_number(void);
 int get_req_copy_number(struct request *req);
-int add_vdi_copy_number(uint32_t vid, int nr_copies);
+int add_vdi_state(uint32_t vid, int nr_copies, bool snapshot);
 int vdi_exist(uint32_t vid);
-int add_vdi(struct vdi_iocb *iocb, uint32_t *new_vid);
-
-int del_vdi(struct request *req, char *data, int data_len, uint32_t *vid,
-	    uint32_t snapid, unsigned int *nr_copies);
-
-int lookup_vdi(const char *name, const char *tag, uint32_t *vid,
-	       uint32_t snapid, unsigned int *nr_copies, uint64_t *ctime);
+int vdi_create(struct vdi_iocb *iocb, uint32_t *new_vid);
+int vdi_delete(struct vdi_iocb *iocb, struct request *req);
+int vdi_lookup(struct vdi_iocb *iocb, struct vdi_info *info);
+void clean_vdi_state(void);
 
 int read_vdis(char *data, int len, unsigned int *rsp_len);
 
@@ -254,16 +251,17 @@ bool have_enough_zones(void);
 struct vnode_info *grab_vnode_info(struct vnode_info *vnode_info);
 struct vnode_info *get_vnode_info(void);
 void put_vnode_info(struct vnode_info *vinfo);
-struct vnode_info *get_vnode_info_epoch(uint32_t epoch);
+struct vnode_info *alloc_vnode_info(const struct sd_node *nodes,
+				    size_t nr_nodes);
+struct vnode_info *get_vnode_info_epoch(uint32_t epoch,
+					struct vnode_info *cur_vinfo);
 void wait_get_vdis_done(void);
 
 int get_nr_copies(struct vnode_info *vnode_info);
 
-void resume_pending_requests(void);
-void resume_wait_epoch_requests(void);
-void resume_wait_obj_requests(uint64_t oid);
-void resume_wait_recovery_requests(void);
-void flush_wait_obj_requests(void);
+void wakeup_requests_on_epoch(void);
+void wakeup_requests_on_oid(uint64_t oid);
+void wakeup_all_requests(void);
 void resume_suspended_recovery(void);
 
 int create_cluster(int port, int64_t zone, int nr_vnodes,
@@ -282,34 +280,36 @@ int set_cluster_flags(uint16_t flags);
 int get_cluster_flags(uint16_t *flags);
 int set_cluster_store(const char *name);
 int get_cluster_store(char *buf);
-int set_cluster_space(uint64_t space);
-int get_cluster_space(uint64_t *space);
+int set_node_space(uint64_t space);
+int get_node_space(uint64_t *space);
 
 int store_file_write(void *buffer, size_t len);
 void *store_file_read(void);
 
 int epoch_log_read(uint32_t epoch, struct sd_node *nodes, int len);
-int epoch_log_read_remote(uint32_t epoch, struct sd_node *nodes, int len);
+int epoch_log_read_with_timestamp(uint32_t epoch, struct sd_node *nodes,
+				int len, time_t *timestamp);
+int epoch_log_read_remote(uint32_t epoch, struct sd_node *nodes, int len,
+			  time_t *timestamp, struct vnode_info *vinfo);
 uint32_t get_latest_epoch(void);
-int init_config_path(const char *base_path);
+void init_config_path(const char *base_path);
+int init_config_file(void);
 int set_cluster_ctime(uint64_t ctime);
 uint64_t get_cluster_ctime(void);
-int get_obj_list(const struct sd_list_req *, struct sd_list_rsp *, void *);
+int get_obj_list(const struct sd_req *, struct sd_rsp *, void *);
 int objlist_cache_cleanup(uint32_t vid);
 
-int start_recovery(struct vnode_info *cur_vinfo, struct vnode_info *old_vinfo);
-void resume_recovery_work(void);
+int start_recovery(struct vnode_info *cur_vinfo, struct vnode_info *, bool);
 bool oid_in_recovery(uint64_t oid);
-bool is_recovery_init(void);
 bool node_in_recovery(void);
 
 int read_backend_object(uint64_t oid, char *data, unsigned int datalen,
-		       uint64_t offset, int nr_copies);
+		       uint64_t offset);
 int write_object(uint64_t oid, char *data, unsigned int datalen,
-		 uint64_t offset, uint16_t flags, bool create, int nr_copies);
+		 uint64_t offset, bool create);
 int read_object(uint64_t oid, char *data, unsigned int datalen,
-		uint64_t offset, int nr_copies);
-int remove_object(uint64_t oid, int nr_copies);
+		uint64_t offset);
+int remove_object(uint64_t oid);
 
 int exec_local_req(struct sd_req *rq, void *data);
 void local_req_init(void);
@@ -338,27 +338,14 @@ int do_process_main(const struct sd_op_template *op, const struct sd_req *req,
 int sheep_do_op_work(const struct sd_op_template *op, struct request *req);
 int gateway_to_peer_opcode(int opcode);
 
-/* Journal */
-struct jrnl_descriptor *jrnl_begin(const void *buf, size_t count, off_t offset,
-				   const char *path, const char *jrnl_dir);
-int jrnl_end(struct jrnl_descriptor *jd);
-int jrnl_recover(const char *jrnl_dir);
-
-static inline bool is_myself(const uint8_t *addr, uint16_t port)
-{
-	return (memcmp(addr, sys->this_node.nid.addr,
-		       sizeof(sys->this_node.nid.addr)) == 0) &&
-		port == sys->this_node.nid.port;
-}
-
 static inline bool vnode_is_local(const struct sd_vnode *v)
 {
-	return is_myself(v->nid.addr, v->nid.port);
+	return node_id_cmp(&v->nid, &sys->this_node.nid) == 0;
 }
 
 static inline bool node_is_local(const struct sd_node *n)
 {
-	return is_myself(n->nid.addr, n->nid.port);
+	return node_eq(n, &sys->this_node);
 }
 
 /* gateway operations */
@@ -366,16 +353,12 @@ int gateway_read_obj(struct request *req);
 int gateway_write_obj(struct request *req);
 int gateway_create_and_write_obj(struct request *req);
 int gateway_remove_obj(struct request *req);
-int gateway_flush_nodes(struct request *req);
 
 /* backend store */
 int peer_read_obj(struct request *req);
 int peer_write_obj(struct request *req);
 int peer_create_and_write_obj(struct request *req);
 int peer_remove_obj(struct request *req);
-int peer_flush(struct request *req);
-
-int default_flush(void);
 
 /* object_cache */
 
@@ -386,14 +369,13 @@ bool object_is_cached(uint64_t oid);
 void object_cache_try_to_reclaim(int);
 int object_cache_handle_request(struct request *req);
 int object_cache_write(uint64_t oid, char *data, unsigned int datalen,
-		       uint64_t offset, uint16_t flags, bool create);
+		       uint64_t offset, bool create);
 int object_cache_read(uint64_t oid, char *data, unsigned int datalen,
 		      uint64_t offset);
 int object_cache_flush_vdi(uint32_t vid);
 int object_cache_flush_and_del(const struct request *req);
 void object_cache_delete(uint32_t vid);
 int object_cache_init(const char *p);
-void object_cache_remove(uint64_t oid);
 
 /* store layout migration */
 int sd_migrate_store(int from, int to);
@@ -414,17 +396,25 @@ void sheep_del_sockfd(const struct node_id *, struct sockfd *);
 int sheep_exec_req(const struct node_id *nid, struct sd_req *hdr, void *data);
 bool sheep_need_retry(uint32_t epoch);
 
-static inline bool is_object_cache_enabled(void)
-{
-	return !!(sys->enabled_cache_type & CACHE_TYPE_OBJECT);
-}
-
-static inline bool is_disk_cache_enabled(void)
-{
-	return !!(sys->enabled_cache_type & CACHE_TYPE_DISK);
-}
-
 /* journal_file.c */
 int journal_file_init(const char *path, size_t size, bool skip);
-int journal_file_write(uint64_t oid, const char *buf, size_t size, off_t, bool);
+void clean_journal_file(const char *p);
+int
+journal_write_store(uint64_t oid, const char *buf, size_t size, off_t, bool);
+int journal_remove_object(uint64_t oid);
+
+/* md.c */
+bool md_add_disk(char *path);
+uint64_t md_init_space(void);
+char *md_get_object_path(uint64_t oid);
+int md_handle_eio(char *);
+bool md_exist(uint64_t oid);
+int md_get_stale_path(uint64_t oid, uint32_t epoch, char *path);
+uint32_t md_get_info(struct sd_md_info *info);
+int md_plug_disks(char *disks);
+int md_unplug_disks(char *disks);
+uint64_t md_get_size(uint64_t *used);
+void kick_node_recover(void);
+void update_node_size(struct sd_node *node);
+
 #endif
