@@ -16,24 +16,25 @@
 #include "sheepdog_proto.h"
 #include "sheep_priv.h"
 
-struct vdi_copy_entry {
+struct vdi_state_entry {
 	uint32_t vid;
 	unsigned int nr_copies;
+	bool snapshot;
 	struct rb_node node;
 };
 
 static uint32_t max_copies;
-static struct rb_root vdi_copy_root = RB_ROOT;
-static pthread_rwlock_t vdi_copy_lock = PTHREAD_RWLOCK_INITIALIZER;
+static struct rb_root vdi_state_root = RB_ROOT;
+static pthread_rwlock_t vdi_state_lock = PTHREAD_RWLOCK_INITIALIZER;
 
-static struct vdi_copy_entry *vdi_copy_search(struct rb_root *root,
-					      uint32_t vid)
+static struct vdi_state_entry *vdi_state_search(struct rb_root *root,
+						uint32_t vid)
 {
 	struct rb_node *n = root->rb_node;
-	struct vdi_copy_entry *t;
+	struct vdi_state_entry *t;
 
 	while (n) {
-		t = rb_entry(n, struct vdi_copy_entry, node);
+		t = rb_entry(n, struct vdi_state_entry, node);
 
 		if (vid < t->vid)
 			n = n->rb_left;
@@ -46,16 +47,16 @@ static struct vdi_copy_entry *vdi_copy_search(struct rb_root *root,
 	return NULL;
 }
 
-static struct vdi_copy_entry *vdi_copy_insert(struct rb_root *root,
-					      struct vdi_copy_entry *new)
+static struct vdi_state_entry *vdi_state_insert(struct rb_root *root,
+						struct vdi_state_entry *new)
 {
 	struct rb_node **p = &root->rb_node;
 	struct rb_node *parent = NULL;
-	struct vdi_copy_entry *entry;
+	struct vdi_state_entry *entry;
 
 	while (*p) {
 		parent = *p;
-		entry = rb_entry(parent, struct vdi_copy_entry, node);
+		entry = rb_entry(parent, struct vdi_state_entry, node);
 
 		if (new->vid < entry->vid)
 			p = &(*p)->rb_left;
@@ -70,16 +71,41 @@ static struct vdi_copy_entry *vdi_copy_insert(struct rb_root *root,
 	return NULL; /* insert successfully */
 }
 
-int get_vdi_copy_number(uint32_t vid)
+static bool vid_is_snapshot(uint32_t vid)
 {
-	struct vdi_copy_entry *entry;
+	struct vdi_state_entry *entry;
 
-	pthread_rwlock_rdlock(&vdi_copy_lock);
-	entry = vdi_copy_search(&vdi_copy_root, vid);
-	pthread_rwlock_unlock(&vdi_copy_lock);
+	pthread_rwlock_rdlock(&vdi_state_lock);
+	entry = vdi_state_search(&vdi_state_root, vid);
+	pthread_rwlock_unlock(&vdi_state_lock);
 
 	if (!entry) {
-		sd_eprintf("No VDI copy entry for %" PRIx32 " found\n", vid);
+		sd_eprintf("No VDI entry for %" PRIx32 " found", vid);
+		return 0;
+	}
+
+	return entry->snapshot;
+}
+
+bool oid_is_readonly(uint64_t oid)
+{
+	/* we allow changing snapshot attributes */
+	if (!is_data_obj(oid))
+		return false;
+
+	return vid_is_snapshot(oid_to_vid(oid));
+}
+
+int get_vdi_copy_number(uint32_t vid)
+{
+	struct vdi_state_entry *entry;
+
+	pthread_rwlock_rdlock(&vdi_state_lock);
+	entry = vdi_state_search(&vdi_state_root, vid);
+	pthread_rwlock_unlock(&vdi_state_lock);
+
+	if (!entry) {
+		sd_eprintf("No VDI copy entry for %" PRIx32 " found", vid);
 		return 0;
 	}
 
@@ -88,13 +114,7 @@ int get_vdi_copy_number(uint32_t vid)
 
 int get_obj_copy_number(uint64_t oid, int nr_zones)
 {
-	uint32_t vid;
-	if (is_vdi_attr_obj(oid))
-		vid = attr_oid_to_vid(oid);
-	else
-		vid = oid_to_vid(oid);
-
-	return min(get_vdi_copy_number(vid), nr_zones);
+	return min(get_vdi_copy_number(oid_to_vid(oid)), nr_zones);
 }
 
 int get_req_copy_number(struct request *req)
@@ -111,73 +131,84 @@ int get_req_copy_number(struct request *req)
 
 int get_max_copy_number(void)
 {
-	return uatomic_read(&max_copies);
+	int nr_copies = uatomic_read(&max_copies);
+
+	if (nr_copies == 0)
+		nr_copies = sys->nr_copies;
+
+	return nr_copies;
 }
 
-int add_vdi_copy_number(uint32_t vid, int nr_copies)
+int add_vdi_state(uint32_t vid, int nr_copies, bool snapshot)
 {
-	struct vdi_copy_entry *entry, *old;
+	struct vdi_state_entry *entry, *old;
 
 	entry = xzalloc(sizeof(*entry));
 	entry->vid = vid;
 	entry->nr_copies = nr_copies;
+	entry->snapshot = snapshot;
 
-	sd_dprintf("%" PRIx32 ", %d\n", vid, nr_copies);
+	sd_dprintf("%" PRIx32 ", %d", vid, nr_copies);
 
-	pthread_rwlock_wrlock(&vdi_copy_lock);
-	old = vdi_copy_insert(&vdi_copy_root, entry);
+	pthread_rwlock_wrlock(&vdi_state_lock);
+	old = vdi_state_insert(&vdi_state_root, entry);
 	if (old) {
 		free(entry);
 		entry = old;
 		entry->nr_copies = nr_copies;
+		entry->snapshot = snapshot;
 	}
 
 	if (uatomic_read(&max_copies) == 0 ||
 	    nr_copies > uatomic_read(&max_copies))
 		uatomic_set(&max_copies, nr_copies);
-	pthread_rwlock_unlock(&vdi_copy_lock);
+	pthread_rwlock_unlock(&vdi_state_lock);
 
 	return SD_RES_SUCCESS;
 }
 
-int fill_vdi_copy_list(void *data)
+int fill_vdi_state_list(void *data)
 {
 	int nr = 0;
 	struct rb_node *n;
-	struct vdi_copy *vc = data;
-	struct vdi_copy_entry *entry;
+	struct vdi_state *vs = data;
+	struct vdi_state_entry *entry;
 
-	pthread_rwlock_rdlock(&vdi_copy_lock);
-	for (n = rb_first(&vdi_copy_root); n; n = rb_next(n)) {
-		entry = rb_entry(n, struct vdi_copy_entry, node);
-		vc->vid = entry->vid;
-		vc->nr_copies = entry->nr_copies;
-		vc++;
+	pthread_rwlock_rdlock(&vdi_state_lock);
+	for (n = rb_first(&vdi_state_root); n; n = rb_next(n)) {
+		entry = rb_entry(n, struct vdi_state_entry, node);
+		memset(vs, 0, sizeof(*vs));
+		vs->vid = entry->vid;
+		vs->nr_copies = entry->nr_copies;
+		vs->snapshot = entry->snapshot;
+		vs++;
 		nr++;
 	}
-	pthread_rwlock_unlock(&vdi_copy_lock);
+	pthread_rwlock_unlock(&vdi_state_lock);
 
-	return nr * sizeof(*vc);
+	return nr * sizeof(*vs);
+}
+
+static inline bool vdi_is_deleted(struct sd_inode *inode)
+{
+	return *inode->name == '\0';
 }
 
 int vdi_exist(uint32_t vid)
 {
-	struct sheepdog_inode *inode;
+	struct sd_inode *inode;
 	int ret = 1;
-	int nr_copies;
 
 	inode = xzalloc(sizeof(*inode));
-	nr_copies = get_vdi_copy_number(vid);
-
 	ret = read_object(vid_to_vdi_oid(vid), (char *)inode,
-			  sizeof(*inode), 0, nr_copies);
+			  sizeof(*inode), 0);
 	if (ret != SD_RES_SUCCESS) {
-		sd_eprintf("fail to read vdi inode (%" PRIx32 ")\n", vid);
+		sd_eprintf("fail to read vdi inode (%" PRIx32 ")", vid);
 		ret = 0;
 		goto out;
 	}
 
-	if (*inode->name == '\0') {
+	if (vdi_is_deleted(inode)) {
 		ret = 0;
 		goto out;
 	}
@@ -189,10 +220,10 @@ out:
 
 /* TODO: should be performed atomically */
 static int create_vdi_obj(struct vdi_iocb *iocb, uint32_t new_vid,
-			  uint32_t cur_vid, uint32_t snapid)
+			  uint32_t cur_vid)
 {
 	/* we are not called concurrently */
-	struct sheepdog_inode *new = NULL, *base = NULL, *cur = NULL;
+	struct sd_inode *new = NULL, *base = NULL, *cur = NULL;
 	struct timeval tv;
 	int ret = SD_RES_NO_MEM;
 	unsigned long block_size = SD_DATA_OBJ_SIZE;
@@ -207,7 +238,7 @@ static int create_vdi_obj(struct vdi_iocb *iocb, uint32_t new_vid,
 
 	if (iocb->base_vid) {
 		ret = read_object(vid_to_vdi_oid(iocb->base_vid), (char *)base,
-				  sizeof(*base), 0, 0);
+				  sizeof(*base), 0);
 		if (ret != SD_RES_SUCCESS) {
 			ret = SD_RES_BASE_VDI_READ;
 			goto out;
@@ -219,12 +250,12 @@ static int create_vdi_obj(struct vdi_iocb *iocb, uint32_t new_vid,
 	if (iocb->create_snapshot) {
 		if (cur_vid != iocb->base_vid) {
 			sd_printf(SDOG_INFO, "tree snapshot %s %" PRIx32 " %"
-				PRIx32 "\n", name, cur_vid, iocb->base_vid);
+				  PRIx32, name, cur_vid, iocb->base_vid);
 
 			ret = read_object(vid_to_vdi_oid(cur_vid), (char *)cur,
-					  SD_INODE_HEADER_SIZE, 0, 0);
+					  SD_INODE_HEADER_SIZE, 0);
 			if (ret != SD_RES_SUCCESS) {
-				sd_printf(SDOG_ERR, "failed\n");
+				sd_printf(SDOG_ERR, "failed");
 				ret = SD_RES_BASE_VDI_READ;
 				goto out;
 			}
@@ -241,7 +272,7 @@ static int create_vdi_obj(struct vdi_iocb *iocb, uint32_t new_vid,
 	new->copy_policy = 0;
 	new->nr_copies = iocb->nr_copies;
 	new->block_size_shift = find_next_bit(&block_size, BITS_PER_LONG, 0);
-	new->snap_id = snapid;
+	new->snap_id = iocb->snapid;
 
 	if (iocb->base_vid) {
 		int i;
@@ -264,9 +295,9 @@ static int create_vdi_obj(struct vdi_iocb *iocb, uint32_t new_vid,
 
 	if (iocb->create_snapshot && cur_vid != iocb->base_vid) {
 		ret = write_object(vid_to_vdi_oid(cur_vid), (char *)cur,
-				   SD_INODE_HEADER_SIZE, 0, 0, false, 0);
+				   SD_INODE_HEADER_SIZE, 0, false);
 		if (ret != 0) {
-			sd_printf(SDOG_ERR, "failed\n");
+			sd_printf(SDOG_ERR, "failed");
 			ret = SD_RES_BASE_VDI_READ;
 			goto out;
 		}
@@ -274,16 +305,16 @@ static int create_vdi_obj(struct vdi_iocb *iocb, uint32_t new_vid,
 
 	if (iocb->base_vid) {
 		ret = write_object(vid_to_vdi_oid(iocb->base_vid), (char *)base,
-				   SD_INODE_HEADER_SIZE, 0, 0, false, 0);
+				   SD_INODE_HEADER_SIZE, 0, false);
 		if (ret != 0) {
-			sd_printf(SDOG_ERR, "failed\n");
+			sd_printf(SDOG_ERR, "failed");
 			ret = SD_RES_BASE_VDI_WRITE;
 			goto out;
 		}
 	}
 
 	ret = write_object(vid_to_vdi_oid(new_vid), (char *)new, sizeof(*new),
-			   0, 0, true, iocb->nr_copies);
+			   0, true);
 	if (ret != 0)
 		ret = SD_RES_VDI_WRITE;
 
@@ -294,232 +325,228 @@ out:
 	return ret;
 }
 
-static int find_first_vdi(unsigned long start, unsigned long end,
-			  const char *name, const char *tag, uint32_t snapid,
-			  uint32_t *vid, unsigned long *deleted_nr,
-			  uint32_t *next_snap, unsigned int *inode_nr_copies,
-			  uint64_t *create_time)
+/*
+ * Return SUCCESS (range of bits set):
+ * Iff we get a bitmap range [left, right) that VDI might be set between. if
+ * right < start, this means a wrap around case where we should examine the
+ * two split ranges, [left, SD_NR_VDIS - 1] and [0, right). 'Right' is the free
+ * bit that might be used by newly created VDI.
+ *
+ * Otherwise:
+ * Return NO_VDI (bit not set) or FULL_VDI (bitmap fully set)
+ */
+static int get_vdi_bitmap_range(const char *name, unsigned long *left,
+				unsigned long *right)
 {
-	struct sheepdog_inode *inode = NULL;
-	unsigned long i;
-	int ret = SD_RES_NO_MEM;
+	*left = fnv_64a_buf(name, strlen(name), FNV1A_64_INIT) &
+		(SD_NR_VDIS - 1);
+	*right = find_next_zero_bit(sys->vdi_inuse, SD_NR_VDIS, *left);
+	if (*left == *right)
+		return SD_RES_NO_VDI;
+
+	if (*right == SD_NR_VDIS) {
+		/* Wrap around */
+		*right = find_next_zero_bit(sys->vdi_inuse, SD_NR_VDIS, 0);
+		if (*right == SD_NR_VDIS)
+			return SD_RES_FULL_VDI;
+	}
+	return SD_RES_SUCCESS;
+}
+
+static inline bool vdi_has_tag(struct vdi_iocb *iocb)
+{
+	if ((iocb->tag && iocb->tag[0]) || iocb->snapid)
+		return true;
+	return false;
+}
+
+static inline bool vdi_tag_match(struct vdi_iocb *iocb,
+			     struct sd_inode *inode)
+{
+	const char *tag = iocb->tag;
+
+	if (inode->tag[0] && !strncmp(inode->tag, tag, sizeof(inode->tag)))
+		return true;
+	if (iocb->snapid == inode->snap_id)
+		return true;
+	return false;
+}
+
+static int fill_vdi_info_range(uint32_t left, uint32_t right,
+			      struct vdi_iocb *iocb, struct vdi_info *info)
+{
+	struct sd_inode *inode;
 	bool vdi_found = false;
-	int nr_copies;
+	int ret;
+	uint32_t i;
+	const char *name = iocb->name;
 
 	inode = malloc(SD_INODE_HEADER_SIZE);
 	if (!inode) {
-		sd_eprintf("failed to allocate memory\n");
+		sd_eprintf("failed to allocate memory");
+		ret = SD_RES_NO_MEM;
 		goto out;
 	}
-
-	for (i = start; i >= end; i--) {
-		nr_copies = get_vdi_copy_number(i);
+	for (i = right - 1; i >= left; i--) {
 		ret = read_object(vid_to_vdi_oid(i), (char *)inode,
-				  SD_INODE_HEADER_SIZE, 0, nr_copies);
-		if (ret != SD_RES_SUCCESS) {
-			ret = SD_RES_EIO;
-			goto out_free_inode;
-		}
+				  SD_INODE_HEADER_SIZE, 0);
+		if (ret != SD_RES_SUCCESS)
+			goto out;
 
-		if (inode->name[0] == '\0') {
-			*deleted_nr = i;
-			continue; /* deleted */
+		if (vdi_is_deleted(inode)) {
+			/* Recycle the deleted inode for fresh vdi create */
+			if (!iocb->create_snapshot)
+				info->free_bit = i;
+			continue;
 		}
 
 		if (!strncmp(inode->name, name, strlen(inode->name))) {
-			*next_snap = inode->snap_id + 1;
-
-			if (!(tag && tag[0]) && !snapid && inode->snap_ctime)
-				break;
-
-			vdi_found = true;
-			if (tag && tag[0] &&
-			    strncmp(inode->tag, tag, sizeof(inode->tag)) != 0)
-				continue;
-			if (snapid && snapid != inode->snap_id)
-				continue;
-
-			*vid = inode->vdi_id;
-			*inode_nr_copies = inode->nr_copies;
-			if (create_time)
-				*create_time = inode->create_time;
-			ret = SD_RES_SUCCESS;
-			goto out_free_inode;
+			sd_dprintf("%s = %s, %u = %u", iocb->tag, inode->tag,
+				   iocb->snapid, inode->snap_id);
+			if (vdi_has_tag(iocb)) {
+				/* Read, delete, clone on snapshots */
+				if (!vdi_is_snapshot(inode)) {
+					vdi_found = true;
+					continue;
+				}
+				if (!vdi_tag_match(iocb, inode))
+					continue;
+			} else {
+				/*
+				 * Rollback & snap create, read, delete on
+				 * current working VDI
+				 */
+				iocb->snapid = inode->snap_id + 1;
+				if (vdi_is_snapshot(inode))
+					/* Current workding VDI is deleted */
+					break;
+			}
+			info->create_time = inode->create_time;
+			info->vid = inode->vdi_id;
+			goto out;
 		}
 	}
-
-	if (vdi_found)
-		ret = SD_RES_NO_TAG;
-	else
-		ret = SD_RES_NO_VDI;
-
-out_free_inode:
-	free(inode);
+	ret = vdi_found ? SD_RES_NO_TAG : SD_RES_NO_VDI;
 out:
+	free(inode);
 	return ret;
 }
 
-static int do_lookup_vdi(const char *name, int namelen, uint32_t *vid,
-			 const char *tag, uint32_t snapid,
-			 uint32_t *next_snapid, unsigned long *right_nr,
-			 unsigned long *deleted_nr, unsigned int *nr_copies,
-			 uint64_t *create_time)
+/* Fill the VDI information from right to left in the bitmap */
+static int fill_vdi_info(unsigned long left, unsigned long right,
+			 struct vdi_iocb *iocb, struct vdi_info *info)
 {
 	int ret;
-	unsigned long nr, start_nr;
 
-	start_nr = fnv_64a_buf(name, namelen, FNV1A_64_INIT) & (SD_NR_VDIS - 1);
+	if (left < right)
+		return fill_vdi_info_range(left, right, iocb, info);
 
-	sd_printf(SDOG_INFO, "looking for %s (%lx)\n", name, start_nr);
-
-	/* bitmap search from the hash point */
-	nr = find_next_zero_bit(sys->vdi_inuse, SD_NR_VDIS, start_nr);
-	*right_nr = nr;
-	if (nr == start_nr) {
-		return SD_RES_NO_VDI;
-	} else if (nr < SD_NR_VDIS) {
-right_side:
-		/* look up on the right side of the hash point */
-		ret = find_first_vdi(nr - 1, start_nr, name,
-				     tag, snapid, vid, deleted_nr, next_snapid,
-				     nr_copies, create_time);
-		return ret;
-	} else {
-		/* round up... bitmap search from the head of the bitmap */
-		nr = find_next_zero_bit(sys->vdi_inuse, SD_NR_VDIS, 0);
-		*right_nr = nr;
-		if (nr >= SD_NR_VDIS)
-			return SD_RES_FULL_VDI;
-		else if (nr) {
-			/* look up on the left side of the hash point */
-			ret = find_first_vdi(nr - 1, 0, name,
-					     tag, snapid, vid, deleted_nr,
-					     next_snapid, nr_copies,
-					     create_time);
-			if (ret == SD_RES_NO_VDI)
-				; /* we need to go to the right side */
-			else
-				return ret;
-		}
-
-		nr = SD_NR_VDIS;
-		goto right_side;
+	ret = fill_vdi_info_range(0, right, iocb, info);
+	switch (ret) {
+	case SD_RES_NO_VDI:
+	case SD_RES_NO_TAG:
+		ret = fill_vdi_info_range(left, SD_NR_VDIS - 1, iocb, info);
+		break;
+	default:
+		break;
 	}
+	return ret;
 }
 
-int lookup_vdi(const char *name, const char *tag, uint32_t *vid,
-	       uint32_t snapid, unsigned int *nr_copies, uint64_t *create_time)
+/* Return SUCCESS if we find targeted VDI specified by iocb and fill info */
+int vdi_lookup(struct vdi_iocb *iocb, struct vdi_info *info)
 {
-	uint32_t dummy0;
-	unsigned long dummy1, dummy2;
+	unsigned long left, right;
+	int ret;
 
-	return do_lookup_vdi(name, strlen(name), vid, tag,
-			     snapid, &dummy0, &dummy1, &dummy2, nr_copies,
-			     create_time);
+
+	ret = get_vdi_bitmap_range(iocb->name, &left, &right);
+	info->free_bit = right;
+	sd_dprintf("%s left %lx right %lx, %x", iocb->name, left, right, ret);
+	switch (ret) {
+	case SD_RES_NO_VDI:
+	case SD_RES_FULL_VDI:
+		return ret;
+	case SD_RES_SUCCESS:
+		break;
+	}
+	return fill_vdi_info(left, right, iocb, info);
 }
 
-static int notify_vdi_add(uint32_t vdi_id, uint32_t nr_copies)
+static int notify_vdi_add(uint32_t vdi_id, uint32_t nr_copies, uint32_t old_vid)
 {
-	struct sd_req hdr;
 	int ret = SD_RES_SUCCESS;
-	char *buf;
+	struct sd_req hdr;
 
 	sd_init_req(&hdr, SD_OP_NOTIFY_VDI_ADD);
-	hdr.flags = SD_FLAG_CMD_WRITE;
-	hdr.data_length = sizeof(vdi_id) + sizeof(nr_copies);
+	hdr.vdi_state.old_vid = old_vid;
+	hdr.vdi_state.new_vid = vdi_id;
+	hdr.vdi_state.copies = nr_copies;
+	hdr.vdi_state.set_bitmap = false;
 
-	buf = xmalloc(sizeof(vdi_id) + sizeof(nr_copies));
-	memcpy(buf, &vdi_id, sizeof(vdi_id));
-	memcpy(buf + sizeof(vdi_id), &nr_copies, sizeof(nr_copies));
-
-	ret = exec_local_req(&hdr, buf);
+	ret = exec_local_req(&hdr, NULL);
 	if (ret != SD_RES_SUCCESS)
-		sd_eprintf("fail to notify vdi add event(%" PRIx32 ", %d)\n",
-			vdi_id, nr_copies);
-
-	free(buf);
+		sd_eprintf("fail to notify vdi add event(%" PRIx32 ", %d, %"
+			   PRIx32 ")", vdi_id, nr_copies, old_vid);
 
 	return ret;
 }
 
-int add_vdi(struct vdi_iocb *iocb, uint32_t *new_vid)
+/*
+ * There are 3 create operation in SD:
+ * 1. fresh create (expect NO_VDI returned from vdi_lookup)
+ * 2. snapshot create (expect SUCCESS)
+ * 3. rollback create (expect NO_VDI)
+ *
+ * Fresh create started with id = 1. Both rollback & snap create started with
+ * current working VDI's snap_id + 1. Working VDI always has the highest snapid.
+ */
+int vdi_create(struct vdi_iocb *iocb, uint32_t *new_vid)
 {
-	uint32_t cur_vid = 0;
-	uint32_t next_snapid = 1;
-	unsigned long nr, deleted_nr = SD_NR_VDIS, right_nr = SD_NR_VDIS;
-	unsigned int dummy;
+	const char *name = iocb->name;
+	struct vdi_info info = {};
 	int ret;
-	const char *name;
 
-	if (iocb->data_len != SD_MAX_VDI_LEN)
-		return SD_RES_INVALID_PARMS;
-
-	name = iocb->name;
-
-	ret = do_lookup_vdi(name, strlen(name), &cur_vid,
-			    NULL, 0, &next_snapid, &right_nr, &deleted_nr,
-			    &dummy, NULL);
-
-	if (iocb->create_snapshot) {
-		if (ret != SD_RES_SUCCESS) {
-			if (ret == SD_RES_NO_VDI)
-				sd_printf(SDOG_CRIT, "VDI %s does not exist\n",
-					name);
+	ret = vdi_lookup(iocb, &info);
+	switch (ret) {
+	case SD_RES_SUCCESS:
+		if (!iocb->create_snapshot)
+			return SD_RES_VDI_EXIST;
+		break;
+	case SD_RES_NO_VDI:
+		if (iocb->create_snapshot)
 			return ret;
-		}
-		nr = right_nr;
-	} else {
-		/* we already have the same VDI or met other errors. */
-		if (ret != SD_RES_NO_VDI) {
-			if (ret == SD_RES_SUCCESS)
-				ret = SD_RES_VDI_EXIST;
-			return ret;
-		}
-
-		if (deleted_nr == SD_NR_VDIS)
-			nr = right_nr;
-		else
-			nr = deleted_nr; /* we can recycle a deleted VDI */
+		break;
+	default:
+		sd_eprintf("%s", sd_strerror(ret));
+		return ret;
 	}
+	if (!iocb->snapid)
+		iocb->snapid = 1;
+	*new_vid = info.free_bit;
+	notify_vdi_add(*new_vid, iocb->nr_copies, info.vid);
 
-	*new_vid = nr;
+	sd_dprintf("%s %s: size %" PRIu64 ", vid %" PRIx32 ", base %" PRIx32
+		   ", cur %" PRIx32 ", copies %d, snapid %"PRIu32,
+		   iocb->create_snapshot ? "snapshot" : "vdi", name, iocb->size,
+		   *new_vid, iocb->base_vid, info.vid, iocb->nr_copies,
+		   iocb->snapid);
 
-	notify_vdi_add(nr, iocb->nr_copies);
-
-	sd_printf(SDOG_INFO, "creating new %s %s: size %" PRIu64 ", vid %"
-		PRIx32 ", base %" PRIx32 ", cur %" PRIx32 ", copies %d\n",
-		iocb->create_snapshot ? "snapshot" : "vdi", name, iocb->size,
-		*new_vid, iocb->base_vid, cur_vid, iocb->nr_copies);
-
-	return create_vdi_obj(iocb, *new_vid, cur_vid, next_snapid);
+	return create_vdi_obj(iocb, *new_vid, info.vid);
 }
 
 static int start_deletion(struct request *req, uint32_t vid);
 
-int del_vdi(struct request *req, char *data, int data_len,
-	    uint32_t *vid, uint32_t snapid, unsigned int *nr_copies)
+int vdi_delete(struct vdi_iocb *iocb, struct request *req)
 {
-	const char *name = data, *tag;
-	uint32_t dummy0;
-	unsigned long dummy1, dummy2;
+	struct vdi_info info;
 	int ret;
 
-	if (data_len == SD_MAX_VDI_LEN + SD_MAX_VDI_TAG_LEN)
-		tag = data + SD_MAX_VDI_LEN;
-	else if (data_len == SD_MAX_VDI_LEN)
-		tag = NULL;
-	else {
-		ret = SD_RES_INVALID_PARMS;
-		goto out;
-	}
-
-	ret = do_lookup_vdi(name, strlen(name), vid, tag,
-			    snapid, &dummy0, &dummy1, &dummy2,
-			    nr_copies, NULL);
+	ret = vdi_lookup(iocb, &info);
 	if (ret != SD_RES_SUCCESS)
 		goto out;
 
-	ret = start_deletion(req, *vid);
+	ret = start_deletion(req, info.vid);
 out:
 	return ret;
 }
@@ -539,11 +566,10 @@ struct deletion_work {
 	uint32_t done;
 
 	struct work work;
-	struct list_head dw_siblings;
+	struct list_head list;
 	struct request *req;
 
 	uint32_t vid;
-	int nr_copies;
 
 	int count;
 	uint32_t *buf;
@@ -553,12 +579,12 @@ static LIST_HEAD(deletion_work_list);
 
 static int delete_inode(struct deletion_work *dw)
 {
-	struct sheepdog_inode *inode = NULL;
+	struct sd_inode *inode = NULL;
 	int ret = SD_RES_SUCCESS;
 
 	inode = xzalloc(sizeof(*inode));
 	ret = read_object(vid_to_vdi_oid(dw->vid), (char *)inode,
-			  SD_INODE_HEADER_SIZE, 0, dw->nr_copies);
+			  SD_INODE_HEADER_SIZE, 0);
 	if (ret != SD_RES_SUCCESS) {
 		ret = SD_RES_EIO;
 		goto out;
@@ -567,7 +593,7 @@ static int delete_inode(struct deletion_work *dw)
 	memset(inode->name, 0, sizeof(inode->name));
 
 	ret = write_object(vid_to_vdi_oid(dw->vid), (char *)inode,
-			   SD_INODE_HEADER_SIZE, 0, 0, false, dw->nr_copies);
+			   SD_INODE_HEADER_SIZE, 0, false);
 	if (ret != 0) {
 		ret = SD_RES_EIO;
 		goto out;
@@ -589,8 +615,8 @@ static int notify_vdi_deletion(uint32_t vdi_id)
 
 	ret = exec_local_req(&hdr, &vdi_id);
 	if (ret != SD_RES_SUCCESS)
-		sd_eprintf("fail to notify vdi deletion(%" PRIx32 "), %d\n",
-			vdi_id, ret);
+		sd_eprintf("fail to notify vdi deletion(%" PRIx32 "), %d",
+			   vdi_id, ret);
 
 	return ret;
 }
@@ -600,27 +626,25 @@ static void delete_one(struct work *work)
 	struct deletion_work *dw = container_of(work, struct deletion_work, work);
 	uint32_t vdi_id = *(dw->buf + dw->count - dw->done - 1);
 	int ret, i, nr_deleted;
-	struct sheepdog_inode *inode = NULL;
-	int nr_copies;
+	struct sd_inode *inode = NULL;
 
-	sd_eprintf("%d %d, %16x\n", dw->done, dw->count, vdi_id);
+	sd_dprintf("%d %d, %16x", dw->done, dw->count, vdi_id);
 
 	inode = malloc(sizeof(*inode));
 	if (!inode) {
-		sd_eprintf("failed to allocate memory\n");
-		goto out;
+		sd_eprintf("failed to allocate memory");
+		return;
 	}
 
-	nr_copies = get_vdi_copy_number(vdi_id);
 	ret = read_backend_object(vid_to_vdi_oid(vdi_id),
-			  (void *)inode, sizeof(*inode), 0, nr_copies);
+			  (void *)inode, sizeof(*inode), 0);
 
 	if (ret != SD_RES_SUCCESS) {
-		sd_eprintf("cannot find VDI object\n");
+		sd_eprintf("cannot find VDI object");
 		goto out;
 	}
 
-	if (inode->vdi_size == 0 && inode->name[0] == '\0')
+	if (inode->vdi_size == 0 && vdi_is_deleted(inode))
 		goto out;
 
 	for (nr_deleted = 0, i = 0; i < MAX_DATA_OBJS; i++) {
@@ -633,14 +657,14 @@ static void delete_one(struct work *work)
 
 		if (inode->data_vdi_id[i] != inode->vdi_id) {
 			sd_dprintf("object %" PRIx64 " is base's data, would"
-				" not be deleted.\n", oid);
+				   " not be deleted.", oid);
 			continue;
 		}
 
-		ret = remove_object(oid, nr_copies);
+		ret = remove_object(oid);
 		if (ret != SD_RES_SUCCESS)
-			sd_eprintf("remove object %" PRIx64 " fail, %d\n",
-				oid, ret);
+			sd_eprintf("remove object %" PRIx64 " fail, %d", oid,
+				   ret);
 
 		nr_deleted++;
 	}
@@ -648,14 +672,14 @@ static void delete_one(struct work *work)
 	if (nr_deleted)
 		notify_vdi_deletion(vdi_id);
 
-	if (*(inode->name) == '\0')
+	if (vdi_is_deleted(inode))
 		goto out;
 
 	inode->vdi_size = 0;
 	memset(inode->name, 0, sizeof(inode->name));
 
 	write_object(vid_to_vdi_oid(vdi_id), (void *)inode,
-		     sizeof(*inode), 0, 0, false, nr_copies);
+		     sizeof(*inode), 0, false);
 out:
 	free(inode);
 }
@@ -671,7 +695,7 @@ static void delete_one_done(struct work *work)
 		return;
 	}
 
-	list_del(&dw->dw_siblings);
+	list_del(&dw->list);
 
 	put_request(req);
 
@@ -680,7 +704,7 @@ static void delete_one_done(struct work *work)
 
 	if (!list_empty(&deletion_work_list)) {
 		dw = list_first_entry(&deletion_work_list,
-				      struct deletion_work, dw_siblings);
+				      struct deletion_work, list);
 
 		queue_work(sys->deletion_wqueue, &dw->work);
 	}
@@ -689,30 +713,28 @@ static void delete_one_done(struct work *work)
 static int fill_vdi_list(struct deletion_work *dw, uint32_t root_vid)
 {
 	int ret, i;
-	struct sheepdog_inode *inode = NULL;
+	struct sd_inode *inode = NULL;
 	int done = dw->count;
 	uint32_t vid;
-	int nr_copies;
 
 	inode = malloc(SD_INODE_HEADER_SIZE);
 	if (!inode) {
-		sd_eprintf("failed to allocate memory\n");
+		sd_eprintf("failed to allocate memory");
 		goto err;
 	}
 
 	dw->buf[dw->count++] = root_vid;
 again:
 	vid = dw->buf[done++];
-	nr_copies = get_vdi_copy_number(vid);
 	ret = read_backend_object(vid_to_vdi_oid(vid), (char *)inode,
-			  SD_INODE_HEADER_SIZE, 0, nr_copies);
+			  SD_INODE_HEADER_SIZE, 0);
 
 	if (ret != SD_RES_SUCCESS) {
-		sd_eprintf("cannot find VDI object\n");
+		sd_eprintf("cannot find VDI object");
 		goto err;
 	}
 
-	if (inode->name[0] != '\0' && vid != dw->vid)
+	if (!vdi_is_deleted(inode) && vid != dw->vid)
 		goto out;
 
 	for (i = 0; i < ARRAY_SIZE(inode->child_vdi_id); i++) {
@@ -734,32 +756,31 @@ out:
 
 static uint64_t get_vdi_root(uint32_t vid, bool *cloned)
 {
-	int ret, nr_copies;
-	struct sheepdog_inode *inode = NULL;
+	int ret;
+	struct sd_inode *inode = NULL;
 
 	*cloned = false;
 
 	inode = malloc(SD_INODE_HEADER_SIZE);
 	if (!inode) {
-		sd_eprintf("failed to allocate memory\n");
+		sd_eprintf("failed to allocate memory");
 		vid = 0;
 		goto out;
 	}
 next:
-	nr_copies = get_vdi_copy_number(vid);
 	ret = read_backend_object(vid_to_vdi_oid(vid), (char *)inode,
-			  SD_INODE_HEADER_SIZE, 0, nr_copies);
+			  SD_INODE_HEADER_SIZE, 0);
 
 	if (vid == inode->vdi_id && inode->snap_id == 1
 			&& inode->parent_vdi_id != 0
 			&& !inode->snap_ctime) {
-		sd_dprintf("vdi %" PRIx32 " is a cloned vdi.\n", vid);
+		sd_dprintf("vdi %" PRIx32 " is a cloned vdi.", vid);
 		/* current vdi is a cloned vdi */
 		*cloned = true;
 	}
 
 	if (ret != SD_RES_SUCCESS) {
-		sd_eprintf("cannot find VDI object\n");
+		sd_eprintf("cannot find VDI object");
 		vid = 0;
 		goto out;
 	}
@@ -789,7 +810,6 @@ static int start_deletion(struct request *req, uint32_t vid)
 	dw->count = 0;
 	dw->vid = vid;
 	dw->req = req;
-	dw->nr_copies = get_vdi_copy_number(vid);
 
 	dw->work.fn = delete_one;
 	dw->work.done = delete_one_done;
@@ -802,21 +822,23 @@ static int start_deletion(struct request *req, uint32_t vid)
 
 	ret = fill_vdi_list(dw, root_vid);
 	if (ret) {
-		/* if the VDI is a cloned VDI, delete its objects
-		 * no matter whether the VDI tree is clear. */
+		/*
+		 * if the VDI is a cloned VDI, delete its objects
+		 * no matter whether the VDI tree is clear.
+		 */
 		if (cloned) {
 			dw->buf[0] = vid;
 			dw->count = 1;
 		} else {
 			sd_dprintf("snapshot chain has valid vdi, "
-				"just mark vdi %" PRIx32 " as deleted.\n",
-				dw->vid);
+				   "just mark vdi %" PRIx32 " as deleted.",
+				   dw->vid);
 			delete_inode(dw);
 			return SD_RES_SUCCESS;
 		}
 	}
 
-	sd_dprintf("%d\n", dw->count);
+	sd_dprintf("%d", dw->count);
 
 	if (dw->count == 0)
 		goto out;
@@ -824,10 +846,10 @@ static int start_deletion(struct request *req, uint32_t vid)
 	uatomic_inc(&req->refcnt);
 
 	if (list_empty(&deletion_work_list)) {
-		list_add_tail(&dw->dw_siblings, &deletion_work_list);
+		list_add_tail(&dw->list, &deletion_work_list);
 		queue_work(sys->deletion_wqueue, &dw->work);
 	} else
-		list_add_tail(&dw->dw_siblings, &deletion_work_list);
+		list_add_tail(&dw->list, &deletion_work_list);
 out:
 	return SD_RES_SUCCESS;
 err:
@@ -845,7 +867,7 @@ int get_vdi_attr(struct sheepdog_vdi_attr *vattr, int data_len,
 	struct sheepdog_vdi_attr tmp_attr;
 	uint64_t oid, hval;
 	uint32_t end;
-	int ret, nr_copies;
+	int ret;
 
 	vattr->ctime = create_time;
 
@@ -859,13 +881,12 @@ int get_vdi_attr(struct sheepdog_vdi_attr *vattr, int data_len,
 	end = *attrid - 1;
 	while (*attrid != end) {
 		oid = vid_to_attr_oid(vid, *attrid);
-		nr_copies = get_vdi_copy_number(vid);
 		ret = read_object(oid, (char *)&tmp_attr,
-				  sizeof(tmp_attr), 0, nr_copies);
+				  sizeof(tmp_attr), 0);
 
 		if (ret == SD_RES_NO_OBJ && wr) {
-			ret = write_object(oid, (char *)vattr,
-					   data_len, 0, 0, true, nr_copies);
+			ret = write_object(oid, (char *)vattr, data_len, 0,
+					   true);
 			if (ret)
 				ret = SD_RES_EIO;
 			else
@@ -886,15 +907,15 @@ int get_vdi_attr(struct sheepdog_vdi_attr *vattr, int data_len,
 			else if (delete) {
 				ret = write_object(oid, (char *)"", 1,
 						   offsetof(struct sheepdog_vdi_attr, name),
-						   0, false, nr_copies);
+						   false);
 				if (ret)
 					ret = SD_RES_EIO;
 				else
 					ret = SD_RES_SUCCESS;
 			} else if (wr) {
 				ret = write_object(oid, (char *)vattr,
-						   SD_ATTR_OBJ_SIZE, 0, 0,
-						   false, nr_copies);
+						   SD_ATTR_OBJ_SIZE, 0,
+						   false);
 
 				if (ret)
 					ret = SD_RES_EIO;
@@ -908,8 +929,25 @@ int get_vdi_attr(struct sheepdog_vdi_attr *vattr, int data_len,
 		(*attrid)++;
 	}
 
-	sd_dprintf("there is no space for new VDIs\n");
+	sd_dprintf("there is no space for new VDIs");
 	ret = SD_RES_FULL_VDI;
 out:
 	return ret;
 }
+
+void clean_vdi_state(void)
+{
+	struct rb_node *current_node = rb_first(&vdi_state_root);
+	struct vdi_state_entry *entry = NULL;
+
+	pthread_rwlock_wrlock(&vdi_state_lock);
+	while (current_node) {
+		entry = rb_entry(current_node, struct vdi_state_entry, node);
+		rb_erase(current_node, &vdi_state_root);
+		free(entry);
+		entry = NULL;
+		current_node = rb_first(&vdi_state_root);
+	}
+	INIT_RB_ROOT(&vdi_state_root);
+	pthread_rwlock_unlock(&vdi_state_lock);
+}
